///////////////////////////////////////////////////////////////////////////////////
//No part of this file can be copied or released without the consent of 
//Avalanche Technology
//										
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
//										
//	Avalanche Technology Inc., Proprietary and Confidential	   *
//										
// 	Release:  3.2    Date 8/8/2023 	
//										
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
//  PART DESCRIPTION:
//
//  Technology: 40nm pMTJ STT-MRAM
//  Part:       AS3001204/AS3004204/AS3008204/AS3016204
//              AS1001204/AS1004204/AS1008204/AS1016204
//
//  Description: 1/4/8/16 Megabit HIGH PERFORMANCE SERIAL PERSISTENT SRAM MEMORY
//
////////////////////////////////////////////////////////////////////////////////////
//  FILE CONTENTS : MODEL OF SPSRAM 
//
////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////
//  DENSITY SELECTION                                                             //
//                                                                                //
//  Uncomment one of the following lines to select the appropriate density. If    //
//  no line is selected, the model will default to 16Mb                           //
//                                                                                //
////////////////////////////////////////////////////////////////////////////////////
//`define DENSITY_1M  // 1Mb
//`define DENSITY_4M  // 4Mb
//`define DENSITY_8M  // 8Mb
//`define DENSITY_16M // 16Mb

////////////////////////////////////////////////////////////////////////////////////
//  VOLTAGE SELECTION                                                             //
//                                                                                //
//  Uncomment one of the following lines to select the appropriate voltage. If    //
//  no line is selected, the model will default to 3V                             //
//                                                                                //
////////////////////////////////////////////////////////////////////////////////////
//`define VOLTAGE_18 // 1.8V
//`define VOLTAGE_3  // 3V

////////////////////////////////////////////////////////////////////////////////////
// MODULE DECLARATION                                                             //
////////////////////////////////////////////////////////////////////////////////////

`timescale 100ps / 10ps

//`define DUMMY_TIMING

//module spsram_as3016204
module spsram_asxxxx204
  ( 
  IO3, 
  WPn_IO2, 
  SO_IO1,
  SI_IO0,
  CSn, 
  CLK
  );
  
  inout IO3; 
  inout SI_IO0; 
  inout SO_IO1;
  inout WPn_IO2;
  input CSn;
  input CLK;

  // Instruction type
  parameter NOOP = 8'h00;
  parameter WREN = 8'h06;
  parameter WRDI = 8'h04;
  parameter WRSR = 8'h01;
  parameter RDSR = 8'h05;
  parameter WRCX = 8'h87;
  parameter RDCX = 8'h46;
  parameter READ = 8'h03;
  parameter RDFT = 8'h0b;
  parameter WRTE = 8'h02;
  parameter WRFT = 8'hDA;
  parameter RDAR = 8'h65;
  parameter WRAR = 8'h71;
  parameter SPIE = 8'hFF;
  parameter DPIE = 8'h37;
  parameter QPIE = 8'h38;
  parameter WRAP = 8'h1A;
  parameter RDAP = 8'h14;
  parameter WRAS = 8'h42;
  parameter RDAS = 8'h4B;
  parameter RDSN = 8'hC3;
  parameter WRSN = 8'hC2;
  parameter HBNE = 8'hBA;
  parameter DPDE = 8'hB9;
  parameter DPDX = 8'hAB;
  parameter SRTE = 8'h66;
  parameter SRST = 8'h99;
  parameter RDC1 = 8'h35;
  parameter RDC2 = 8'h3F;
  parameter RDC3 = 8'h44;
  parameter RDC4 = 8'h45;
  parameter RDID = 8'h9F;
  parameter RUID = 8'h4C;
  parameter DRFR = 8'h0D;
  parameter DRFW = 8'hDE;
  parameter RDDO = 8'h3B;
  parameter RDQO = 8'h6B;
  parameter RDDI = 8'hBB;
  parameter RDQI = 8'hEB;
  parameter WDUI = 8'hA2;
  parameter WQDI = 8'h32;
  parameter WDIO = 8'hA1;
  parameter WQIO = 8'hD2;
  parameter DRDI = 8'hBD;
  parameter DRQI = 8'hED;
  parameter DWQI = 8'h31;
  parameter DWQO = 8'hD1;

  //SPI DPI and QPI modes
  parameter SPIEN   = 8'hFF;
  parameter DPIEN   = 8'h37;
  parameter QPIEN   = 8'h38;

  ////////////////////////////////////////////////////////////
  // Timing Violation checks
  ///////////////////////////////////////////////////////////
  //$setup, $hold, $width violations
  reg tCSS_notifier;
  reg tCSH_notifier;
  reg tCH_notifier;
  reg tCL_notifier;
  reg tCS_notifier;
  reg tSU_notifier;
  reg tHD_notifier;

  specify 
    specparam tPOR_time  = 1; 
    specparam tCSS       = 50; 
    specparam tCSH       = 40; 
    specparam tCS        = 1300; 
    specparam tSU        = 20;
    specparam tHD        = 20;
    specparam tCO        = 70;
`ifdef DDR_TIMING
    specparam tCH        = 92.5; 
    specparam tCL        = 92.5; 
`else
    specparam tCH        = 50; 
    specparam tCL        = 50; 
`endif
`ifdef DUMMY_TIMING
    specparam tEXHIB     = 500; //dummy
    specparam tENTHIB    = 500; //dummy
    specparam tCSDPD     = 500;
    specparam tEXDPD     = 500; //dummy
    specparam tEDPD      = 500; //dummy
    specparam tSRST      = 500; //dummy
`else
    specparam tEXHIB     = 4500000;
    specparam tENTHIB    = 30000;
    specparam tCSDPD     = 500;
    specparam tEXDPD     = 4000000;
    specparam tEDPD      = 30000; 
    specparam tSRST      = 500000;
`endif

    $setup (negedge CSn, posedge CLK, tCSS,  tCSS_notifier);
    $hold  (negedge CLK, posedge CSn, tCSH,  tCSH_notifier);

    $width (posedge CLK, tCH, 0, tCH_notifier); 
    $width (negedge CLK, tCL, 0, tCL_notifier);
    $width (posedge CSn, tCS, 0, tCS_notifier);

    $setup (SI_IO0, posedge CLK, tSU,  tSU_notifier);
    $hold  (posedge CLK, SI_IO0, tHD,  tHD_notifier);

  endspecify 

  /* States */
  parameter S_IDLE     = 8'd0;
  parameter S_CMD      = 8'd1;
  parameter S_ADDR     = 8'd2;
  parameter S_DATAIN   = 8'd3;
  parameter S_DATAOUT  = 8'd4;
  parameter S_LATENCY  = 8'd5;
  parameter S_MODE     = 8'd6;
  parameter S_HIBER    = 8'd7;
  parameter S_HBENT    = 8'd8;
  parameter S_HBEXT    = 8'd9;
  parameter S_DEEP     = 8'd10;
  parameter S_DPENT    = 8'd11;
  parameter S_DPEXT    = 8'd12;
  parameter S_DPCMD    = 8'd13;
  parameter SD_ADDR    = 8'd14; // DDR 
  parameter SD_MODE    = 8'd15; // DDR 
  parameter SD_DATAIN  = 8'd16; // DDR
  parameter SD_DATAOUT = 8'd17; // DDR
  parameter S_IGNORE   = 8'hff;
  reg [7:0] state;
  reg s_cmd_done;
  reg s_addr_done;
  reg s_datain_done;
  reg s_dataout_done;
  reg s_latency_done;
  reg s_mode_done;
  reg modesw_valid;

  reg PORn;
  reg [7:0] mode;
  reg si_dq0_reg; 
  reg so_dq1_reg;
  reg wp_dq2_reg;
  reg hold_dq3_reg; 
  reg [3:0] dqout_en;

  /* Address */
  reg [23:0] addr;
  //wire [23:0] pre_addr;
  wire [23:0] sr_addr;
  wire [23:0] sr_addr_ddr;
  localparam ADDR_SR  = 24'h0;
  localparam ADDR_CR1 = 24'h2;
  localparam ADDR_CR2 = 24'h3;
  localparam ADDR_CR3 = 24'h4;
  localparam ADDR_CR4 = 24'h5;
  localparam ADDR_ID  = 24'h30;
  localparam ADDR_UID = 24'h40;

  /* Device config/status registers */
  reg [7:0] SR;
  localparam SR_WRMASK = 8'hFC;
  localparam SR_MAPLK  = 8'hC0;
  localparam SR_DEFAULT = 8'h00;
  reg [7:0] CR1, CR2, CR3, CR4;
  localparam CR1_WRMASK = 8'h05;
  localparam CR2_WRMASK = 8'h0F;
  localparam CR3_WRMASK = 8'hF7;
  localparam CR4_WRMASK = 8'hFF;
  localparam CR1_DEFAULT = 8'h00;
  localparam CR2_DEFAULT = 8'h00;
  localparam CR3_DEFAULT = 8'h00;
  localparam CR4_DEFAULT = 8'h05;
  wire [31:0] IDREG;
`ifdef VOLTAGE_18
  localparam VID = 4'b0010;
`elsif VOLTAGE_3
  localparam VID = 4'b0001;
`else
  localparam VID = 4'b0001;
`endif
`ifdef DENSITY_1M
  assign IDREG = {8'hE6, 4'h0, VID, 4'h0, 4'h1, 8'h1};
`elsif DENSITY_4M
  assign IDREG = {8'hE6, 4'h0, VID, 4'h0, 4'h2, 8'h1};
`elsif DENSITY_8M
  assign IDREG = {8'hE6, 4'h0, VID, 4'h0, 4'h3, 8'h1};
`elsif DENSITY_16M
  assign IDREG = {8'hE6, 4'h0, VID, 4'h0, 4'h4, 8'h1};
`else
  assign IDREG = {8'hE6, 4'h0, VID, 4'h0, 4'h4, 8'h1};
`endif
  wire [63:0] UIDREG;
  assign UIDREG = 64'h0123456789abcdef;
  reg [7:0] ASP;
  localparam ASP_DEFAULT = 8'h00;
  reg [63:0] SN;
  localparam SN_DEFAULT = 64'h0;

  wire wren;
  wire [2:0] bpsel;
  wire tbsel;
  wire snpen;
  wire wpnen;
  assign wren = SR[1];
  assign bpsel = SR[4:2];
  assign tbsel = SR[5];
  assign snpen = SR[6];
  assign wpnen = SR[7];
  wire maplk, asplk;
  assign maplk = CR1[2];
  assign asplk = CR1[0];
  wire [3:0] mlats;
  assign mlats = CR2[3:0];
  wire [2:0] odsel;
  wire wraps;
  wire [2:0] wrpls;
  assign odsel = CR3[7:5];
  assign wraps = CR3[4];
  assign wrpls = CR3[2:0];
  wire [1:0] wrens;
  assign wrens = CR4[1:0];

  wire wr_allow; // WREN + WRENS logic
  assign wr_allow = (((wrens == 2'b00) && wren) ||
                     (wrens == 2'b01) ||
                     ((wrens == 2'b10) && wren)
                    )? 1'b1 : 0;

  reg swrst_en;

  /* Memory Array */
`ifdef DENSITY_1M // 1Mb
  localparam ARRAY_BYTES = 24'h20000;
  localparam UPPER_1_64 = 24'h1F800;
  localparam UPPER_1_32 = 24'h1F000;
  localparam UPPER_1_16 = 24'h1E000;
  localparam UPPER_1_8  = 24'h1C000;
  localparam UPPER_1_4  = 24'h18000;
  localparam UPPER_1_2  = 24'h10000;
  localparam LOWER_1_64 = 24'h007FF;
  localparam LOWER_1_32 = 24'h00FFF;
  localparam LOWER_1_16 = 24'h01FFF;
  localparam LOWER_1_8  = 24'h03FFF;
  localparam LOWER_1_4  = 24'h07FFF;
  localparam LOWER_1_2  = 24'h0FFFF;
`elsif DENSITY_4M // 4Mb
  localparam ARRAY_BYTES = 24'h80000;
  localparam UPPER_1_64 = 24'h7E000;
  localparam UPPER_1_32 = 24'h7C000;
  localparam UPPER_1_16 = 24'h78000;
  localparam UPPER_1_8  = 24'h70000;
  localparam UPPER_1_4  = 24'h60000;
  localparam UPPER_1_2  = 24'h40000;
  localparam LOWER_1_64 = 24'h01FFF;
  localparam LOWER_1_32 = 24'h03FFF;
  localparam LOWER_1_16 = 24'h07FFF;
  localparam LOWER_1_8  = 24'h0FFFF;
  localparam LOWER_1_4  = 24'h1FFFF;
  localparam LOWER_1_2  = 24'h3FFFF;
`elsif DENSITY_8M // 8Mb
  localparam ARRAY_BYTES = 24'h100000;
  localparam UPPER_1_64 = 24'hFC000;
  localparam UPPER_1_32 = 24'hF8000;
  localparam UPPER_1_16 = 24'hF0000;
  localparam UPPER_1_8  = 24'hE0000;
  localparam UPPER_1_4  = 24'hC0000;
  localparam UPPER_1_2  = 24'h80000;
  localparam LOWER_1_64 = 24'h03FFF;
  localparam LOWER_1_32 = 24'h07FFF;
  localparam LOWER_1_16 = 24'h0FFFF;
  localparam LOWER_1_8  = 24'h1FFFF;
  localparam LOWER_1_4  = 24'h3FFFF;
  localparam LOWER_1_2  = 24'h7FFFF;
`elsif DENSITY_16M // 16Mb
  localparam ARRAY_BYTES = 24'h200000;
  localparam UPPER_1_64 = 24'h1F8000;
  localparam UPPER_1_32 = 24'h1F0000;
  localparam UPPER_1_16 = 24'h1E0000;
  localparam UPPER_1_8  = 24'h1C0000;
  localparam UPPER_1_4  = 24'h180000;
  localparam UPPER_1_2  = 24'h100000;
  localparam LOWER_1_64 = 24'h007FFF;
  localparam LOWER_1_32 = 24'h00FFFF;
  localparam LOWER_1_16 = 24'h01FFFF;
  localparam LOWER_1_8  = 24'h03FFFF;
  localparam LOWER_1_4  = 24'h07FFFF;
  localparam LOWER_1_2  = 24'h0FFFFF;
`else
  localparam ARRAY_BYTES = 24'h200000;
  localparam UPPER_1_64 = 24'h1F8000;
  localparam UPPER_1_32 = 24'h1F0000;
  localparam UPPER_1_16 = 24'h1E0000;
  localparam UPPER_1_8  = 24'h1C0000;
  localparam UPPER_1_4  = 24'h180000;
  localparam UPPER_1_2  = 24'h100000;
  localparam LOWER_1_64 = 24'h007FFF;
  localparam LOWER_1_32 = 24'h00FFFF;
  localparam LOWER_1_16 = 24'h01FFFF;
  localparam LOWER_1_8  = 24'h03FFFF;
  localparam LOWER_1_4  = 24'h07FFFF;
  localparam LOWER_1_2  = 24'h0FFFFF;
`endif

/*SCK period*/
  real t0, t1;
  real SCK_period;

  always@(negedge CSn) begin
    @(posedge CLK) t0 = $realtime;
    @(posedge CLK) t1 = $realtime;
    SCK_period = t1 - t0;
  end

  reg [7:0] mem_array [0:ARRAY_BYTES-1];
  reg [23:0] wr_addr, rd_addr;

  /* Augmented Storage Array */
  localparam AUGMENTED_BYTES = 24'h100;
  reg [7:0] aug_array [0:AUGMENTED_BYTES-1];

  /* Address protect */
  reg addr_prot;
  always@(*)
    if (wr_addr >= ARRAY_BYTES) begin // Above capacity
      addr_prot <= 1'b1;
    end
    else begin
      if (tbsel == 0) // Top Block Protection
        case(bpsel)
          3'b001: addr_prot <= (wr_addr >= UPPER_1_64)? 1'b1 : 0;
          3'b010: addr_prot <= (wr_addr >= UPPER_1_32)? 1'b1 : 0;
          3'b011: addr_prot <= (wr_addr >= UPPER_1_16)? 1'b1 : 0;
          3'b100: addr_prot <= (wr_addr >= UPPER_1_8)? 1'b1 : 0;
          3'b101: addr_prot <= (wr_addr >= UPPER_1_4)? 1'b1 : 0;
          3'b110: addr_prot <= (wr_addr >= UPPER_1_2)? 1'b1 : 0;
          3'b111: addr_prot <= 1'b1; 
          default: addr_prot <= 0;
        endcase
      else            // Bottom Block Protection
        case(bpsel)
          3'b001: addr_prot <= (wr_addr <= LOWER_1_64)? 1'b1 : 0;
          3'b010: addr_prot <= (wr_addr <= LOWER_1_32)? 1'b1 : 0;
          3'b011: addr_prot <= (wr_addr <= LOWER_1_16)? 1'b1 : 0;
          3'b100: addr_prot <= (wr_addr <= LOWER_1_8)? 1'b1 : 0;
          3'b101: addr_prot <= (wr_addr <= LOWER_1_4)? 1'b1 : 0;
          3'b110: addr_prot <= (wr_addr <= LOWER_1_2)? 1'b1 : 0;
          3'b111: addr_prot <= 1'b1; 
          default: addr_prot <= 0;
        endcase
    end

  /* Augmented Array Protection */
  reg aug_prot;
  always@(*)
    if (wr_addr[23:8] != 16'd0) // Above capacity
      aug_prot <= 1'b1;
    else
      case(wr_addr[7:5])
        3'd0: aug_prot <= (ASP[0])? 1'b1 : 0;
        3'd1: aug_prot <= (ASP[1])? 1'b1 : 0;
        3'd2: aug_prot <= (ASP[2])? 1'b1 : 0;
        3'd3: aug_prot <= (ASP[3])? 1'b1 : 0;
        3'd4: aug_prot <= (ASP[4])? 1'b1 : 0;
        3'd5: aug_prot <= (ASP[5])? 1'b1 : 0;
        3'd6: aug_prot <= (ASP[6])? 1'b1 : 0;
        3'd7: aug_prot <= (ASP[7])? 1'b1 : 0;
      endcase
  
  /* Read address range check */
  reg addr_oor;
  always@(*)
    if (rd_addr[23:21] != 3'b000) // Above capacity (16Mb)
      addr_oor <= 1'b1;
    else
      addr_oor <= 0;

  /* Augmented Read address range check */
  reg aug_oor;
  always@(*)
    if (rd_addr[23:8] != 16'd0)
      aug_oor <= 1'b1;
    else
      aug_oor <= 0;

  /* SPI Pin write protect */
  reg hw_wp;
  always@(*)
    if (wpnen && (mode == SPIEN))
      if (!WPn_IO2)
        hw_wp <= 1'b1;
      else
        hw_wp <= 0;
    else
      hw_wp <= 0;

  /* Latches */
  reg [7:0] cmd_latch;
  reg [7:0] mode_latch; // mode byte

  wire [7:0] datain_reg_ser;
  wire [7:0] datain_reg_dual;
  wire [7:0] datain_reg_quad;
  reg  [7:0] datain_reg;
  srdin_ser DATAIN_LATCH_SERIAL(
    .sclk (CLK),
    .cs_n (CSn),
    .din  (SI_IO0),
    .dout (datain_reg_ser)
    );

  srdin_dual DATAIN_LATCH_DUAL(
    .sclk (CLK),
    .cs_n (CSn),
    .din  ({SO_IO1, SI_IO0}),
    .dout (datain_reg_dual)
    );

  srdin_quad DATAIN_LATCH_QUAD(
    .sclk (CLK),
    .cs_n (CSn),
    .din  ({IO3, WPn_IO2, SO_IO1, SI_IO0}),
    .dout (datain_reg_quad)
    );

  wire [7:0] datain_reg_ddr_ser;
  wire [7:0] datain_reg_ddr_dual;
  wire [7:0] datain_reg_ddr_quad;
  //reg  [7:0] datain_reg;
  srdin_ser_ddr DATAIN_LATCH_SERIAL_DDR(
    .sclk (CLK),
    .cs_n (CSn),
    .din  (SI_IO0),
    .dout (datain_reg_ddr_ser)
    );

  srdin_dual_ddr DATAIN_LATCH_DUAL_DDR(
    .sclk (CLK),
    .cs_n (CSn),
    .din  ({SO_IO1, SI_IO0}),
    .dout (datain_reg_ddr_dual)
    );

  srdin_quad_ddr DATAIN_LATCH_QUAD_DDR(
    .sclk (CLK),
    .cs_n (CSn),
    .din  ({IO3, WPn_IO2, SO_IO1, SI_IO0}),
    .dout (datain_reg_ddr_quad)
    );

  always@(*)
    case(state)
      SD_ADDR, SD_DATAIN:
        case(mode)
          SPIEN: begin
            case(cmd_latch)
              DWQI, DWQO:
                datain_reg <= datain_reg_ddr_quad;
              default:
                datain_reg <= datain_reg_ddr_ser;
            endcase
          end
          DPIEN:   datain_reg <= datain_reg_ddr_dual;
          QPIEN:   datain_reg <= datain_reg_ddr_quad;
          default: datain_reg <= 0;
        endcase
        
      S_MODE: // XIP mode byte input width
        case(mode)
          SPIEN: begin
            case(cmd_latch)
              RDDI, WDIO:
                datain_reg <= datain_reg_dual;
              RDQI, WQIO:
                datain_reg <= datain_reg_quad;
              default:
                datain_reg <= datain_reg_ser;
            endcase
          end
          DPIEN:   datain_reg <= datain_reg_dual;
          QPIEN:   datain_reg <= datain_reg_quad;
          default: datain_reg <= 0;
        endcase

      SD_MODE: // XIP mode byte input width
        case(mode)
          SPIEN: begin
            case(cmd_latch)
              DRDI:
                datain_reg <= datain_reg_ddr_dual;
              DRQI, DWQO:
                datain_reg <= datain_reg_ddr_quad;
              default:
                datain_reg <= datain_reg_ddr_ser;
            endcase
          end
          DPIEN:   datain_reg <= datain_reg_ddr_dual;
          QPIEN:   datain_reg <= datain_reg_ddr_quad;
          default: datain_reg <= 0;
        endcase

      S_DATAIN:
        case(mode)
          SPIEN: begin
            case(cmd_latch)
              RDDI, WDUI, WDIO:
                datain_reg <= datain_reg_dual;
              RDQI, WQDI, WQIO:
                datain_reg <= datain_reg_quad;
              default:
                datain_reg <= datain_reg_ser;
            endcase
          end
          DPIEN:   datain_reg <= datain_reg_dual;
          QPIEN:   datain_reg <= datain_reg_quad;
          default: datain_reg <= 0;
        endcase

      default:
        case(mode)
          SPIEN:   datain_reg <= datain_reg_ser;
          DPIEN:   datain_reg <= datain_reg_dual;
          QPIEN:   datain_reg <= datain_reg_quad;
          default: datain_reg <= 0;
        endcase
    endcase

  reg [2:0]  addr_mode_sel;
  addr_reg ADDR_LATCH(
    .sclk     (CLK),
    .cs_n     (CSn),
    .mode_sel (addr_mode_sel),
    .din      ({IO3, WPn_IO2, SO_IO1, SI_IO0}),
    .addr     (sr_addr)
    );

  addr_reg_ddr ADDR_LATCH_DDR(
    .sclk     (CLK),
    .cs_n     (CSn),
    .mode_sel (addr_mode_sel),
    .din      ({IO3, WPn_IO2, SO_IO1, SI_IO0}),
    .addr     (sr_addr_ddr)
    );


  always@(posedge CLK)
    if (s_addr_done) begin
      if (state == SD_ADDR)
        addr <= sr_addr_ddr;
      else
        addr <= sr_addr;
    end

  always@(*)
    case(mode)
      SPIEN: begin
        case(cmd_latch)
          RDDI, DRDI, WDIO:
            addr_mode_sel <= 3'b010;
          RDQI, DRQI, WQIO, DWQO:
            addr_mode_sel <= 3'b100;
          default:
            addr_mode_sel <= 3'b001;
        endcase
      end
      DPIEN: addr_mode_sel <= 3'b010;
      QPIEN: addr_mode_sel <= 3'b100;
      default: addr_mode_sel <= 3'b000;
    endcase

  /* XIP mode enable */
  reg xip_en;

  /* SCLK counter */
  reg [5:0] sclk_re_ctr;

  always@(posedge CLK or negedge PORn or posedge CSn)
    if (!PORn || CSn)
      sclk_re_ctr <= 0;
    else
      case(mode)
        SPIEN:
          case(state)
            S_CMD, S_DPCMD:
              //if (sclk_re_ctr == 6'd7)
              if (s_cmd_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1;

            S_ADDR, SD_ADDR:
              if (s_addr_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1; 

            S_MODE, SD_MODE:
              if (s_mode_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1; 

            S_DATAIN, S_DATAOUT, SD_DATAIN, SD_DATAOUT: 
              case(cmd_latch)
                WRCX, RDCX:
                  if (sclk_re_ctr == 6'd31)
                    sclk_re_ctr <= 0;
                  else
                    sclk_re_ctr <= sclk_re_ctr + 1;

                WRSN, RDSN, RDDO, RDQO, RDDI, RDQI, WDUI, WQDI, WDIO, WQIO, 
                DRFR, DRDI, DRQI, DWQI, DWQO:
                  if (s_datain_done || s_dataout_done)
                    sclk_re_ctr <= 0;
                  else
                    sclk_re_ctr <= sclk_re_ctr + 1;

                default:
                  if (sclk_re_ctr == 6'd7)
                    sclk_re_ctr <= 0;
                  else
                    sclk_re_ctr <= sclk_re_ctr + 1;
              endcase

            S_LATENCY:
              if (s_latency_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1;

            default:
              sclk_re_ctr <= 0;
          endcase

        DPIEN:
          case(state)
            S_CMD, S_DPCMD:
              //if (sclk_re_ctr == 6'd3)
              if (s_cmd_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1;

            S_ADDR, SD_ADDR:
              if (s_addr_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1; 

            S_MODE, SD_MODE:
              if (s_mode_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1; 

            S_DATAIN, S_DATAOUT, SD_DATAIN, SD_DATAOUT:
              case(cmd_latch)
                WRCX, RDCX:
                  if (sclk_re_ctr == 6'd15)
                    sclk_re_ctr <= 0;
                  else
                    sclk_re_ctr <= sclk_re_ctr + 1;

                WRSN, RDSN:
                  if (s_datain_done || s_dataout_done)
                    sclk_re_ctr <= 0;
                  else
                    sclk_re_ctr <= sclk_re_ctr + 1;

                default:
                  if (sclk_re_ctr == 6'd3)
                    sclk_re_ctr <= 0;
                  else
                    sclk_re_ctr <= sclk_re_ctr + 1;
                 
              endcase

            S_LATENCY:
              if (s_latency_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1;

            default:
              sclk_re_ctr <= 0;
          endcase

        QPIEN:
          case(state)
            S_CMD, S_DPCMD:
              if (s_cmd_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1;

            S_ADDR, SD_ADDR:
              if (s_addr_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1; 

            S_MODE, SD_MODE:
              if (s_mode_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1; 

            S_DATAIN, S_DATAOUT, SD_DATAIN, SD_DATAOUT:
              case(cmd_latch)
                WRCX, RDCX:
                  if (sclk_re_ctr == 6'd7)
                    sclk_re_ctr <= 0;
                  else
                    sclk_re_ctr <= sclk_re_ctr + 1;

                WRSN, RDSN:
                  if (s_datain_done || s_dataout_done)
                    sclk_re_ctr <= 0;
                  else
                    sclk_re_ctr <= sclk_re_ctr + 1;

                default:
                  if (sclk_re_ctr == 6'd1)
                    sclk_re_ctr <= 0;
                  else
                    sclk_re_ctr <= sclk_re_ctr + 1;
              endcase

            S_LATENCY:
              if (s_latency_done)
                sclk_re_ctr <= 0;
              else
                sclk_re_ctr <= sclk_re_ctr + 1;

            default:
              sclk_re_ctr <= 0;
          endcase

      endcase

  /*
   * Byte strobe
   */
  reg byte_strobe;
  reg [7:0] byte_strobe_ctr;
  always@(*)
    case(mode)
      SPIEN:
        case(sclk_re_ctr)
          6'd7, 6'd15, 6'd23, 6'd31, 6'd39, 6'd47, 6'd55, 6'd63:
            byte_strobe = 1'b1;
          6'd3, 6'd11, 6'd19, 6'd27, 6'd35, 6'd43, 6'd51, 6'd59:
            case(cmd_latch)
              DRFR, DRFW, RDDO, RDDI, WDUI, WDIO:
                byte_strobe = 1'b1;
              default:
                byte_strobe = 0;
            endcase
          6'd1:
            case(cmd_latch)
              RDQO, RDQI, WQDI, WQIO, DRDI:
                byte_strobe = 1'b1;
              default:
                byte_strobe = 0;
            endcase
          6'd0:
            case(cmd_latch)
              DRQI, DWQI, DWQO:
                byte_strobe = 1'b1;
              default:
                byte_strobe = 0;
            endcase
          default:
            byte_strobe = 0;
        endcase

      DPIEN:
        case(sclk_re_ctr)
          6'd3, 6'd7, 6'd11, 6'd15, 6'd19, 6'd23, 6'd27, 6'd31:
            byte_strobe = 1'b1;
          6'd1, 6'd5, 6'd9, 6'd13, 6'd17, 6'd21, 6'd25, 6'd29:
            case(cmd_latch)
              DRFR, DRFW:
                byte_strobe = 1'b1;
              default:
                byte_strobe = 0;
            endcase
          default:
            byte_strobe = 0;
        endcase

      QPIEN:
        case(sclk_re_ctr)
          6'd1, 6'd3, 6'd5, 6'd7, 6'd9, 6'd11, 6'd13, 6'd15:
            byte_strobe = 1'b1;
          6'd0, 6'd2, 6'd4, 6'd6, 6'd8, 6'd10, 6'd12, 6'd14:
            case(cmd_latch)
              DRFR, DRFW:
                byte_strobe = 1'b1;
              default:
                byte_strobe = 0;
            endcase
          default:
            byte_strobe = 0;
        endcase

      default:
        byte_strobe = 0;
    endcase

  always@(posedge CLK)
    case(state)
      S_DATAIN, S_DATAOUT, SD_DATAIN, SD_DATAOUT:
        if (byte_strobe)
          byte_strobe_ctr = byte_strobe_ctr + 1;
      default:
        byte_strobe_ctr = 0;
    endcase

  // Power on reset
  integer i;
  initial
  begin
    /* Initialize Array */
    for (i=0; i<ARRAY_BYTES; i=i+1) begin
      mem_array[i] = 0;
    end
    /* Initialize Augmented Storage */
    for (i=0; i<AUGMENTED_BYTES; i=i+1) begin
      aug_array[i] = 0;
    end
    PORn = 1'b0;
    #tPOR_time PORn = 1'b1;
    mode = SPIEN;
  end

  always@(posedge CSn) begin
    if ((state == S_CMD) && (cmd_latch == SRST) && swrst_en) begin
      $display("[MODEL] Software Reset Initiated");
      PORn = 1'b0;
      #tSRST PORn= 1'b1;
      mode = SPIEN;
    end
  end
  
  // Initial Reset
  always @(posedge PORn)
  begin
    hold_dq3_reg = 0; 
    si_dq0_reg = 0; 
    so_dq1_reg = 0;
    wp_dq2_reg = 0;
  end

  /* Output pin assignment */
  assign #(tCO) SI_IO0  = (dqout_en[0]) ? si_dq0_reg   : 1'bz;
  assign #(tCO) SO_IO1  = (dqout_en[1]) ? so_dq1_reg   : 1'bz;
  assign #(tCO) WPn_IO2 = (dqout_en[2]) ? wp_dq2_reg   : 1'bz;
  assign #(tCO) IO3     = (dqout_en[3]) ? hold_dq3_reg : 1'bz;

  reg [7:0] dataout_reg;
  reg dataout_reg_shift;
  reg dataout_reg_latch;
  
  //wire dataout_reg_ser_latch;
  //wire dataout_reg_ser_shift;
  wire dataout_reg_ser_dout;
  wire dataout_reg_ser_ddr;
  //wire dataout_reg_dual_latch;
  //wire dataout_reg_dual_shift;
  //wire dataout_reg_quad_latch;
  //wire dataout_reg_quad_shift;
  wire [1:0] dataout_reg_dual_dout;
  wire [1:0] dataout_reg_dual_ddr;
  wire [3:0] dataout_reg_quad_dout;
  wire [3:0] dataout_reg_quad_ddr;
  //assign dataout_reg_ser_latch = ((mode == SPIEN) && dataout_reg_latch)? 1'b1 : 0;
  //assign dataout_reg_ser_shift = ((mode == SPIEN) && dataout_reg_shift)? 1'b1 : 0;
  //assign dataout_reg_dual_latch = ((mode == DPIEN) && dataout_reg_latch)? 1'b1 : 0;
  //assign dataout_reg_dual_shift = ((mode == DPIEN) && dataout_reg_shift)? 1'b1 : 0;
  //assign dataout_reg_quad_latch = ((mode == QPIEN) && dataout_reg_latch)? 1'b1 : 0;
  //assign dataout_reg_quad_shift = ((mode == QPIEN) && dataout_reg_shift)? 1'b1 : 0;

  srdout_ser DATAOUT_LATCH_SERIAL(
    .sclk  (CLK),
    .cs_n  (CSn),
    //.latch (dataout_reg_ser_latch),
    .latch (dataout_reg_latch),
    //.shift (dataout_reg_ser_shift),
    .shift (dataout_reg_shift),
    .din   (dataout_reg),
    .dout  (dataout_reg_ser_dout),
    .ddr   (dataout_reg_ser_ddr)
    );

  srdout_dual DATAOUT_LATCH_DUAL(
    .sclk  (CLK),
    .cs_n  (CSn),
    //.latch (dataout_reg_dual_latch),
    .latch (dataout_reg_latch),
    //.shift (dataout_reg_dual_shift),
    .shift (dataout_reg_shift),
    .din   (dataout_reg),
    .dout  (dataout_reg_dual_dout),
    .ddr   (dataout_reg_dual_ddr)
    );

  srdout_quad DATAOUT_LATCH_QUAD(
    .sclk  (CLK),
    .cs_n  (CSn),
    //.latch (dataout_reg_quad_latch),
    .latch (dataout_reg_latch),
    //.shift (dataout_reg_quad_shift),
    .shift (dataout_reg_shift),
    .din   (dataout_reg),
    .dout  (dataout_reg_quad_dout),
    .ddr   (dataout_reg_quad_ddr)
    );

  /* Data out register to pin output */
  always@(*)
    case(mode)
      SPIEN: begin
        case(cmd_latch)
          DRFR:
            so_dq1_reg = dataout_reg_ser_ddr;
          DRDI:
          begin
            si_dq0_reg = dataout_reg_dual_ddr[0];
            so_dq1_reg = dataout_reg_dual_ddr[1];
          end
          DRQI:
          begin
            si_dq0_reg   = dataout_reg_quad_ddr[0];
            so_dq1_reg   = dataout_reg_quad_ddr[1];
            wp_dq2_reg   = dataout_reg_quad_ddr[2];
            hold_dq3_reg = dataout_reg_quad_ddr[3];
          end
          RDDO, RDDI:
          begin
            si_dq0_reg = dataout_reg_dual_dout[0];
            so_dq1_reg = dataout_reg_dual_dout[1];
          end
          RDQO, RDQI:
          begin
            si_dq0_reg   = dataout_reg_quad_dout[0];
            so_dq1_reg   = dataout_reg_quad_dout[1];
            wp_dq2_reg   = dataout_reg_quad_dout[2];
            hold_dq3_reg = dataout_reg_quad_dout[3];
          end
          default:
            so_dq1_reg = dataout_reg_ser_dout;
        endcase
      end

      DPIEN: begin
        case(cmd_latch)
          DRFR: begin
            si_dq0_reg = dataout_reg_dual_ddr[0];
            so_dq1_reg = dataout_reg_dual_ddr[1];
          end
          default: begin
            si_dq0_reg = dataout_reg_dual_dout[0];
            so_dq1_reg = dataout_reg_dual_dout[1];
          end
        endcase
      end

      QPIEN: begin
        case(cmd_latch)
          DRFR: begin
            si_dq0_reg   = dataout_reg_quad_ddr[0];
            so_dq1_reg   = dataout_reg_quad_ddr[1];
            wp_dq2_reg   = dataout_reg_quad_ddr[2];
            hold_dq3_reg = dataout_reg_quad_ddr[3];
          end
          default: begin
            si_dq0_reg   = dataout_reg_quad_dout[0];
            so_dq1_reg   = dataout_reg_quad_dout[1];
            wp_dq2_reg   = dataout_reg_quad_dout[2];
            hold_dq3_reg = dataout_reg_quad_dout[3];
          end
        endcase
      end
    endcase

  always@(negedge CLK or posedge CSn)
    if (CSn)
      dqout_en = 0;
    else begin
      dqout_en = 0;
      case(state)
        S_MODE, SD_MODE: // Special case for MLATS = 0
          //if (((cmd_latch == RDFT) || (cmd_latch == DRFR))
          //    && (mlats == 0) && s_mode_done)
          if ((mlats == 0) && s_mode_done) begin
            case(cmd_latch)
              RDFT, DRFR:
                case(mode)
                  SPIEN: dqout_en = 4'b0010;
                  DPIEN: dqout_en = 4'b0011;
                  QPIEN: dqout_en = 4'b1111;
                endcase

              RDDO, RDDI, DRDI:
                dqout_en = 4'b0011;

              RDQO, RDQI, DRQI:
                dqout_en = 4'b1111;

              default:
                dqout_en = 0;
            endcase
          end
          else
            dqout_en = 0;

        S_ADDR:
          if (s_addr_done)
            case(cmd_latch)
              READ:
                case(mode)
                  SPIEN: dqout_en = 4'b0010;
                  DPIEN: dqout_en = 4'b0011;
                  QPIEN: dqout_en = 4'b1111;
                endcase
              default:
                dqout_en = 0;
            endcase
          
        S_DATAOUT, SD_DATAOUT:
          case(cmd_latch)
            READ, RDFT, RDAS, DRFR:
              case(mode)
                SPIEN: dqout_en = 4'b0010;
                DPIEN: dqout_en = 4'b0011;
                QPIEN: dqout_en = 4'b1111;
                default: dqout_en = 0;
              endcase

            RDDO, RDDI, DRDI:
              dqout_en = 4'b0011;

            RDQO, RDQI, DRQI:
              dqout_en = 4'b1111;

            default: 
              if (!s_dataout_done)
                case(mode)
                  SPIEN: dqout_en = 4'b0010;
                  DPIEN: dqout_en = 4'b0011;
                  QPIEN: dqout_en = 4'b1111;
                  default: dqout_en = 0;
                endcase
          endcase

        S_CMD:
          if (s_cmd_done)
            case(datain_reg)
              RDSR, RDCX, RDAP, RDSN, RDC1, RDC2, RDC3, RDC4, RDID, RUID:
                case(mode)
                  SPIEN: dqout_en = 4'b0010;
                  DPIEN: dqout_en = 4'b0011;
                  QPIEN: dqout_en = 4'b1111;
                  default: dqout_en = 0;
                endcase
            endcase
          else
            dqout_en = 0;
            
        S_LATENCY:
          if (s_latency_done)
            case(cmd_latch) 
              RDDO, RDDI, DRDI:
                dqout_en = 4'b0011;

              RDQO, RDQI, DRQI:
                dqout_en = 4'b1111;

              default:
                case(mode)
                  SPIEN: dqout_en = 4'b0010;
                  DPIEN: dqout_en = 4'b0011;
                  QPIEN: dqout_en = 4'b1111;
                  default: dqout_en = 0;
                endcase
            endcase

          else
            dqout_en = 0;

      endcase
    end

  /* Hibernate/Deep Sleep Triggers */
  reg hb_ent_done, hb_ext_done;
  reg dp_ent_done, dp_ext_done;
  reg dp_cswake;

  /* State machine */
  always@(posedge CLK or negedge PORn or posedge CSn or negedge CSn
          or posedge hb_ent_done or posedge hb_ext_done
          or posedge dp_ent_done or posedge dp_ext_done or posedge dp_cswake)
          
    if ((state == S_HBENT) && hb_ent_done) begin
      state <= S_HIBER;
    end

    else if ((state == S_HBEXT) && hb_ext_done) begin
      state <= S_IDLE;
    end

    else if ((state == S_DPENT) && dp_ent_done) begin
      state <= S_DEEP;
    end

    else if ((state == S_DPEXT) && dp_ext_done) begin
      state <= S_IDLE;
    end

    else if (!CSn && (state == S_HIBER)) begin
      state <= S_HBEXT;
      $display("[MODEL] Exiting Hibernate");
    end

    else if (CSn) begin
      if ((state == S_CMD) && (cmd_latch == HBNE)) begin
        state <= S_HBENT;
        $display("[MODEL] Entering Hibernate");
      end
      else if ((state == S_CMD) && (cmd_latch == DPDE))begin
        state <= S_DPENT;
        $display("[MODEL] Entering Deep Power Down");
      end
      else if (((state == S_DEEP) || (state == S_DPCMD)) && dp_cswake) begin
        state <= S_DPEXT;
        $display("[MODEL] Exiting Deep Power Down (CS trigger)");
      end
      else if (((state == S_DEEP) || (state == S_DPCMD)) && !dp_cswake && (cmd_latch != DPDX)) begin
        state <= S_DEEP;
      end
      else if ((state == S_DPCMD) && s_cmd_done) begin
        if (cmd_latch == DPDX) begin
          state <= S_DPEXT;
          $display("[MODEL] Exiting Deep Power Down (DPDX)");
        end
        else begin
          state <= S_DEEP;
        end
      end
      else begin
        state <= S_IDLE;
      end
    end

    else if (CLK) begin
      case (state)
        S_IDLE:
          if (xip_en) begin /* XIP transition direct to Addr states */
            case(cmd_latch)
              DRFR, DRFW, DRDI, DRQI, DWQI, DWQO:
                state <= SD_ADDR;
              default:
                state <= S_ADDR;
            endcase
          end
          else begin
            state <= S_CMD;
          end

        S_CMD:
          if (s_cmd_done)
            case(cmd_latch)
              RDSR, RDCX, RDAP, RDSN, RDC1, RDC2, RDC3, RDC4, RDID, RUID:
                state <= S_DATAOUT;

              WRSR, WRCX, WRAP, WRSN:
                state <= S_DATAIN;

              READ, RDAS:
                if (mode == SPIEN) begin
                  state <= S_ADDR;
                end
                else begin
                  state <= S_IGNORE;
                  $display("[MODEL] READ/RDAS ignored (restricted to SPI)");
                end

              RDAR, WRAR, RDFT:
                state <= S_ADDR;

              //RDDO, RDQO, RDDI, RDQI, WDUI, WQDI:
              RDDO, RDQO, RDDI, RDQI:
                if (mode == SPIEN)
                  state <= S_ADDR;
                else begin
                  state <= S_IGNORE;
                  $display("[MODEL] RDDO/RDQO/RDDI/RDQI ignored (restricted to SPI)");
                end

              WRAS:
                if (mode == SPIEN) begin
                  if (wren) begin
                    state <= S_ADDR;
                  end
                  else begin
                    state <= S_IGNORE; 
                    $display("[MODEL] WREN not set. WRAS ignored");
                  end
                end
                else begin
                  state <= S_IGNORE;
                  $display("[MODEL] WRAS ignored (restricted to SPI)");
                end
                
              WRTE:
                if (mode == SPIEN) begin
                  if (wr_allow) begin
                    state <= S_ADDR;
                  end
                  else begin
                    if ((wrens == 2'b00) || (wrens == 2'b10)) begin
                      $display("[MODEL] WREN not set. WRTE ignored");
                      state <= S_IGNORE;
                    end
                    else begin
                      $display("[MODEL] Invalid WRENS value 0x%X. WRTE ignored", wrens);
                      state <= S_IGNORE;
                    end
                  end
                end
                else begin
                  state <= S_IGNORE;
                  $display("[MODEL] WRTE ignored (restricted to SPI)");
                end

              WRFT, WDUI, WQDI, WDIO, WQIO:
                if (wr_allow) begin
                  state <= S_ADDR;
                end
                else begin
                  state <= S_IGNORE;
                  if ((wrens == 2'b00) || (wrens == 2'b10)) begin
                    $display("[MODEL] WREN not set. Write ignored");
                  end
                  else begin
                    $display("[MODEL] Invalid WRENS value 0x%X. Write ignored", wrens);
                  end
                end

              DRFR, DRDI, DRQI:
                state <= SD_ADDR;

              DRFW, DWQI, DWQO:
                if (wr_allow) begin
                  state <= SD_ADDR;
                end
                else begin
                  state <= S_IGNORE;
                  if ((wrens == 2'b00) || (wrens == 2'b10)) begin
                    $display("[MODEL] WREN not set. Write ignored");
                  end
                  else begin
                    $display("[MODEL] Invalid WRENS value 0x%X. DRFW ignored", wrens);
                  end
                end

              default:
                state <= S_IGNORE;
            endcase

        S_ADDR:
          if (s_addr_done)
            case(cmd_latch)
              RDAR, RDAS:
                state <= S_LATENCY;
              WRAR:  begin
                state <= S_DATAIN;
                case(sr_addr)
                  ADDR_SR, ADDR_CR1, ADDR_CR2, ADDR_CR3, ADDR_CR4:
                    $display("[MODEL] WRAR address valid");
                  ADDR_ID, ADDR_UID:
                    $display("[MODEL] WRAR address not written (Read-only register)");
                  default:
                    $display("[MODEL] WRAR address invalid: 0x%x", sr_addr);
                endcase
              end
              WRTE, WRAS:
                state <= S_DATAIN;
              READ:
                state <= S_DATAOUT;
              WRFT, RDFT, RDDO, RDQO, RDDI, RDQI, WDUI, WQDI, WDIO, WQIO:
                state <= S_MODE;
              default:
                state <= S_IGNORE;
            endcase

        SD_ADDR:
          if (s_addr_done)
            case(cmd_latch)
              DRFR, DRFW, DRDI, DRQI, DWQI, DWQO:
                state <= SD_MODE;
              default:
                state <= S_IGNORE;
            endcase

        S_MODE:
          if (s_mode_done)
            case(cmd_latch)
              WRFT, WDUI, WQDI, WDIO, WQIO:
                state <= S_DATAIN;
              RDFT, RDDO, RDQO, RDDI, RDQI:
                if (mlats != 0)
                  state <= S_LATENCY;
                else
                  state <= S_DATAOUT;
              default:
                state <= S_IGNORE;
            endcase

        SD_MODE:
          if (s_mode_done)
            case(cmd_latch)
              DRFR, DRDI, DRQI:
                if (mlats != 0)
                  state <= S_LATENCY;
                else
                  state <= SD_DATAOUT;
              DRFW, DWQI, DWQO:
                state <= SD_DATAIN;
              default:
                state <= S_IGNORE;
            endcase

        S_LATENCY:
          if (s_latency_done)
            case(cmd_latch)
              RDAR, RDFT, RDAS, RDDO, RDQO, RDDI, RDQI:
                state <= S_DATAOUT;
              DRFR, DRDI, DRQI:
                state <= SD_DATAOUT;
              default:
                state <= S_IGNORE;
            endcase

        S_DATAIN:
          if (s_datain_done)
            case(cmd_latch)
              WRSR, WRCX, WRAR, WRAP, WRSN:
                state <= S_IDLE;
              WRTE, WRFT, WRAS, WDUI, WQDI, WDIO, WQIO:
                state <= S_DATAIN; // continue until CSn deasserted
              default:
                state <= S_IGNORE;
            endcase

        SD_DATAIN:
          if (s_datain_done)
            case(cmd_latch)
              DRFW, DWQI, DWQO:
                state <= SD_DATAIN; // continue until CSn deasserted
              default:
                state <= S_IGNORE;
            endcase

        S_DATAOUT:
          if (s_dataout_done)
            case(cmd_latch)
              RDSR, RDCX, RDAR, RDAP, RDSN, RDC1, RDC2, RDC3, RDC4, RDID, RUID:
                state <= S_IDLE;
              READ, RDFT, RDAS, RDDO, RDQO, RDDI, RDQI:
                state <= S_DATAOUT; // continue until CSn deasserted
              default:
                state <= S_IGNORE;
            endcase

        SD_DATAOUT:
          case(cmd_latch)
            DRFR, DRDI, DRQI:
              state <= SD_DATAOUT; // continue until CSn deasserted
            default:
              state <= S_IGNORE;
          endcase

        S_HIBER:
          state <= S_HIBER;

        S_DEEP:
          state <= S_DPCMD;

        S_DPCMD:
          state <= S_DPCMD;

        S_IGNORE:
          state <= S_IGNORE;

        default:
          state <= S_IDLE;
      endcase
    end

  /* Command Notifier */
  always@(negedge CLK)
    if (s_cmd_done && (state == S_DPCMD)) begin
      if (datain_reg == DPDX)
        $display("[MODEL] Command received 0x%x (DPDX)", datain_reg);
      else
        $display("[MODEL] Unsupported command 0x%X. Currently in Deep Power Down", datain_reg);
    end
    else if (s_cmd_done)
      case(datain_reg)
        WREN: $display("[MODEL] Command received 0x%x (WREN)", datain_reg);
        WRDI: $display("[MODEL] Command received 0x%x (WRDI)", datain_reg);
        RDSR: $display("[MODEL] Command received 0x%x (RDSR)", datain_reg);
        WRSR: $display("[MODEL] Command received 0x%x (WRSR)", datain_reg);
        SPIE: 
          if (modesw_valid) 
            $display("[MODEL] Command received 0x%x (SPIE)", datain_reg);
          else
            $display("[MODEL] Command received 0x%x (SPIE), but already in SPI mode", datain_reg);
        DPIE: 
          if (modesw_valid)
            $display("[MODEL] Command received 0x%x (DPIE)", datain_reg);
          else
            $display("[MODEL] Command received 0x%x (DPIE), but already in DPI mode", datain_reg);
        QPIE: 
          if (modesw_valid)
            $display("[MODEL] Command received 0x%x (QPIE)", datain_reg);
          else
            $display("[MODEL] Command received 0x%x (QPIE), but already in QPI mode", datain_reg);

        WRCX: $display("[MODEL] Command received 0x%x (WRCX)", datain_reg);
        RDCX: $display("[MODEL] Command received 0x%x (RDCX)", datain_reg);
        RDAR: $display("[MODEL] Command received 0x%x (RDAR)", datain_reg);
        WRAR: $display("[MODEL] Command received 0x%x (WRAR)", datain_reg);
        WRTE: $display("[MODEL] Command received 0x%x (WRTE)", datain_reg);
        READ: $display("[MODEL] Command received 0x%x (READ)", datain_reg);
        WRFT: $display("[MODEL] Command received 0x%x (WRFT)", datain_reg);
        RDFT: $display("[MODEL] Command received 0x%x (RDFT)", datain_reg);
        WRAP: $display("[MODEL] Command received 0x%x (WRAP)", datain_reg);
        RDAP: $display("[MODEL] Command received 0x%x (RDAP)", datain_reg);
        WRAS: $display("[MODEL] Command received 0x%x (WRAS)", datain_reg);
        RDAS: $display("[MODEL] Command received 0x%x (RDAS)", datain_reg);
        RDSN: $display("[MODEL] Command received 0x%x (RDSN)", datain_reg);
        WRSN: $display("[MODEL] Command received 0x%x (WRSN)", datain_reg);
        HBNE: $display("[MODEL] Command received 0x%x (HBNE)", datain_reg);
        DPDE: $display("[MODEL] Command received 0x%x (DPDE)", datain_reg);
        DPDX: $display("[MODEL] Command received 0x%x (DPDX), but ignored since not in Deep Power Down", 
                datain_reg);
        SRTE: $display("[MODEL] Command received 0x%x (SRTE)", datain_reg);
        SRST: $display("[MODEL] Command received 0x%x (SRST)", datain_reg);
        RDC1: $display("[MODEL] Command received 0x%x (RDC1)", datain_reg);
        RDC2: $display("[MODEL] Command received 0x%x (RDC2)", datain_reg);
        RDC3: $display("[MODEL] Command received 0x%x (RDC3)", datain_reg);
        RDC4: $display("[MODEL] Command received 0x%x (RDC4)", datain_reg);
        RDID: $display("[MODEL] Command received 0x%x (RDID)", datain_reg);
        RUID: $display("[MODEL] Command received 0x%x (RUID)", datain_reg);
        DRFR: $display("[MODEL] Command received 0x%x (DRFR)", datain_reg);
        DRFW: $display("[MODEL] Command received 0x%x (DRFW)", datain_reg);
        RDDO: $display("[MODEL] Command received 0x%x (RDDO)", datain_reg);
        RDQO: $display("[MODEL] Command received 0x%x (RDQO)", datain_reg);
        RDDI: $display("[MODEL] Command received 0x%x (RDDI)", datain_reg);
        RDQI: $display("[MODEL] Command received 0x%x (RDQI)", datain_reg);
        WDUI: $display("[MODEL] Command received 0x%x (WDUI)", datain_reg);
        WQDI: $display("[MODEL] Command received 0x%x (WQDI)", datain_reg);
        WDIO: $display("[MODEL] Command received 0x%x (WDIO)", datain_reg);
        WQIO: $display("[MODEL] Command received 0x%x (WQIO)", datain_reg);
        DRDI: $display("[MODEL] Command received 0x%x (DRDI)", datain_reg);
        DRQI: $display("[MODEL] Command received 0x%x (DRQI)", datain_reg);
        DWQI: $display("[MODEL] Command received 0x%x (DWQI)", datain_reg);
        DWQO: $display("[MODEL] Command received 0x%x (DWQO)", datain_reg);
        default: $display("[MODEL] Unsupported command 0x%x", datain_reg);
      endcase

  /* Dataout cycles LUT */
  reg [7:0] s_dataout_cc;
  always@(*)
    case(cmd_latch)
      RDAR:
        case(addr)
          ADDR_SR, ADDR_CR1, ADDR_CR2, ADDR_CR3, ADDR_CR4:
            s_dataout_cc <= 8'd8; // 8-1
          ADDR_ID:
            s_dataout_cc <= 8'd31; // 32-1
          ADDR_UID:
            s_dataout_cc <= 8'd63; // 64-1
          default:
            s_dataout_cc <= 8'd8; // 8-1
        endcase

      default:
       s_dataout_cc <= 0;
    endcase

  /* State done signals */
  always@(*)
    if ((state == S_CMD) || (state == S_DPCMD))
      case(mode)
        SPIEN: s_cmd_done <= (sclk_re_ctr == 6'd7)? 1'b1 : 0;
        DPIEN: s_cmd_done <= (sclk_re_ctr == 6'd3)? 1'b1 : 0;
        QPIEN: s_cmd_done <= (sclk_re_ctr == 6'd1)? 1'b1 : 0;
        default: s_cmd_done <= 0;
      endcase
    else
      s_cmd_done <= 0;

  always@(*)
    if (state == S_ADDR)
      case(mode)
        SPIEN: begin
          case(cmd_latch)
            RDDI, WDIO:
              s_addr_done <= (sclk_re_ctr == 6'd11)? 1'b1 : 0;
            RDQI, WQIO:
              s_addr_done <= (sclk_re_ctr == 6'd5)? 1'b1 : 0;
            default:
              s_addr_done <= (sclk_re_ctr == 6'd23)? 1'b1 : 0;
          endcase
        end
        DPIEN: s_addr_done <= (sclk_re_ctr == 6'd11)? 1'b1 : 0;
        QPIEN: s_addr_done <= (sclk_re_ctr == 6'd5) ? 1'b1 : 0;
        default: s_addr_done <= 0;
      endcase
    else if (state == SD_ADDR)
      case(mode)
        SPIEN: begin
          case(cmd_latch)
            DRDI:
              s_addr_done <= (sclk_re_ctr == 6'd5)? 1'b1 : 0;
            DRQI, DWQO:
              s_addr_done <= (sclk_re_ctr == 6'd2)? 1'b1 : 0;
            default:
              s_addr_done <= (sclk_re_ctr == 6'd11)? 1'b1 : 0;
          endcase
        end
        DPIEN: s_addr_done <= (sclk_re_ctr == 6'd5) ? 1'b1 : 0;
        QPIEN: s_addr_done <= (sclk_re_ctr == 6'd2) ? 1'b1 : 0;
        default: s_addr_done <= 0;
      endcase
    else
      s_addr_done <= 0;

  always@(*)
    if (state == S_MODE)
      case(mode)
        SPIEN: begin 
          case(cmd_latch)
            RDDI, WDIO:
              s_mode_done <= (sclk_re_ctr == 6'd3)? 1'b1 : 0;
            RDQI, WQIO:
              s_mode_done <= (sclk_re_ctr == 6'd1)? 1'b1 : 0;
            default:
              s_mode_done <= (sclk_re_ctr == 6'd7)? 1'b1 : 0;
          endcase
        end
        DPIEN: s_mode_done <= (sclk_re_ctr == 6'd3)? 1'b1 : 0;
        QPIEN: s_mode_done <= (sclk_re_ctr == 6'd1)? 1'b1 : 0;
        default: s_mode_done <= 0;
      endcase
    else if (state == SD_MODE)
      case(mode)
        SPIEN: begin
          case(cmd_latch)
            DRDI:
              s_mode_done <= (sclk_re_ctr == 6'd1)? 1'b1 : 0;
            DRQI, DWQO:
              s_mode_done <= 1'b1; //1cc done
            default:
              s_mode_done <= (sclk_re_ctr == 6'd3)? 1'b1 : 0;
          endcase
        end
        DPIEN: s_mode_done <= (sclk_re_ctr == 6'd1)? 1'b1 : 0;
        QPIEN: s_mode_done <= (sclk_re_ctr == 6'd0)? 1'b1 : 0;
        default: s_mode_done <= 0;
      endcase
    else
      s_mode_done <= 0;

  always@(*)
    if ((state == S_DATAIN) || (state == SD_DATAIN))
      case(cmd_latch)
        WRSR, WRAR, WRTE, WRFT, WRAP, WRAS: 
          case(mode)
            SPIEN: s_datain_done <= (sclk_re_ctr == 6'd7)? 1'b1 : 0;
            DPIEN: s_datain_done <= (sclk_re_ctr == 6'd3)? 1'b1 : 0;
            QPIEN: s_datain_done <= (sclk_re_ctr == 6'd1)? 1'b1 : 0;
            default: s_datain_done <= 0;
          endcase

        WRCX:
          case(mode)
            SPIEN: s_datain_done <= (sclk_re_ctr == 6'd31)? 1'b1 : 0;
            DPIEN: s_datain_done <= (sclk_re_ctr == 6'd15)? 1'b1 : 0;
            QPIEN: s_datain_done <= (sclk_re_ctr == 6'd7)? 1'b1 : 0;
            default: s_datain_done <= 0;
          endcase

        WRSN:
          case(mode)
            SPIEN: s_datain_done <= (sclk_re_ctr == 6'd63)? 1'b1 : 0;
            DPIEN: s_datain_done <= (sclk_re_ctr == 6'd31)? 1'b1 : 0;
            QPIEN: s_datain_done <= (sclk_re_ctr == 6'd15)? 1'b1 : 0;
            default: s_datain_done <= 0;
          endcase

        DRFW, WDUI, WQDI, WDIO, WQIO, DWQI, DWQO:
          /*case(mode)
            SPIEN: s_datain_done <= (sclk_re_ctr == 6'd3)? 1'b1 : 0;
            DPIEN: s_datain_done <= (sclk_re_ctr == 6'd1)? 1'b1 : 0;
            QPIEN: s_datain_done <= 1'b1;
            default: s_datain_done <= 0;
          endcase*/
          s_datain_done <= byte_strobe;

        default:
          s_datain_done <= 0;
      endcase
    else
      s_datain_done <= 0;

  always@(*)
    //if (state == S_DATAOUT)
    case(state)
      S_DATAOUT:
        case(cmd_latch)
          RDSR, READ, RDFT, RDAP, RDAS, RDC1, RDC2, RDC3, RDC4:
            case(mode)
              SPIEN: s_dataout_done <= (sclk_re_ctr == 6'd7)? 1'b1 : 0;
              DPIEN: s_dataout_done <= (sclk_re_ctr == 6'd3)? 1'b1 : 0;
              QPIEN: s_dataout_done <= (sclk_re_ctr == 6'd1)? 1'b1 : 0;
              default: s_dataout_done <= 0;
            endcase

          RDDO, RDQO, RDDI, RDQI:
            s_dataout_done <= byte_strobe;

          RDCX, RDID:
            case(mode)
              SPIEN: s_dataout_done <= (sclk_re_ctr == 6'd31)? 1'b1 : 0;
              DPIEN: s_dataout_done <= (sclk_re_ctr == 6'd15)? 1'b1 : 0;
              QPIEN: s_dataout_done <= (sclk_re_ctr == 6'd7)? 1'b1 : 0;
              default: s_dataout_done <= 0;
            endcase

          RUID:
            case(mode)
              SPIEN: s_dataout_done <= (sclk_re_ctr == 6'd63)? 1'b1 : 0;
              DPIEN: s_dataout_done <= (sclk_re_ctr == 6'd31)? 1'b1 : 0;
              QPIEN: s_dataout_done <= (sclk_re_ctr == 6'd15)? 1'b1 : 0;
              default: s_dataout_done <= 0;
            endcase

          RDAR:
            case(mode)
              SPIEN: s_dataout_done <= (sclk_re_ctr == (s_dataout_cc - 8'd1))? 1'b1 : 0;
              DPIEN: s_dataout_done <= (sclk_re_ctr == ((s_dataout_cc/2) - 8'd1))? 1'b1 : 0;
              DPIEN: s_dataout_done <= (sclk_re_ctr == ((s_dataout_cc/4) - 8'd1))? 1'b1 : 0;
              default: s_dataout_done <= 0;
            endcase

          RDSN:
            case(mode)
              SPIEN: s_dataout_done <= (sclk_re_ctr == 6'd63)? 1'b1 : 0;
              DPIEN: s_dataout_done <= (sclk_re_ctr == 6'd31)? 1'b1 : 0;
              QPIEN: s_dataout_done <= (sclk_re_ctr == 6'd15)? 1'b1 : 0;
              default: s_dataout_done <= 0;
            endcase

          default:
            s_dataout_done <= 0;
        endcase

      SD_DATAOUT:
        case(cmd_latch)
          DRFR, DRDI, DRQI:
            s_dataout_done <= byte_strobe;

          default:
            s_dataout_done <= 0;
        endcase

    //else
      default:
        s_dataout_done <= 0;

    endcase

  always@(*)
    if (state == S_LATENCY)
      case(cmd_latch)
        RDAR:
          case(mode)
            SPIEN: s_latency_done <= (sclk_re_ctr == 6'd7)? 1'b1 : 0;
            DPIEN: s_latency_done <= (sclk_re_ctr == 6'd3)? 1'b1 : 0;
            QPIEN: s_latency_done <= (sclk_re_ctr == 6'd1)? 1'b1 : 0;
          endcase

        RDFT, RDAS, DRFR, RDDO, RDQO, RDDI, RDQI, DRDI, DRQI:
          if (sclk_re_ctr == ({1'd0, mlats} - 6'd1))
            s_latency_done <= 1'b1;
          else
            s_latency_done <= 1'b0;

        default: s_latency_done <= 0;
      endcase
    else
      s_latency_done <= 0;

  /* Command Latch */
  always@(negedge CLK or negedge CSn)
    if (!CSn && (state == S_HIBER)) begin
      cmd_latch <= NOOP; // prevents state machine from entering Hibernate on posedge CSn
    end
    else if (!CSn && (state == S_DEEP)) begin
      cmd_latch <= NOOP;
    end
    else if (s_cmd_done) begin
      cmd_latch <= datain_reg;
    end

  /* Mode */
  always@(negedge CLK)
    case(state)
      S_CMD:
        if (s_cmd_done)
          case(datain_reg)
            SPIE:
              if ((mode == DPIEN) || (mode == QPIEN)) begin
                mode = SPIEN;
                $display("[MODEL] Mode set to SPI");
              end

            DPIE:
              if ((mode == SPIEN) || (mode == QPIEN)) begin
                mode = DPIEN;
                $display("[MODEL] Mode set to DPI");
              end

            QPIE:
              if ((mode == SPIEN) || (mode == DPIEN)) begin
                mode = QPIEN;
                $display("[MODEL] Mode set to QPI");
              end

          endcase
    endcase

  /* Status Register
   * WREN bit may be cleared earlier than end of transaction since this is
   * just a model
   */
  always@(negedge CLK)
    if (!hw_wp) begin
      case(state)
        S_CMD:
          //if (sclk_re_ctr == 3'd7)
          if (s_cmd_done)
            case(datain_reg)
              WREN: begin
                SR <= SR | 8'h02;
                $display("[MODEL] WREN bit set. SR=0x%x", (SR | 8'h02));
              end
              WRDI: begin
                SR <= SR & 8'hFD;
                $display("[MODEL] WREN bit reset. SR=0x%x", (SR & 8'hFD));
              end
            endcase

        S_ADDR, SD_ADDR:
          if (s_addr_done)
            case(cmd_latch)
              /*WRTE, WRFT, DRFW, WDUI, WQDI, DWQI, WDIO, WQIO, DWQO:
              begin
                if (wrens == 2'b00)
                  SR <= SR & 8'hFD; // clear WREN
                // else, don't clear (for wrens == 2'b01 and 2'b10)
              end*/

              WRAS:
                SR <= SR & 8'hFD; // clear WREN
            endcase

        S_DATAIN, SD_DATAIN:
          if (s_datain_done)
            case(cmd_latch)
              WRTE, WRFT, DRFW, WDUI, WQDI, DWQI, WDIO, WQIO, DWQO:
              begin
                if ((wrens == 2'b00) && !xip_en)
                  SR <= SR & 8'hFD; // clear WREN
                // else, don't clear (for wrens == 2'b01 and 2'b10)
                // or Normal mode XIP enabled
              end

              WRSR: begin
                if (wren) begin
                  if (!maplk) begin
                    SR <= ((SR & ~SR_WRMASK) | (datain_reg & SR_WRMASK)) & 8'hFD; // clear WREN as well
                    $display("[MODEL] SR written. SR=0x%x", 
                      ((SR & ~SR_WRMASK) | (datain_reg & SR_WRMASK)) & 8'hFD);
                    end
                  else begin
                    SR <= ((SR & ~SR_MAPLK) | (datain_reg & SR_MAPLK)) & 8'hFD; // clear WREN as well
                    $display("[MODEL] SR written (MAPLK set). SR=0x%x", 
                      ((SR & ~SR_MAPLK) | (datain_reg & SR_MAPLK)) & 8'hFD);
                  end
                end
                else
                  $display("[MODEL] WREN bit not set. SR not written. SR=0x%x", SR);
              end

              WRCX, WRAP, WRSN:
                SR <= SR & 8'hFD; // clear WREN

              WRAR: begin
                if (addr == ADDR_SR) begin
                  if (wren) begin
                    if (!maplk) begin
                      SR <= ((SR & ~SR_WRMASK) | (datain_reg & SR_WRMASK)) & 8'hFD; // clear WREN as well
                      $display("[MODEL] SR written. SR=0x%x", 
                        ((SR & ~SR_WRMASK) | (datain_reg & SR_WRMASK)) & 8'hFD);
                      end
                    else begin
                      SR <= ((SR & ~SR_MAPLK) | (datain_reg & SR_MAPLK)) & 8'hFD; // clear WREN as well
                      $display("[MODEL] SR written (MAPLK set). SR=0x%x", 
                        ((SR & ~SR_MAPLK) | (datain_reg & SR_MAPLK)) & 8'hFD);
                    end
                  end
                  else begin
                    $display("[MODEL] WREN bit not set. SR not written. SR=0x%x", SR);
                  end
                end
                else begin
                  SR <= SR & 8'hFD; // clear WREN
                end
              end

            endcase
        endcase
    end

  /* Mode switch valid */
  //reg modesw_valid;
  always@(*)
    case(mode)
      SPIE:
        case(datain_reg)
          DPIEN, QPIEN: modesw_valid = 1'b1;
          default:      modesw_valid = 0;
        endcase
      DPIE:
        case(datain_reg)
          SPIEN, QPIEN: modesw_valid = 1'b1;
          default:      modesw_valid = 0;
        endcase
      QPIE:
        case(datain_reg)
          DPIEN, SPIEN: modesw_valid = 1'b1;
          default:      modesw_valid = 0;
        endcase
    endcase

  /* Config Registers */
  always@(negedge CLK)
    if (!hw_wp) begin
      case(state)
        S_CMD:
          if (s_cmd_done && modesw_valid)
            case(datain_reg)
              SPIE: begin
                //if ((mode == DPIEN) || (mode == QPIEN)) begin //QPISL,DPISL update
                  {CR2[6], CR2[4]} <= 2'b00;
                  $display("[MODEL] CR2 updated. CR2=0x%x", {CR2[7],1'b0,CR2[5],1'b0,CR2[3:0]});
                end
              DPIE: begin
                //if ((mode == SPIEN) || (mode == QPIEN)) begin //QPISL,DPISL update
                  {CR2[6], CR2[4]} <= 2'b01;
                  $display("[MODEL] CR2 updated. CR2=0x%x", {CR2[7],1'b0,CR2[5],1'b1,CR2[3:0]});
                end
              QPIE: begin
                //if ((mode == SPIEN) || (mode == DPIEN)) begin //QPISL,DPISL update
                  {CR2[6], CR2[4]} <= 2'b10;
                  $display("[MODEL] CR2 updated. CR2=0x%x", {CR2[7],1'b1,CR2[5],1'b0,CR2[3:0]});
                end

            endcase

        S_DATAIN:
          case(cmd_latch)
            WRAR: begin
              if (wren) begin
                if (s_datain_done) begin
                  case(addr)
                    ADDR_CR1: begin
                      CR1 <= (CR1 & ~CR1_WRMASK) | (datain_reg & CR1_WRMASK);
                      $display("[MODEL] CR1 written. CR1=0x%x",
                        (CR1 & ~CR1_WRMASK) | (datain_reg & CR1_WRMASK));
                    end
                    ADDR_CR2: begin
                      CR2 <= (CR2 & ~CR2_WRMASK) | (datain_reg & CR2_WRMASK);
                      $display("[MODEL] CR2 written. CR2=0x%x",
                        (CR2 & ~CR2_WRMASK) | (datain_reg & CR2_WRMASK));
                    end
                    ADDR_CR3: begin
                      CR3 <= (CR3 & ~CR3_WRMASK) | (datain_reg & CR3_WRMASK);
                      $display("[MODEL] CR3 written. CR3=0x%x",
                        (CR3 & ~CR3_WRMASK) | (datain_reg & CR3_WRMASK));
                    end
                    ADDR_CR4: begin
                      CR4 <= (CR4 & ~CR4_WRMASK) | (datain_reg & CR4_WRMASK);
                      $display("[MODEL] CR4 written. CR4=0x%x",
                        (CR4 & ~CR4_WRMASK) | (datain_reg & CR4_WRMASK));
                    end
                  endcase
                end
              end
              else begin
                if (s_datain_done)
                  case(addr)
                    ADDR_CR1, ADDR_CR2, ADDR_CR3, ADDR_CR4:
                      $display("[MODEL] WREN bit not set CRx not written");
                  endcase
              end

            end

            WRCX: begin
              if (wren) begin
                if (byte_strobe) begin
                  case(byte_strobe_ctr)
                    8'd0: begin
                      CR1 <= (CR1 & ~CR1_WRMASK) | (datain_reg & CR1_WRMASK);
                      $display("[MODEL] CR1 written. CR1=0x%x",
                        (CR1 & ~CR1_WRMASK) | (datain_reg & CR1_WRMASK));
                    end
                    8'd1: begin
                      CR2 <= (CR2 & ~CR2_WRMASK) | (datain_reg & CR2_WRMASK);
                      $display("[MODEL] CR2 written. CR2=0x%x",
                        (CR2 & ~CR2_WRMASK) | (datain_reg & CR2_WRMASK));
                    end
                    8'd2: begin
                      CR3 <= (CR3 & ~CR3_WRMASK) | (datain_reg & CR3_WRMASK);
                      $display("[MODEL] CR3 written. CR3=0x%x", 
                        (CR3 & ~CR3_WRMASK) | (datain_reg & CR3_WRMASK));
                    end
                    8'd3: begin
                      CR4 <= (CR4 & ~CR4_WRMASK) | (datain_reg & CR4_WRMASK);
                      $display("[MODEL] CR4 written. CR4=0x%x",
                        (CR4 & ~CR4_WRMASK) | (datain_reg & CR4_WRMASK));
                    end
                  endcase
                end
              end
              else begin
                if (s_datain_done)
                  $display("[MODEL] WREN bit not set CRx not written");
              end
            end

          endcase
      endcase
    end

  /* Augmented Array Protection Register */
  always@(negedge CLK)
    case(state)
      S_DATAIN:
        if (s_datain_done && (cmd_latch == WRAP))
          if (wren) begin
            if (!asplk) begin
              ASP <= datain_reg;
              $display("[MODEL] ASP written. ASP=0x%X", datain_reg);
            end
            else begin
              $display("[MODEL] ASP not written (ASPLK set)");
            end
          end
          else begin
            $display("[MODEL] WREN bit not set. ASP not written");
          end

    endcase

  /* Serial Number Register */
  always@(negedge CLK)
    if ((state == S_DATAIN) &&
        (cmd_latch == WRSN) &&
        byte_strobe) begin
      if (!snpen) begin
        if (wren) begin
          case(byte_strobe_ctr)
            8'd0: SN[63:56] <= datain_reg;
            8'd1: SN[55:48] <= datain_reg;
            8'd2: SN[47:40] <= datain_reg;
            8'd3: SN[39:32] <= datain_reg;
            8'd4: SN[31:24] <= datain_reg;
            8'd5: SN[23:16] <= datain_reg;
            8'd6: SN[15:8]  <= datain_reg;
            8'd7: begin 
              SN[7:0]   <= datain_reg;
              $display("[MODEL] SN written. SN=0x%X", {SN[63:16],datain_reg});
            end
          endcase
        end
        else begin
          if (byte_strobe_ctr == 0)
            $display("[MODEL] WREN bit not set. SN not written");
        end
      end
      else begin
        if (byte_strobe_ctr == 0)
          $display("[MODEL] SNPEN is set. SN not written");
      end
    end

  /* Data out control */
  always@(*) begin
    dataout_reg_latch = 0;
    dataout_reg_shift = 0;
    case(state)
      S_CMD:
        if (s_cmd_done)
          case(datain_reg)
            RDSR: begin dataout_reg_latch = 1'b1; dataout_reg = SR; end
            RDCX: begin dataout_reg_latch = 1'b1; dataout_reg = CR1; end
            RDAP: begin dataout_reg_latch = 1'b1; dataout_reg = ASP; end
            RDSN: begin dataout_reg_latch = 1'b1; dataout_reg = SN[63:56]; end
            RDC1: begin dataout_reg_latch = 1'b1; dataout_reg = CR1; end
            RDC2: begin dataout_reg_latch = 1'b1; dataout_reg = CR2; end
            RDC3: begin dataout_reg_latch = 1'b1; dataout_reg = CR3; end
            RDC4: begin dataout_reg_latch = 1'b1; dataout_reg = CR4; end
            RDID: begin dataout_reg_latch = 1'b1; dataout_reg = IDREG[31:24]; end
            RUID: begin dataout_reg_latch = 1'b1; dataout_reg = UIDREG[63:56]; end
          endcase

      S_ADDR:
        if (s_addr_done)
          case(cmd_latch)
            READ: begin
              dataout_reg_latch = 1'b1;
              dataout_reg = mem_array[rd_addr];
            end
            RDAS: begin
              if (!aug_oor) begin
                dataout_reg_latch = 1'b1;
                dataout_reg = aug_array[rd_addr[7:0]];
              end
              else begin
                $display("[MODEL] Augmented array read address out of range: 0x%X", rd_addr);
              end
            end
          endcase

      S_MODE, SD_MODE:
        if (s_mode_done && (mlats == 4'd0)) // Special case 0 latency
        begin
          case(cmd_latch)
            RDFT, DRFR, RDDO, RDQO, RDDI, RDQI, DRDI, DRQI:
            begin
              dataout_reg_latch = 1'b1;
              dataout_reg = mem_array[rd_addr];
            end
          endcase
        end

      S_LATENCY:
        if (s_latency_done)
          case(cmd_latch)
            RDFT, DRFR, RDDO, RDQO, RDDI, RDQI, DRDI, DRQI: 
            begin
              dataout_reg_latch = 1'b1;
              dataout_reg = mem_array[rd_addr];
            end

            RDAR: begin
              dataout_reg_latch = 1'b1;
              case(addr)
                ADDR_SR : dataout_reg = SR;
                ADDR_CR1: dataout_reg = CR1;
                ADDR_CR2: dataout_reg = CR2;
                ADDR_CR3: dataout_reg = CR3;
                ADDR_CR4: dataout_reg = CR4;
                ADDR_ID : dataout_reg = IDREG[31:24];
                ADDR_UID: dataout_reg = UIDREG[63:56];
                default : begin 
                  dataout_reg = 0;
                  $display("[MODEL] Invalid Register Address 0x%x", addr);
                end
              endcase
            end
          endcase

      S_DATAOUT, SD_DATAOUT:
        case(cmd_latch)
          RDSR, RDAP, RDC1, RDC2, RDC3, RDC4: 
            dataout_reg_shift = 1'b1;

          RDID: begin
            dataout_reg_latch = byte_strobe;
            dataout_reg_shift = ~byte_strobe;
            case(byte_strobe_ctr)
              8'd0: dataout_reg = IDREG[23:16];
              8'd1: dataout_reg = IDREG[15:8];
              8'd2: dataout_reg = IDREG[7:0];
            endcase
          end

          RUID: begin
            dataout_reg_latch = byte_strobe;
            dataout_reg_shift = ~byte_strobe;
            case(byte_strobe_ctr)
              8'd0: dataout_reg = UIDREG[55:48];
              8'd1: dataout_reg = UIDREG[47:40];
              8'd2: dataout_reg = UIDREG[39:32];
              8'd3: dataout_reg = UIDREG[31:24];
              8'd4: dataout_reg = UIDREG[23:16];
              8'd5: dataout_reg = UIDREG[15:8];
              8'd6: dataout_reg = UIDREG[7:0];
            endcase
          end

          RDCX: begin
            dataout_reg_latch = byte_strobe;
            dataout_reg_shift = ~byte_strobe;
            case(byte_strobe_ctr)
              8'd0: dataout_reg = CR2;
              8'd1: dataout_reg = CR3;
              8'd2: dataout_reg = CR4;
              default: dataout_reg = 0;
            endcase
          end

          RDAR: begin
            dataout_reg_latch = byte_strobe;
            dataout_reg_shift = ~byte_strobe;
            case(addr)
              ADDR_ID:
                case(byte_strobe_ctr)
                  8'd0: dataout_reg = IDREG[23:16];
                  8'd1: dataout_reg = IDREG[15:8];
                  8'd2: dataout_reg = IDREG[7:0];
                endcase

              ADDR_UID:
                case(byte_strobe_ctr)
                  8'd0: dataout_reg = UIDREG[55:48];
                  8'd1: dataout_reg = UIDREG[47:40];
                  8'd2: dataout_reg = UIDREG[39:32];
                  8'd3: dataout_reg = UIDREG[31:24];
                  8'd4: dataout_reg = UIDREG[23:16];
                  8'd5: dataout_reg = UIDREG[15:8];
                  8'd6: dataout_reg = UIDREG[7:0];
                endcase
            endcase
          end

          RDSN: begin
            dataout_reg_latch = byte_strobe;
            dataout_reg_shift = ~byte_strobe;
            case(byte_strobe_ctr)
              8'd0: dataout_reg = SN[55:48];
              8'd1: dataout_reg = SN[47:40];
              8'd2: dataout_reg = SN[39:32];
              8'd3: dataout_reg = SN[31:24];
              8'd4: dataout_reg = SN[23:16];
              8'd5: dataout_reg = SN[15:8];
              8'd6: dataout_reg = SN[7:0];
            endcase
          end

          READ, RDFT, DRFR, RDDO, RDQO, RDDI, RDQI, DRDI, DRQI:
          begin
            dataout_reg_latch = byte_strobe;
            dataout_reg_shift = ~byte_strobe;
            dataout_reg = mem_array[rd_addr];
          end

          RDAS: begin
            if (!aug_oor) begin
              dataout_reg_latch = byte_strobe; 
              dataout_reg_shift = ~byte_strobe;
              dataout_reg = aug_array[rd_addr];
            end
          end

        endcase

    endcase
  end

  /* Read address out of range notifier */
  reg dataout_reg_latch_d, aug_oor_d;
  always@(posedge CLK) begin
    dataout_reg_latch_d <= dataout_reg_latch;
    aug_oor_d <= aug_oor;
  end
 
  always@(negedge CLK) begin
    if ((cmd_latch == RDAS) &&
        (dataout_reg_latch_d & aug_oor_d) &&
        (sclk_re_ctr == 0)) begin
      $display("[MODEL] Augmented array read address out of range: 0x%X", rd_addr);
    end
  end

  /* Array write addr */
  reg [23:0] wr_ctr;
  assign wr_addr = addr + wr_ctr;
  always@(posedge CLK)
    if ((state == S_DATAIN) || (state == SD_DATAIN)) begin
      case(cmd_latch)
        WRTE, WRFT, WRAS, DRFW, WDUI, WQDI, WDIO, WQIO, DWQI, DWQO:
          if (s_datain_done)
            wr_ctr <= wr_ctr + 1;
        default:
          wr_ctr <= 0;
      endcase
    end
    else begin
      wr_ctr <= 0;
    end

  /* Array read addr */
  reg [23:0] rd_ctr;
  always@(posedge CLK)
    if ((state == S_DATAOUT) || (state == SD_DATAOUT)) begin
      case(cmd_latch)
        READ, RDFT, RDAS, DRFR, RDDO, RDQO, RDDI, RDQI, DRDI, DRQI:
         begin
          if (s_dataout_done) begin

            if (!wraps)
              rd_ctr <= rd_ctr + 1;

            else // Wrap enabled
              case(wrpls)
                3'b000: begin
                  if (rd_ctr[3:0] == 4'hF)
                    rd_ctr[3:0] <= 4'h0;
                  else
                    rd_ctr[3:0] <= rd_ctr[3:0] + 1;
                end

                3'b001: begin
                  if (rd_ctr[4:0] == 5'h1F)
                    rd_ctr[4:0] <= 5'h0;
                  else
                    rd_ctr[4:0] <= rd_ctr[4:0] + 1;
                end

                3'b010: begin
                  if (rd_ctr[5:0] == 6'h3F)
                    rd_ctr[5:0] <= 6'h0;
                  else
                    rd_ctr[5:0] <= rd_ctr[5:0] + 1;
                end

                3'b011: begin
                  if (rd_ctr[6:0] == 7'h7F)
                    rd_ctr[6:0] <= 7'h0;
                  else
                    rd_ctr[6:0] <= rd_ctr[6:0] + 1;
                end

                3'b100: begin
                  if (rd_ctr[7:0] == 8'hFF)
                    rd_ctr[7:0] <= 8'h0;
                  else
                    rd_ctr[7:0] <= rd_ctr[7:0] + 1;
                end

                default: begin
                  rd_ctr <= rd_ctr + 1;
                end
              endcase

          end
          if (rd_ctr == 1) begin
            case(cmd_latch)
             READ, RDAS: if(SCK_period < 200)
                     $display("[MODEL] Flagging READ. Frequency is greater than 50MHz violation");
             DRFR, DRDI, DRQI: if(SCK_period < 185)
                     $display("[MODEL] Flagging READ. Frequency is greater than 54MHz violation");
             RDFT, RDDO, RDQO, RDDI, RDQI: if(SCK_period < 92.5)
                     $display("[MODEL] Flagging READ. Frequency is greater than 108MHz violation");
            endcase
          end
         end
        default:
          rd_ctr <= 1;
      endcase
    end
    else begin // State before S_DATAOUT
      if (!wraps)
        rd_ctr <= 1;
      else // take lower bits of first address
        if (cmd_latch == READ) begin
          case(wrpls)
            3'b000:
              rd_ctr <= {20'd0, sr_addr[3:0]} + 1;
            3'b001:
              rd_ctr <= {19'd0, sr_addr[4:0]} + 1;
            3'b010:
              rd_ctr <= {18'd0, sr_addr[5:0]} + 1;
            3'b011:
              rd_ctr <= {17'd0, sr_addr[6:0]} + 1;
            3'b100:
              rd_ctr <= {16'd0, sr_addr[7:0]} + 1;
            default:
              rd_ctr <= 1;
          endcase
        end
        else begin
          case(wrpls)
            3'b000:
              rd_ctr <= {20'd0, addr[3:0]} + 1;
            3'b001:
              rd_ctr <= {19'd0, addr[4:0]} + 1;
            3'b010:
              rd_ctr <= {18'd0, addr[5:0]} + 1;
            3'b011:
              rd_ctr <= {17'd0, addr[6:0]} + 1;
            3'b100:
              rd_ctr <= {16'd0, addr[7:0]} + 1;
            default:
              rd_ctr <= 1;
          endcase
        end
    end

  always@(*)
    case(cmd_latch)
      READ, RDFT, RDAS, DRFR, RDDO, RDQO, RDDI, RDQI, DRDI, DRQI: 
      begin
        case(state)
          S_ADDR, SD_ADDR:
            if (s_addr_done) begin 
              case(cmd_latch)
                DRFR, DRDI, DRQI:
                  rd_addr <= sr_addr_ddr;
                default:
                  rd_addr <= sr_addr;
              endcase
            end
            else
              rd_addr <= 0;

          S_MODE, SD_MODE, S_LATENCY:
            rd_addr <= addr;

          S_DATAOUT, SD_DATAOUT:
            if (!wraps)
              rd_addr <= addr + rd_ctr;
            else
              case(wrpls)
                3'b000: rd_addr <= {addr[23:4], rd_ctr[3:0]};
                3'b001: rd_addr <= {addr[23:5], rd_ctr[4:0]};
                3'b010: rd_addr <= {addr[23:6], rd_ctr[5:0]};
                3'b011: rd_addr <= {addr[23:7], rd_ctr[6:0]};
                3'b100: rd_addr <= {addr[23:8], rd_ctr[7:0]};
                default: rd_addr <= addr + rd_ctr;
              endcase
             
        endcase
      end

      default: begin
        rd_addr <= 0;
      end
    endcase


  /* Memory Array: SDR writes */
  always@(negedge CLK)
    case(cmd_latch)
      WRTE, WRFT, WDUI, WQDI, WDIO, WQIO: begin
        if (s_datain_done)
          if (!addr_prot) begin
            //mem_array[wr_addr[20:0]) = datain_reg;
            mem_array[wr_addr] = datain_reg;
            $display("[MODEL] Data 0x%X written to Addr 0x%X", datain_reg, wr_addr);
          end
          else
            $display("[MODEL] Address protected or out of range: 0x%X", wr_addr);
      end
    endcase

  /* Memory Array: DDR writes */
  always@(posedge CLK or posedge CSn)
    case(cmd_latch)
      DRFW, DWQI, DWQO: begin
        if (s_datain_done)
          if (!addr_prot) begin
            mem_array[wr_addr] = datain_reg;
            $display("[MODEL] Data 0x%X written to Addr 0x%X", datain_reg, wr_addr);
          end
          else
            $display("[MODEL] Address protected or out of range: 0x%X", wr_addr);
      end
    endcase

  /* Augmented Storage Array */
  always@(negedge CLK)
    if ((cmd_latch == WRAS) && s_datain_done)
      if (!aug_prot) begin
        aug_array[wr_addr] = datain_reg;
        $display("[MODEL] Data 0x%X written to Augmented array addr 0x%X", datain_reg, wr_addr);
      end
      else begin
        $display("[MODEL] Augmented array address protected or out of range: 0x%X", wr_addr);
      end

  /* Mode Latch */
  //always@(negedge CLK)
  always@(posedge CLK)
    if (s_mode_done) 
      case(datain_reg[7:4]) // Latch only correct values of mode byte
        4'hA: begin
          mode_latch <= datain_reg;
          $display("[MODEL] XIP enabled");
        end
        4'hF: begin
          mode_latch <= datain_reg;
          $display("[MODEL] XIP disabled");
        end
        default: begin
          if (xip_en)
            $display("[MODEL] Invalid XIP mode byte. Retaining XIP enabled");
          else
            $display("[MODEL] Invalid XIP mode byte. Retaining XIP disabled");
        end
      endcase

  /* XIP mode enabled */
  always@(*)
    case(mode_latch[7:4])
      4'hA: xip_en = 1'b1; 
      4'hF: xip_en = 0; 
      default: xip_en = 0;
    endcase

  /* Initializations */
  always@(posedge PORn) begin
    state = S_IDLE;
    dqout_en = 0;
    SR = SR_DEFAULT;
    CR1 = CR1_DEFAULT; 
    CR2 = CR2_DEFAULT; 
    CR3 = CR3_DEFAULT; 
    CR4 = CR4_DEFAULT; 
    mode_latch = 8'hF0; // XIP disabled
    ASP = ASP_DEFAULT;
    SN = SN_DEFAULT;
    swrst_en = 0;
  end

  /* Software reset enable  */
  always@(negedge CLK)
    if (s_cmd_done) begin
      if (cmd_latch == SRTE)
        swrst_en = 1'b1;
      else
        swrst_en = 0; // disable SRTE when another command sent
    end

  /* Hibernate Enter/Exit counter */
  reg [31:0] hb_ent_ctr, hb_ext_ctr;

  always begin
    #1
    if (state == S_HBENT)
      hb_ent_ctr <= hb_ent_ctr + 1;
    else
      hb_ent_ctr <= 0;
  end

  always begin
    #1
    if (state == S_HBEXT)
      hb_ext_ctr <= hb_ext_ctr + 1;
    else
      hb_ext_ctr <= 0;
  end

  /* Hibernate Sleep/Wake delay triggers */
  always@(*) begin
    if (hb_ent_ctr >= tENTHIB)
      hb_ent_done <= 1'b1;
    else
      hb_ent_done <= 0;
  end

  always@(*) begin
    if (hb_ext_ctr >= tEXHIB)
      hb_ext_done <= 1'b1;
    else
      hb_ext_done <= 0;
  end

  /* Deep Power Down Enter/Exit counter */
  reg [31:0] dp_ent_ctr, dp_ext_ctr;

  always begin
    #1
    if (state == S_DPENT)
      dp_ent_ctr <= dp_ent_ctr + 1;
    else
      dp_ent_ctr <= 0;
  end

  always begin
    #1
    if (state == S_DPEXT)
      dp_ext_ctr <= dp_ext_ctr + 1;
    else
      dp_ext_ctr <= 0;
  end

  /* Deep Power Down CS_active counter */
  reg [31:0] dp_csact_ctr;
  always begin
    #1
    if (((state == S_DEEP) || (state == S_DPCMD)) && !CSn)
      dp_csact_ctr <= dp_csact_ctr + 1;
    else
      dp_csact_ctr <= 0;
  end
  
  /* Deep Power Down Sleep/Wake delay triggers */
  always@(*) begin
    if (dp_ent_ctr >= tEDPD)
      dp_ent_done <= 1'b1;
    else
      dp_ent_done <= 0;
  end

  always@(*) begin
    if (dp_ext_ctr >= tEXDPD)
      dp_ext_done <= 1'b1;
    else
      dp_ext_done <= 0;
  end

  /* Deep Power Down CS wake trigger */
  always@(*) begin
    if (dp_csact_ctr >= (tCSDPD-1))
      dp_cswake <= 1'b1;
    else
      dp_cswake <= 0;
  end

endmodule

/*
 * Shift register data in: Serial
 */
module srdin_ser(
  input  wire       sclk,
  input  wire       cs_n,
  input  wire       din,
  output reg  [7:0] dout
  );

  always@(posedge sclk)
    if (!cs_n)
      dout <= {dout[6:0], din};

endmodule

/*
 * Shift register data in: Dual
 */
module srdin_dual(
  input  wire       sclk,
  input  wire       cs_n,
  input  wire [1:0] din,
  output reg  [7:0] dout
  );

  always@(posedge sclk)
    if (!cs_n)
      dout <= {dout[5:0], din};

endmodule

/*
 * Shift register data in: Quad
 */
module srdin_quad(
  input  wire       sclk,
  input  wire       cs_n,
  input  wire [3:0] din,
  output reg  [7:0] dout
  );

  always@(posedge sclk)
    if (!cs_n)
      dout <= {dout[3:0], din};

endmodule

/*
 * Shift register data in: Serial DDR
 */
module srdin_ser_ddr(
  input  wire       sclk,
  input  wire       cs_n,
  input  wire       din,
  output reg  [7:0] dout
  );

  always@(posedge sclk or negedge sclk)
    if (!cs_n)
      dout <= {dout[6:0], din};

endmodule

/*
 * Shift register data in: Dual DDR
 */
module srdin_dual_ddr(
  input  wire       sclk,
  input  wire       cs_n,
  input  wire [1:0] din,
  output reg  [7:0] dout
  );

  always@(posedge sclk or negedge sclk)
    if (!cs_n)
      dout <= {dout[5:0], din};

endmodule

/*
 * Shift register data in: Quad DDR
 */
module srdin_quad_ddr(
  input  wire       sclk,
  input  wire       cs_n,
  input  wire [3:0] din,
  output reg  [7:0] dout
  );

  always@(posedge sclk or negedge sclk)
    if (!cs_n)
      dout <= {dout[3:0], din};

endmodule

/*
 * Shift register data out: Serial
 */
module srdout_ser(
  input  wire       sclk,
  input  wire       cs_n,
  input  wire       latch,
  input  wire       shift,
  input  wire [7:0] din,
  output wire       dout,
  output wire       ddr
  );

  reg [7:0] sr;
  assign dout = sr[7];
  reg [7:0] sr_ddr;
  assign ddr = sr_ddr[7];

  initial begin
    sr = 0;
    sr_ddr = 0;
  end

  always@(negedge sclk)
    if (!cs_n)
      case({latch, shift})
        2'b10: sr = din;
        2'b01: sr = {sr[6:0], sr[7]};
      endcase

  always@(negedge sclk or posedge sclk)
    if (!cs_n)
      case({latch, shift})
        2'b10: begin
          if (!sclk)
            sr_ddr = din; // negedge latch
          else
            sr_ddr = {sr_ddr[6:0], sr_ddr[7]}; // posedge shift
        end
        2'b01: sr_ddr = {sr_ddr[6:0], sr_ddr[7]};
      endcase

endmodule

/*
 * Shift register data out: Dual
 */
module srdout_dual(
  input  wire       sclk,
  input  wire       cs_n,
  input  wire       latch,
  input  wire       shift,
  input  wire [7:0] din,
  output wire [1:0] dout,
  output wire [1:0] ddr
  );

  reg [7:0] sr;
  assign dout = sr[7:6];
  reg [7:0] sr_ddr;
  assign ddr  = sr_ddr[7:6];

  initial begin
    sr = 0;
    sr_ddr = 0;
  end

  always@(negedge sclk)
    if (!cs_n)
      case({latch, shift})
        2'b10: sr = din;
        2'b01: sr = {sr[5:0], sr[7:6]};
      endcase

  always@(negedge sclk or posedge sclk)
    if (!cs_n)
      case({latch, shift})
        2'b10: begin
          if (!sclk)
            sr_ddr = din;
          else
            sr_ddr = {sr_ddr[5:0], sr_ddr[7:6]};
        end
        2'b01: sr_ddr = {sr_ddr[5:0], sr_ddr[7:6]};
      endcase

endmodule

/*
 * Shift register data out: Quad
 */
module srdout_quad(
  input  wire       sclk,
  input  wire       cs_n,
  input  wire       latch,
  input  wire       shift,
  input  wire [7:0] din,
  output wire [3:0] dout,
  output wire [3:0] ddr
  );

  reg [7:0] sr;
  assign dout = sr[7:4];

  reg [7:0] sr_ddr;
  assign ddr = sr_ddr[7:4];

  initial begin
    sr = 0;
    sr_ddr = 0;
  end

  always@(negedge sclk)
    if (!cs_n)
      case({latch, shift})
        2'b10: sr = din;
        2'b01: sr = {sr[3:0], sr[7:4]};
      endcase

  always@(negedge sclk or posedge sclk)
    if (!cs_n)
      case({latch, shift})
        2'b10: begin
          if (!sclk)
            sr_ddr = din;
          else
            sr_ddr = {sr_ddr[3:0], sr_ddr[7:4]};
        end
        2'b01: sr_ddr = {sr_ddr[3:0], sr_ddr[7:4]};
      endcase

endmodule

/*
 * Address register
 */
module addr_reg(
  input  wire        sclk,
  input  wire        cs_n,
  input  wire [2:0]  mode_sel,
  input  wire [3:0]  din,
  output reg  [23:0] addr
  //output reg  [23:0] pre_addr
  );

  // SPI register
  reg [23:0] spi_sr;
  always@(posedge sclk)
    if (!cs_n)
      spi_sr <= {spi_sr[22:0], din[0]};
 
  // DPI register
  reg [24:0] dpi_sr;
  always@(posedge sclk)
    if (!cs_n)
      dpi_sr <= {dpi_sr[21:0], din[1:0]};
 
  // QPI register
  reg [24:0] qpi_sr;
  always@(posedge sclk)
    if (!cs_n)
      qpi_sr <= {qpi_sr[19:0], din};
 
  always@(*)
    case(mode_sel)
      3'b001: addr <= spi_sr;
      3'b010: addr <= dpi_sr;
      3'b100: addr <= qpi_sr;
      default: addr <= 0;
    endcase

endmodule

/*
 * Address register DDR
 */
module addr_reg_ddr(
  input  wire        sclk,
  input  wire        cs_n,
  input  wire [2:0]  mode_sel,
  input  wire [3:0]  din,
  output reg  [23:0] addr
  //output reg  [23:0] pre_addr
  );

  // SPI register
  reg [23:0] spi_sr;
  always@(posedge sclk or negedge sclk)
    if (!cs_n)
      spi_sr <= {spi_sr[22:0], din[0]};
 
  // DPI register
  reg [24:0] dpi_sr;
  always@(posedge sclk or negedge sclk)
    if (!cs_n)
      dpi_sr <= {dpi_sr[21:0], din[1:0]};
 
  // QPI register
  reg [24:0] qpi_sr;
  always@(posedge sclk or negedge sclk)
    if (!cs_n)
      qpi_sr <= {qpi_sr[19:0], din};
 
  always@(*)
    case(mode_sel)
      3'b001: addr <= spi_sr;
      3'b010: addr <= dpi_sr;
      3'b100: addr <= qpi_sr;
      default: addr <= 0;
    endcase

endmodule
