`timescale 100ps / 10ps

/*
 * Options for modeling byte transmission (uncomment at most 1)
 * - SUHDCHK_ZHD
 *   - Line is released (data=HiZ) after hold time
 * - SUHDCK_CKFE
 *   - Data is set during falling edge of CLK, line released at falling edge
 *     of CLK
 * - default (no option uncommented)
 *   - Data is set to 0 after hold time, line released (tCL-tSU) after
 *     last falling edge of CLK
 *
 */
//`define SUHDCHK_ZHD
`define SUHDCHK_CKFE

module stimulus 
	(
		SO_IO1,
		CLK,
		IO3, 
		SI_IO0 ,
		WPn_IO2,
		CSn
	);

  inout SO_IO1;
	inout IO3;
	inout SI_IO0;
	inout WPn_IO2;
	
	output CSn;
	output CLK;
        
	reg CSn = 1;
	reg CLK = 0;
        
	reg SI_IO0_reg=0;
  reg SO_IO1_reg=0;
  reg WPn_IO2_reg=0;
  reg IO3_reg=0;
        
  reg [3:0] tx;

  parameter SPI = 3'h01;
  parameter DPI = 3'h02;
  parameter QPI = 3'h04;

  `include "../shr/ref/task.v"
	 
  assign SI_IO0 = tx[0] ? SI_IO0_reg : 1'bz;  
  assign SO_IO1 = tx[1] ? SO_IO1_reg : 1'bz;  
  assign WPn_IO2 = tx[2] ? WPn_IO2_reg : 1'bz;  
  assign IO3 = tx[3] ? IO3_reg : 1'bz;  

  initial begin
    tx = 4'd0;
    #10000;

    /* Top block protect, All */
    $display("==== [TEST] Top block protect: All ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h1234, 8'h50, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1234, 8'h50, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b0, 3'b111, 2'h0});
    WriteSeq(SPI, 24'h1234, 8'hB0, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1234, 8'hB0, 8'h7, 8'h0, 24'd4);

    /* Top block protect, Upper 1/2 */
    $display("==== [TEST] Top block protect: Upper 1/2 (16Mbit)  ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'hFFFFE, 8'h50, 8'h7, 24'd4);
    ReadSeq(SPI, 24'hFFFFE, 8'h50, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b0, 3'b110, 2'h0});
    WriteSeq(SPI, 24'hFFFFE, 8'hB0, 8'h7, 24'd4);
    ReadSeq(SPI, 24'hFFFFE, 8'hB0, 8'h7, 8'h0, 24'd4);

    /* Top block protect, Upper 1/4 */
    $display("==== [TEST] Top block protect: Upper 1/4 (16Mbit)  ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h17FFFE, 8'h50, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h17FFFE, 8'h50, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b0, 3'b101, 2'h0});
    WriteSeq(SPI, 24'h17FFFE, 8'hB0, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h17FFFE, 8'hB0, 8'h7, 8'h0, 24'd4);

    /* Top block protect, Upper 1/8 */
    $display("==== [TEST] Top block protect: Upper 1/8 (16Mbit)  ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h1BFFFE, 8'h50, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1BFFFE, 8'h50, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b0, 3'b100, 2'h0});
    WriteSeq(SPI, 24'h1BFFFE, 8'hB0, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1BFFFE, 8'hB0, 8'h7, 8'h0, 24'd4);

    /* Top block protect, Upper 1/16 */
    $display("==== [TEST] Top block protect: Upper 1/16 (16Mbit)  ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h1DFFFE, 8'h50, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1DFFFE, 8'h50, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b0, 3'b011, 2'h0});
    WriteSeq(SPI, 24'h1DFFFE, 8'hB0, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1DFFFE, 8'hB0, 8'h7, 8'h0, 24'd4);

    /* Top block protect, Upper 1/32 */
    $display("==== [TEST] Top block protect: Upper 1/32 (16Mbit)  ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h1EFFFE, 8'h50, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1EFFFE, 8'h50, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b0, 3'b010, 2'h0});
    WriteSeq(SPI, 24'h1EFFFE, 8'hB0, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1EFFFE, 8'hB0, 8'h7, 8'h0, 24'd4);

    /* Top block protect, Upper 1/64 */
    $display("==== [TEST] Top block protect: Upper 1/64 (16Mbit)  ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h1F7FFE, 8'h50, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1F7FFE, 8'h50, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b0, 3'b001, 2'h0});
    WriteSeq(SPI, 24'h1F7FFE, 8'hB0, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1F7FFE, 8'hB0, 8'h7, 8'h0, 24'd4);

    /* Bottom block protect, All */
    $display("==== [TEST] Bottom block protect: All ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h1fabcd, 8'h66, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1fabcd, 8'h66, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b1, 3'b111, 2'h0});
    WriteSeq(SPI, 24'h1fabcd, 8'hC6, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h1fabcd, 8'hC6, 8'h7, 8'h0, 24'd4);

    /* Bottom block protect, Upper 1/2 */
    $display("==== [TEST] Bottom block protect: Lower 1/2 (16Mbit) ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h0FFFFE, 8'h66, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h0FFFFE, 8'h66, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b1, 3'b110, 2'h0});
    WriteSeq(SPI, 24'h0FFFFE, 8'hC6, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h0FFFFE, 8'hC6, 8'h7, 8'h0, 24'd4);

    /* Bottom block protect, Upper 1/4 */
    $display("==== [TEST] Bottom block protect: Lower 1/4 (16Mbit) ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h07FFFE, 8'h66, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h07FFFE, 8'h66, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b1, 3'b101, 2'h0});
    WriteSeq(SPI, 24'h07FFFE, 8'hC6, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h07FFFE, 8'hC6, 8'h7, 8'h0, 24'd4);

    /* Bottom block protect, Upper 1/8 */
    $display("==== [TEST] Bottom block protect: Lower 1/8 (16Mbit) ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h03FFFE, 8'h66, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h03FFFE, 8'h66, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b1, 3'b100, 2'h0});
    WriteSeq(SPI, 24'h03FFFE, 8'hC6, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h03FFFE, 8'hC6, 8'h7, 8'h0, 24'd4);

    /* Bottom block protect, Upper 1/16 */
    $display("==== [TEST] Bottom block protect: Lower 1/16 (16Mbit) ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h01FFFE, 8'h66, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h01FFFE, 8'h66, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b1, 3'b011, 2'h0});
    WriteSeq(SPI, 24'h01FFFE, 8'hC6, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h01FFFE, 8'hC6, 8'h7, 8'h0, 24'd4);

    /* Bottom block protect, Upper 1/32 */
    $display("==== [TEST] Bottom block protect: Lower 1/32 (16Mbit) ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h00FFFE, 8'h66, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h00FFFE, 8'h66, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b1, 3'b010, 2'h0});
    WriteSeq(SPI, 24'h00FFFE, 8'hC6, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h00FFFE, 8'hC6, 8'h7, 8'h0, 24'd4);

    /* Bottom block protect, Upper 1/64 */
    $display("==== [TEST] Bottom block protect: Lower 1/64 (16Mbit) ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h0);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h0, 8'h08, 8'h0, 8'h01);
    WriteSeq(SPI, 24'h007FFE, 8'h66, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h007FFE, 8'h66, 8'h7, 8'h0, 24'd4);

    WriteEnable(SPI);
    WriteStatus(SPI, {2'h0, 1'b1, 3'b001, 2'h0});
    WriteSeq(SPI, 24'h007FFE, 8'hC6, 8'h7, 24'd4);
    ReadSeq(SPI, 24'h007FFE, 8'hC6, 8'h7, 8'h0, 24'd4);

    $finish;
  end
endmodule
