`timescale 100ps / 10ps

/*
 * Options for modeling byte transmission (uncomment at most 1)
 * - SUHDCHK_ZHD
 *   - Line is released (data=HiZ) after hold time
 * - SUHDCK_CKFE
 *   - Data is set during falling edge of CLK, line released at falling edge
 *     of CLK
 * - default (no option uncommented)
 *   - Data is set to 0 after hold time, line released (tCL-tSU) after
 *     last falling edge of CLK
 *
 */
//`define SUHDCHK_ZHD
`define SUHDCHK_CKFE

module stimulus 
	(
		SO_IO1,
		CLK,
		IO3, 
		SI_IO0 ,
		WPn_IO2,
		CSn
	);

  inout SO_IO1;
	inout IO3;
	inout SI_IO0;
	inout WPn_IO2;
	
	output CSn;
	output CLK;
        
	reg CSn = 1;
	reg CLK = 0;
        
	reg SI_IO0_reg=0;
  reg SO_IO1_reg=0;
  reg WPn_IO2_reg=0;
  reg IO3_reg=0;
        
  reg [3:0] tx;

  parameter SPI = 3'h01;
  parameter DPI = 3'h02;
  parameter QPI = 3'h04;

  `include "../shr/ref/task.v"
	 
  assign SI_IO0 = tx[0] ? SI_IO0_reg : 1'bz;  
  assign SO_IO1 = tx[1] ? SO_IO1_reg : 1'bz;  
  assign WPn_IO2 = tx[2] ? WPn_IO2_reg : 1'bz;  
  assign IO3 = tx[3] ? IO3_reg : 1'bz;  

  /* Expected ERROR: 88 */
  initial begin
    tx = 4'd0;
    #10000;
    $display("==== [TEST] WREN and WRDI (SPI) ====");
    WriteEnable(SPI);
    WriteDisable(SPI);
    WriteEnable(SPI);
    ReadStatus(SPI, 8'h02);
    WriteDisable(SPI);
    ReadStatus(SPI, 8'h00);

    /* Successful write to SR */
    $display("==== [TEST] Successful write (SPI) to SR ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'hFF);
    ReadStatus(SPI, 8'hFF);

    /* Failed write to SR (No WREN) */
    $display("==== [TEST] Failed write (SPI) to SR (no WREN) ==== ");
    WriteStatus(SPI, 8'hAA);
    ReadStatus(SPI, 8'hAA);

    /* Switch from SPI to DPI */
    $display("==== [TEST] Switch to DPI then DPI RDSR ==== ");
    DPIEnable(SPI);
    ReadStatus(DPI, 8'hFF);

    /* Switch from DPI to SPI */
    $display("==== [TEST] Switch to SPI then SPI RDSR ==== ");
    SPIEnable(DPI);
    ReadStatus(SPI, 8'hFF);

    /* Switch from SPI to QPI */
    $display("==== [TEST] Switch to QPI then QPI RDSR ==== ");
    QPIEnable(SPI);
    ReadStatus(QPI, 8'hFF);

    /* Switch from QPI to SPI */
    $display("==== [TEST] Switch to SPI then SPI RDSR ==== ");
    SPIEnable(QPI);
    ReadStatus(SPI, 8'hFF);

    /* Switch from SPI to DPI */
    $display("==== [TEST] Switch to DPI ==== ");
    DPIEnable(SPI);
    ReadStatus(DPI, 8'hFF);

    /* DPI Register Function checks */
    $display("==== [TEST] DPI register functions  ==== ");
    WriteEnable(DPI);
    WriteStatus(DPI, 8'hA5);
    ReadStatus(DPI, 8'hA5);
    $display("==== [TEST] DPI No WREN  ==== ");
    WriteEnable(DPI);
    WriteDisable(DPI);
    WriteStatus(DPI, 8'hB7);
    ReadStatus(DPI, 8'hB7);

    /* Switch from DPI to QPI */
    $display("==== [TEST] Switch to QPI then QPI RDSR ==== ");
    QPIEnable(DPI);
    ReadStatus(QPI, 8'hA5);

    /* QPI Register Function checks */
    $display("==== [TEST] QPI register functions  ==== ");
    WriteEnable(QPI);
    WriteStatus(QPI, 8'h5A);
    ReadStatus(QPI, 8'h5A);
    $display("==== [TEST] QPI No WREN  ==== ");
    WriteEnable(QPI);
    WriteDisable(QPI);
    WriteStatus(QPI, 8'hB7);
    ReadStatus(QPI, 8'hB7);

    /* Switch from QPI to DPI */
    $display("==== [TEST] Switch to DPI then DPI RDSR ==== ");
    DPIEnable(QPI);
    ReadStatus(DPI, 8'h5A);

    /* Switch to SPI */
    SPIEnable(DPI);
    SPIEnable(SPI);

    /* Test Config Registers */
    $display("==== [TEST] SPI write/read Config reg ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'hFF, 8'hFF, 8'hFF, 8'hFF);
    ReadConfig(SPI, 8'hFF, 8'hFF, 8'hFF, 8'hFF);

    /* Test Config Registers No WREN */
    $display("==== [TEST] SPI write/read Config reg No WREN ==== ");
    WriteConfig(SPI, 8'hAA, 8'hAA, 8'hAA, 8'hAA);
    ReadConfig(SPI, 8'hAA, 8'hAA, 8'hAA, 8'hAA);

    /* Test Config Registers: DPI */
    $display("==== [TEST] DPI write/read Config reg ==== ");
    DPIEnable(SPI);
    DPIEnable(DPI);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'hBB, 8'hBB, 8'hBB, 8'hBB);
    ReadConfig(DPI, 8'hBB, 8'hBB, 8'hBB, 8'hBB);

    /* Test Config Registers No WREN: DPI */
    $display("==== [TEST] DPI write/read Config reg No WREN ==== ");
    WriteConfig(DPI, 8'hAA, 8'hAA, 8'hAA, 8'hAA);
    ReadConfig(DPI, 8'hAA, 8'hAA, 8'hAA, 8'hAA);

    /* Test Config Registers: QPI */
    $display("==== [TEST] QPI write/read Config reg ==== ");
    QPIEnable(DPI);
    QPIEnable(QPI);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'hCD, 8'hCD, 8'hCD, 8'hCD);
    ReadConfig(QPI, 8'hCD, 8'hCD, 8'hCD, 8'hCD);

    /* Test Config Registers No WREN: QPI */
    $display("==== [TEST] QPI write/read Config reg No WREN ==== ");
    WriteConfig(QPI, 8'hAA, 8'hAA, 8'hAA, 8'hAA);
    ReadConfig(QPI, 8'hAA, 8'hAA, 8'hAA, 8'hAA);
    SPIEnable(QPI);

    /* MAPLK Test */
    $display("==== [TEST] MAPLK test (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h00); // MAPLK = 0
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    ReadStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h04, 8'h00, 8'h00, 8'h00); // MAPLK = 1
    WriteEnable(SPI);
    WriteStatus(SPI, 8'hFF); // {TBSEL, BPSEL} = 4'b1111
    ReadStatus(SPI, 8'hFF);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h00); // MAPLK = 0
    WriteEnable(SPI);
    WriteStatus(SPI, 8'hFF); // {TBSEL, BPSEL} = 4'b1111
    ReadStatus(SPI, 8'hFF);

    /* RDAR Test: SPI */
    $display("==== [TEST] RDAR test (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h01, 8'h08, 8'hEF, 8'hA5); // MAPLK = 0
    WriteEnable(SPI);
    WriteStatus(SPI, 8'hFF);
    ReadReg(SPI, 24'habcd12, 8'h00);
    ReadReg(SPI, 24'h0, 8'hFF);
    ReadReg(SPI, 24'h2, 8'h01);
    ReadReg(SPI, 24'h3, 8'h08);
    ReadReg(SPI, 24'h4, 8'hEF);
    ReadReg(SPI, 24'h5, 8'hA5);

    /* RDAR Test: DPI */
    $display("==== [TEST] RDAR test (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h00, 8'h00, 8'h5A); // MAPLK = 0
    WriteEnable(DPI);
    WriteStatus(DPI, 8'h00);
    //ReadReg(DPI, 24'h0, 8'h00);
    ReadReg(DPI, 24'hcd1234, 8'h00);
    ReadReg(DPI, 24'h0, 8'h00);
    ReadReg(DPI, 24'h2, 8'h00);
    ReadReg(DPI, 24'h3, 8'h00);
    ReadReg(DPI, 24'h4, 8'h00);
    ReadReg(DPI, 24'h5, 8'h5A);

    /* RDAR Test: QPI */
    $display("==== [TEST] RDAR test (QPI) ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'hF0, 8'hFF, 8'h5A, 8'hA5); // MAPLK = 0
    WriteEnable(QPI);
    WriteStatus(QPI, 8'hFF);
    //ReadReg(QPI, 24'h0, 8'h00);
    ReadReg(QPI, 24'hef1256, 8'h00);
    ReadReg(QPI, 24'h0, 8'hFF);
    ReadReg(QPI, 24'h2, 8'hF0);
    ReadReg(QPI, 24'h3, 8'hFF);
    ReadReg(QPI, 24'h4, 8'h5A);
    ReadReg(QPI, 24'h5, 8'hA5);
    SPIEnable(QPI);

    /* RDAR ID, UID: SPI */
    $display("==== [TEST] RDAR test for ID, UID (SPI) ==== ");
    ReadReg(SPI, 24'h30, 8'h00);
    ReadReg(SPI, 24'h40, 8'h00);
    DPIEnable(SPI);

    /* RDAR ID, UID: DPI */
    $display("==== [TEST] RDAR test for ID, UID (DPI) ==== ");
    ReadReg(DPI, 24'h30, 8'h00);
    ReadReg(DPI, 24'h40, 8'h00);
    QPIEnable(DPI);

    /* RDAR ID, UID: QPI */
    $display("==== [TEST] RDAR test for ID, UID (QPI) ==== ");
    ReadReg(QPI, 24'h30, 8'h00);
    ReadReg(QPI, 24'h40, 8'h00);
    SPIEnable(QPI);

    /* WRAR Test: SPI */
    $display("==== [TEST] WRAR test (SPI) ==== ");
    WriteEnable(SPI);
    WriteReg(SPI, 24'h0, 8'hA3); // SR
    WriteEnable(SPI);
    WriteReg(SPI, 24'h2, 8'hA3); // CR1
    WriteEnable(SPI);
    WriteReg(SPI, 24'h3, 8'hA3); // CR2 
    WriteEnable(SPI);
    WriteReg(SPI, 24'h4, 8'hA3); // CR3
    WriteEnable(SPI);
    WriteReg(SPI, 24'h5, 8'hA3); // CR4
    ReadReg(SPI, 24'h0, 8'hA3); // SR
    ReadReg(SPI, 24'h2, 8'hA3); // CR1
    ReadReg(SPI, 24'h3, 8'hA3); // CR2 
    ReadReg(SPI, 24'h4, 8'hA3); // CR3
    ReadReg(SPI, 24'h5, 8'hA3); // CR4

    /* WRAR Test: DPI */
    $display("==== [TEST] WRAR test (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteReg(DPI, 24'h0, 8'h7A); // SR
    WriteEnable(DPI);
    WriteReg(DPI, 24'h2, 8'h7A); // CR1
    WriteEnable(DPI);
    WriteReg(DPI, 24'h3, 8'h7A); // CR2 
    WriteEnable(DPI);
    WriteReg(DPI, 24'h4, 8'h7A); // CR3
    WriteEnable(DPI);
    WriteReg(DPI, 24'h5, 8'h7A); // CR4
    ReadReg(DPI, 24'h0, 8'h7A); // SR
    ReadReg(DPI, 24'h2, 8'h7A); // CR1
    ReadReg(DPI, 24'h3, 8'h7A); // CR2 
    ReadReg(DPI, 24'h4, 8'h7A); // CR3
    ReadReg(DPI, 24'h5, 8'h7A); // CR4

    /* WRAR Test: QPI */
    $display("==== [TEST] WRAR test (QPI) ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteReg(QPI, 24'h0, 8'hBB); // SR
    WriteEnable(QPI);
    WriteReg(QPI, 24'h2, 8'hBB); // CR1
    WriteEnable(QPI);
    WriteReg(QPI, 24'h3, 8'hBB); // CR2 
    WriteEnable(QPI);
    WriteReg(QPI, 24'h4, 8'hBB); // CR3
    WriteEnable(QPI);
    WriteReg(QPI, 24'h5, 8'hBB); // CR4
    ReadReg(QPI, 24'h0, 8'hBB); // SR
    ReadReg(QPI, 24'h2, 8'hBB); // CR1
    ReadReg(QPI, 24'h3, 8'hBB); // CR2 
    ReadReg(QPI, 24'h4, 8'hBB); // CR3
    ReadReg(QPI, 24'h5, 8'hBB); // CR4

    /* WRAR Test no WREN: SPI */
    $display("==== [TEST] WRAR test no WREN (SPI) ==== ");
    SPIEnable(QPI);
    WriteReg(SPI, 24'h0, 8'h52); // SR
    WriteReg(SPI, 24'h2, 8'h52); // CR1
    WriteReg(SPI, 24'h3, 8'h52); // CR2 
    WriteReg(SPI, 24'h4, 8'h52); // CR3
    WriteReg(SPI, 24'h5, 8'h52); // CR4
    ReadReg(SPI, 24'h0, 8'h52); // SR
    ReadReg(SPI, 24'h2, 8'h52); // CR1
    ReadReg(SPI, 24'h3, 8'h52); // CR2 
    ReadReg(SPI, 24'h4, 8'h52); // CR3
    ReadReg(SPI, 24'h5, 8'h52); // CR4

    /* WRAR Test no WREN: DPI */
    $display("==== [TEST] WRAR test no WREN (DPI) ==== ");
    DPIEnable(SPI);
    WriteReg(DPI, 24'h0, 8'h52); // SR
    WriteReg(DPI, 24'h2, 8'h52); // CR1
    WriteReg(DPI, 24'h3, 8'h52); // CR2 
    WriteReg(DPI, 24'h4, 8'h52); // CR3
    WriteReg(DPI, 24'h5, 8'h52); // CR4
    ReadReg(DPI, 24'h0, 8'h52); // SR
    ReadReg(DPI, 24'h2, 8'h52); // CR1
    ReadReg(DPI, 24'h3, 8'h52); // CR2 
    ReadReg(DPI, 24'h4, 8'h52); // CR3
    ReadReg(DPI, 24'h5, 8'h52); // CR4

    /* WRAR Test no WREN: QPI */
    $display("==== [TEST] WRAR test no WREN (QPI) ==== ");
    QPIEnable(DPI);
    WriteReg(QPI, 24'h0, 8'h52); // SR
    WriteReg(QPI, 24'h2, 8'h52); // CR1
    WriteReg(QPI, 24'h3, 8'h52); // CR2 
    WriteReg(QPI, 24'h4, 8'h52); // CR3
    WriteReg(QPI, 24'h5, 8'h52); // CR4
    ReadReg(QPI, 24'h0, 8'h52); // SR
    ReadReg(QPI, 24'h2, 8'h52); // CR1
    ReadReg(QPI, 24'h3, 8'h52); // CR2 
    ReadReg(QPI, 24'h4, 8'h52); // CR3
    ReadReg(QPI, 24'h5, 8'h52); // CR4

    /* WRAR Test Read-only and Invalid address: SPI */
    $display("==== [TEST] WRAR test Read-only/Invalid addr (SPI) ==== ");
    SPIEnable(QPI);
    WriteEnable(SPI);
    WriteReg(SPI, 24'h30, 8'hff); // ID
    WriteEnable(SPI);
    WriteReg(SPI, 24'h40, 8'hff); // UID
    WriteEnable(SPI);
    WriteReg(SPI, 24'h123456, 8'hff); // Invalid reg addr

    /* WRAR Test Read-only and Invalid address: DPI */
    $display("==== [TEST] WRAR test Read-only/Invalid addr (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteReg(DPI, 24'h30, 8'hff); // ID
    WriteEnable(DPI);
    WriteReg(DPI, 24'h40, 8'hff); // UID
    WriteEnable(DPI);
    WriteReg(DPI, 24'h123456, 8'hff); // Invalid reg addr

    /* WRAR Test Read-only and Invalid address: QPI */
    $display("==== [TEST] WRAR test Read-only/Invalid addr (QPI) ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteReg(QPI, 24'h30, 8'hff); // ID
    WriteEnable(QPI);
    WriteReg(QPI, 24'h40, 8'hff); // UID
    WriteEnable(QPI);
    WriteReg(QPI, 24'h123456, 8'hff); // Invalid reg addr

    /* WRTE Test (SPI) */
    $display("==== [TEST] WRTE test (SPI) ==== ");
    SPIEnable(QPI);
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h00);
    WriteEnable(SPI);
    WriteSeq(SPI, 24'h123400, 8'hA0, 8'h1, 24'd20);
    ReadReg(SPI, 24'h0, 8'h00);
    ReadSeq(SPI, 24'h123400, 8'hA0, 8'h1, 8'h0, 24'd20);

    /* WRTE with WRENS=0b01 (SPI) */
    $display("==== [TEST] WRTE test: WRENS=0b01, WREN set (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h01);
    WriteEnable(SPI);
    WriteSeq(SPI, 24'h00abcd, 8'h70, 8'h2, 24'd30);
    ReadSeq(SPI, 24'h00abcd, 8'h70, 8'h2, 8'h0, 24'd30);
    $display("==== [TEST] WRTE test: WRENS=0b01, WREN cleared (SPI) ==== ");
    WriteDisable(SPI);
    WriteSeq(SPI, 24'h001234, 8'h39, 8'h2, 24'd30);
    ReadSeq(SPI, 24'h001234, 8'h39, 8'h2, 8'h0, 24'd30);

    /* WRTE with WRENS=0b10 (SPI) */
    $display("==== [TEST] WRTE test: WRENS=0b10, WREN set (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h02);
    WriteEnable(SPI);
    WriteSeq(SPI, 24'h00be00, 8'h70, 8'h1, 24'h10);
    WriteSeq(SPI, 24'h00be10, 8'h80, 8'h1, 24'h10);
    ReadSeq(SPI, 24'h00be00, 8'h70, 8'h1, 8'h0, 24'h20);
    $display("==== [TEST] WRTE test: WRENS=0b10, WREN cleared halfway (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h02);
    WriteEnable(SPI);
    WriteSeq(SPI, 24'h00be00, 8'h30, 8'h1, 24'h10);
    WriteDisable(SPI);
    WriteSeq(SPI, 24'h00be10, 8'h40, 8'h1, 24'h10);
    ReadSeq(SPI, 24'h00be00, 8'h30, 8'h1, 8'h0, 24'h20);

    /* WRFT Test (SPI) */
    $display("==== [TEST] WRFT/RDFT test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    //// Latency = 8
    //// Always allow array writes
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(SPI, 24'h6985, 8'ha1, 8'h5, 24'd15);
    ReadSeq(SPI, 24'h6985, 8'ha1, 8'h5, 8'h0, 24'd15);
    ReadReg(SPI, 24'h0, 8'h00); // Clear model address latch
    FastReadSeq(SPI, 24'h6985, 8'ha1, 8'h5, 8'h8, 24'd15);
    $display("==== [TEST] RDFT test: MLATS=0 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h01);
    FastReadSeq(SPI, 24'h6985, 8'ha1, 8'h5, 8'h0, 24'd15);
    $display("==== [TEST] RDFT test: MLATS=15 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h0F, 8'h00, 8'h01);
    FastReadSeq(SPI, 24'h6985, 8'ha1, 8'h5, 8'hF, 24'd15);
  
    /* WRFT Test (DPI) */
    $display("==== [TEST] WRFT/RDFT test: WRENS=SRAM(0b01), MLATS=8 (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    //// Latency = 8
    //// Always allow array writes
    WriteConfig(DPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(DPI, 24'ha1b2, 8'ha1, 8'h5, 24'd15);
    FastReadSeq(DPI, 24'ha1b2, 8'ha1, 8'h5, 8'h8, 24'd15);
    $display("==== [TEST] RDFT test: MLATS=0 (DPI) ==== ");
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h00, 8'h00, 8'h01);
    FastReadSeq(DPI, 24'ha1b2, 8'ha1, 8'h5, 8'h0, 24'd15);
    $display("==== [TEST] RDFT test: MLATS=15 (DPI) ==== ");
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h0F, 8'h00, 8'h01);
    FastReadSeq(DPI, 24'ha1b2, 8'ha1, 8'h5, 8'hF, 24'd15);
    
    /* WRFT Test (QPI) */
    $display("==== [TEST] WRFT/RDFT test: WRENS=SRAM(0b01), MLATS=8 (QPI) ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    //// Latency = 8
    //// Always allow array writes
    WriteConfig(QPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(QPI, 24'hd8b4, 8'ha1, 8'h5, 24'd15);
    ReadSeq(QPI, 24'hd8b4, 8'ha1, 8'h5, 8'h0, 24'd15);
    ReadReg(QPI, 24'h0, 8'h00); // Clear model address latch
    FastReadSeq(QPI, 24'hd8b4, 8'ha1, 8'h5, 8'h8, 24'd15);
    $display("==== [TEST] RDFT test: MLATS=0 (QPI) ==== ");
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h00, 8'h00, 8'h01);
    FastReadSeq(QPI, 24'hd8b4, 8'ha1, 8'h5, 8'h0, 24'd15);
    $display("==== [TEST] RDFT test: MLATS=15 (QPI) ==== ");
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h0F, 8'h00, 8'h01);
    FastReadSeq(QPI, 24'hd8b4, 8'ha1, 8'h5, 8'hF, 24'd15);

    /* RDFT XIP test (SPI) */
    $display("==== [TEST] RDFT XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    SPIEnable(QPI);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(SPI, 24'h100, 8'h70, 8'h1, 24'd10);
    FastReadSeqXIP(SPI, 24'h100, 8'h70, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    FastReadSeqXIP(SPI, 24'h105, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    FastReadSeqXIP(SPI, 24'h105, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    FastReadSeqXIP(SPI, 24'h100, 8'h70, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    FastReadSeqXIP(SPI, 24'h100, 8'h70, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);
    
    /* RDFT XIP test (DPI) */
    $display("==== [TEST] RDFT XIP test: WRENS=SRAM(0b01), MLATS=8 (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(DPI, 24'h200, 8'h80, 8'h1, 24'd10);
    FastReadSeqXIP(DPI, 24'h200, 8'h80, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    FastReadSeqXIP(DPI, 24'h205, 8'h85, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    FastReadSeqXIP(DPI, 24'h205, 8'h85, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    FastReadSeqXIP(DPI, 24'h200, 8'h80, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    FastReadSeqXIP(DPI, 24'h200, 8'h80, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);
    
    /* RDFT XIP test (QPI) */
    $display("==== [TEST] RDFT XIP test: WRENS=SRAM(0b01), MLATS=8 (QPI) ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(QPI, 24'h300, 8'h90, 8'h1, 24'd10);
    FastReadSeqXIP(QPI, 24'h300, 8'h90, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    FastReadSeqXIP(QPI, 24'h305, 8'h95, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    FastReadSeqXIP(QPI, 24'h305, 8'h95, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    FastReadSeqXIP(QPI, 24'h300, 8'h90, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    FastReadSeqXIP(QPI, 24'h300, 8'h90, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);
    
    /* WRFT XIP test (SPI) */
    $display("==== [TEST] WRFT XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    SPIEnable(QPI);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeqXIP(SPI, 24'h100, 8'hb0, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    FastWriteSeqXIP(SPI, 24'h105, 8'hb5, 8'h1, 24'd5, 1'b1, 8'hC9);
    FastWriteSeqXIP(SPI, 24'h10a, 8'hba, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(SPI, 24'h100, 8'hb0, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);
  
    /* WRFT XIP test (DPI) */
    $display("==== [TEST] WRFT XIP test: WRENS=SRAM(0b01), MLATS=8 (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeqXIP(DPI, 24'h100, 8'hb0, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    FastWriteSeqXIP(DPI, 24'h105, 8'hb5, 8'h1, 24'd5, 1'b1, 8'hC9);
    FastWriteSeqXIP(DPI, 24'h10a, 8'hba, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(DPI, 24'h100, 8'hb0, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);
  
    /* WRFT XIP test (QPI) */
    $display("==== [TEST] WRFT XIP test: WRENS=SRAM(0b01), MLATS=8 (QPI) ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeqXIP(QPI, 24'h100, 8'hb0, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    FastWriteSeqXIP(QPI, 24'h105, 8'hb5, 8'h1, 24'd5, 1'b1, 8'hC9);
    FastWriteSeqXIP(QPI, 24'h10a, 8'hba, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(QPI, 24'h100, 8'hb0, 8'h1, 8'h8, 24'd15, 1'b0, 8'h69);

    /* READ/WRTE DPI/QPI Test: Mode not supported */
    $display("==== [TEST] READ/WRTE DPI Test: Mode not supported ==== ");
    DPIEnable(QPI);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteSeq(DPI, 24'h500, 8'h10, 8'h1, 24'd5);
    ReadSeq(DPI, 24'h500, 8'h10, 8'h1, 8'h0, 24'd5);

    $display("==== [TEST] READ/WRTE QPI Test: Mode not supported ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteSeq(QPI, 24'h500, 8'h10, 8'h1, 24'd5);
    ReadSeq(QPI, 24'h500, 8'h10, 8'h1, 8'h0, 24'd5);

    /* WRAP Test (SPI) */
    $display("==== [TEST] WRAP Test (SPI) ==== ");
    SPIEnable(QPI);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h00); // ASPLK=0
    WriteEnable(SPI);
    WriteASP(SPI, 8'h5A);
    ReadASP(SPI, 8'h5A);
    $display("==== [TEST] WRAP Test: WREN cleared (SPI) ==== ");
    WriteASP(SPI, 8'hA5); // No WREN
    ReadASP(SPI, 8'hA5);
    $display("==== [TEST] WRAP Test: ASPLK set (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h01, 8'h00, 8'h00, 8'h00); // ASPLK=1
    WriteEnable(SPI);
    WriteASP(SPI, 8'hFF);
    ReadASP(SPI, 8'hFF);
  
    /* WRAP Test (DPI) */
    $display("==== [TEST] WRAP Test (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h00, 8'h00, 8'h00); // ASPLK=0
    WriteEnable(DPI);
    WriteASP(DPI, 8'h6A);
    ReadASP(DPI, 8'h6A);
    $display("==== [TEST] WRAP Test: WREN cleared (DPI) ==== ");
    WriteASP(DPI, 8'hA6); // No WREN
    ReadASP(DPI, 8'hA6);
    $display("==== [TEST] WRAP Test: ASPLK set (DPI) ==== ");
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h01, 8'h00, 8'h00, 8'h00); // ASPLK=1
    WriteEnable(DPI);
    WriteASP(DPI, 8'hFF);
    ReadASP(DPI, 8'hFF);
  
    /* WRAP Test (QPI) */
    $display("==== [TEST] WRAP Test (QPI) ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h00, 8'h00, 8'h00); // ASPLK=0
    WriteEnable(QPI);
    WriteASP(QPI, 8'h7A);
    ReadASP(QPI, 8'h7A);
    $display("==== [TEST] WRAP Test: WREN cleared (QPI) ==== ");
    WriteASP(QPI, 8'hA7); // No WREN
    ReadASP(QPI, 8'hA7);
    $display("==== [TEST] WRAP Test: ASPLK set (QPI) ==== ");
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h01, 8'h00, 8'h00, 8'h00); // ASPLK=1
    WriteEnable(QPI);
    WriteASP(QPI, 8'hFF);
    ReadASP(QPI, 8'hFF);

    /* WRAS/RDAS Test: Successful write */
    $display("==== [TEST] WRAS/RDAS ==== ");
    SPIEnable(QPI);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00); 
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h10, 8'hA0, 8'h1, 24'h20);
    WriteEnable(SPI);
    WriteSeq(SPI, 24'h10, 8'hF0, 8'h7, 24'h20);
    ReadAugSeq(SPI, 24'h10, 8'hA0, 8'h1, 8'h8, 24'h10);
    ReadAugSeq(SPI, 24'h20, 8'hB0, 8'h1, 8'h8, 24'h10);
    ReadSeq(SPI, 24'h10, 8'hF0, 8'h7, 8'h0, 24'h20);

    /* WRAS/RDAS Test: No WREN */
    $display("==== [TEST] WRAS/RDAS: No WREN ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00); 
    WriteAugSeq(SPI, 24'h10, 8'h11, 8'h1, 24'h1);
    ReadAugSeq(SPI, 24'h10, 8'h11, 8'h1, 8'h8, 24'h1);

    /* WRAS/RDAS Test: ASP[0] */
    $display("==== [TEST] WRAS/RDAS Protect Section 0 ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00);
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h10, 8'h00, 8'h0, 24'h1); // Init to 0
    WriteEnable(SPI);
    WriteASP(SPI, 8'h01); 
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h10, 8'h66, 8'h1, 24'h1);
    ReadAugSeq(SPI, 24'h10, 8'h66, 8'h1, 8'h8, 24'h1);

    /* WRAS/RDAS Test: ASP[1] */
    $display("==== [TEST] WRAS/RDAS Protect Section 1 ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00);
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h30, 8'h00, 8'h0, 24'h1); // Init to 0
    WriteEnable(SPI);
    WriteASP(SPI, 8'h02); 
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h30, 8'h66, 8'h1, 24'h1);
    ReadAugSeq(SPI, 24'h30, 8'h66, 8'h1, 8'h8, 24'h1);

    /* WRAS/RDAS Test: ASP[2] */
    $display("==== [TEST] WRAS/RDAS Protect Section 2 ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00);
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h50, 8'h00, 8'h0, 24'h1); // Init to 0
    WriteEnable(SPI);
    WriteASP(SPI, 8'h04); 
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h50, 8'h66, 8'h1, 24'h1);
    ReadAugSeq(SPI, 24'h50, 8'h66, 8'h1, 8'h8, 24'h1);

    /* WRAS/RDAS Test: ASP[3] */
    $display("==== [TEST] WRAS/RDAS Protect Section 3 ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00);
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h70, 8'h00, 8'h0, 24'h1); // Init to 0
    WriteEnable(SPI);
    WriteASP(SPI, 8'h08); 
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h70, 8'h66, 8'h1, 24'h1);
    ReadAugSeq(SPI, 24'h70, 8'h66, 8'h1, 8'h8, 24'h1);

    /* WRAS/RDAS Test: ASP[4] */
    $display("==== [TEST] WRAS/RDAS Protect Section 4 ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00);
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h8F, 8'h00, 8'h0, 24'h1); // Init to 0
    WriteEnable(SPI);
    WriteASP(SPI, 8'h10); 
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h8F, 8'h66, 8'h1, 24'h1);
    ReadAugSeq(SPI, 24'h8F, 8'h66, 8'h1, 8'h8, 24'h1);

    /* WRAS/RDAS Test: ASP[5] */
    $display("==== [TEST] WRAS/RDAS Protect Section 5 ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00);
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'hAF, 8'h00, 8'h0, 24'h1); // Init to 0
    WriteEnable(SPI);
    WriteASP(SPI, 8'h20); 
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'hAF, 8'h66, 8'h1, 24'h1);
    ReadAugSeq(SPI, 24'hAF, 8'h66, 8'h1, 8'h8, 24'h1);

    /* WRAS/RDAS Test: ASP[6] */
    $display("==== [TEST] WRAS/RDAS Protect Section 6 ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00);
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'hCF, 8'h00, 8'h0, 24'h1); // Init to 0
    WriteEnable(SPI);
    WriteASP(SPI, 8'h40); 
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'hCF, 8'h66, 8'h1, 24'h1);
    ReadAugSeq(SPI, 24'hCF, 8'h66, 8'h1, 8'h8, 24'h1);

    /* WRAS/RDAS Test: ASP[7] */
    $display("==== [TEST] WRAS/RDAS Protect Section 7 ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00);
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'hEF, 8'h00, 8'h0, 24'h1); // Init to 0
    WriteEnable(SPI);
    WriteASP(SPI, 8'h80); 
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'hEF, 8'h66, 8'h1, 24'h1);
    ReadAugSeq(SPI, 24'hEF, 8'h66, 8'h1, 8'h8, 24'h1);

    /* WRAS/RDAS Test: Address out of range */
    $display("==== [TEST] WRAS/RDAS Address out of range ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00); // ASPLK=0;
    WriteEnable(SPI);
    WriteASP(SPI, 8'h00);
    WriteEnable(SPI);
    WriteAugSeq(SPI, 24'h12EF, 8'hFF, 8'h0, 24'h1);
    ReadAugSeq(SPI, 24'h12EF, 8'hFF, 8'h0, 8'h8, 24'h1);
  
    /* WRSN/RDSN Test: SNPEN disabled */
    $display("==== [TEST] WRSN/RDSN Test: SNPEN disabled (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h00);
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteSN(SPI, 64'h0123456789abcdef);
    ReadSN(SPI, 64'h0123456789abcdef);
    
    /* WRSN/RDSN Test: SNPEN enabled */
    $display("==== [TEST] WRSN/RDSN Test: SNPEN enabled (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h00);
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h40);
    WriteEnable(SPI);
    WriteSN(SPI, 64'hff23456789abcdef);
    ReadSN(SPI, 64'hff23456789abcdef);
    
    /* WRSN/RDSN Test: WREN not set */
    $display("==== [TEST] WRSN/RDSN Test: WREN not set (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h00);
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteSN(SPI, 64'hffff456789abcdef);
    ReadSN(SPI, 64'hffff456789abcdef);
    
    /* WRSN/RDSN Test: SNPEN disabled */
    $display("==== [TEST] WRSN/RDSN Test: SNPEN disabled (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h00, 8'h00, 8'h00);
    WriteEnable(DPI);
    WriteStatus(DPI, 8'h00);
    WriteEnable(DPI);
    WriteSN(DPI, 64'habcdef0123456789);
    ReadSN(DPI, 64'habcdef0123456789);
    
    /* WRSN/RDSN Test: SNPEN enabled */
    $display("==== [TEST] WRSN/RDSN Test: SNPEN enabled (DPI) ==== ");
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h00, 8'h00, 8'h00);
    WriteEnable(DPI);
    WriteStatus(DPI, 8'h40);
    WriteEnable(DPI);
    WriteSN(DPI, 64'hff23456789abcdef);
    ReadSN(DPI, 64'hff23456789abcdef);
    
    /* WRSN/RDSN Test: WREN not set */
    $display("==== [TEST] WRSN/RDSN Test: WREN not set (DPI) ==== ");
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h00, 8'h00, 8'h00);
    WriteEnable(DPI);
    WriteStatus(DPI, 8'h00);
    WriteSN(DPI, 64'hffff456789abcdef);
    ReadSN(DPI, 64'hffff456789abcdef);
    
    /* WRSN/RDSN Test: SNPEN disabled */
    $display("==== [TEST] WRSN/RDSN Test: SNPEN disabled (QPI) ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h00, 8'h00, 8'h00);
    WriteEnable(QPI);
    WriteStatus(QPI, 8'h00);
    WriteEnable(QPI);
    WriteSN(QPI, 64'h0123456789abcdef);
    ReadSN(QPI, 64'h0123456789abcdef);
    
    /* WRSN/RDSN Test: SNPEN enabled */
    $display("==== [TEST] WRSN/RDSN Test: SNPEN enabled (QPI) ==== ");
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h00, 8'h00, 8'h00);
    WriteEnable(QPI);
    WriteStatus(QPI, 8'h40);
    WriteEnable(QPI);
    WriteSN(QPI, 64'hff23456789abcdef);
    ReadSN(QPI, 64'hff23456789abcdef);
    
    /* WRSN/RDSN Test: WREN not set */
    $display("==== [TEST] WRSN/RDSN Test: WREN not set (QPI) ==== ");
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h00, 8'h00, 8'h00);
    WriteEnable(QPI);
    WriteStatus(QPI, 8'h00);
    WriteSN(QPI, 64'hffff456789abcdef);
    ReadSN(QPI, 64'hffff456789abcdef);

    /* HBNE Test (SPI) */
    $display("==== [TEST] HBNE (SPI) ==== ");
    SPIEnable(QPI);
    EnterHibernate(SPI);
    #(10000); // Wait to hibernate
    SelectChip();
    #(10000); // Wait to wake
    DeSelChip();

    /* HBNE Test (DPI) */
    $display("==== [TEST] HBNE (DPI) ==== ");
    DPIEnable(SPI);
    EnterHibernate(DPI);
    #(10000); // Wait to hibernate
    SelectChip();
    #(10000); // Wait to wake
    DeSelChip();
    
    /* HBNE Test (QPI) */
    $display("==== [TEST] HBNE (QPI) ==== ");
    QPIEnable(DPI);
    EnterHibernate(QPI);
    #(10000); // Wait to hibernate
    SelectChip();
    #(10000); // Wait to wake
    DeSelChip();

    /* DPDE, CS wake Test (SPI) */
    $display("==== [TEST] DPDE, CS wake (SPI) ====");
    SPIEnable(QPI);
    EnterDPD(SPI);
    #(10000); // Wait to DPD
    SelectChip();
    #(500); // Wake trigger 
    DeSelChip();

    /* DPDE, CS wake Test (DPI) */
    $display("==== [TEST] DPDE, CS wake (DPI) ====");
    DPIEnable(SPI);
    EnterDPD(DPI);
    #(10000); // Wait to DPD
    SelectChip();
    #(500); // Wake trigger 
    DeSelChip();

    /* DPDE, CS wake Test (QPI) */
    $display("==== [TEST] DPDE, CS wake (QPI) ====");
    QPIEnable(DPI);
    EnterDPD(QPI);
    #(10000); // Wait to DPD
    SelectChip();
    #(500); // Wake trigger 
    DeSelChip();

    /* DPDX Test (SPI) */
    $display("==== [TEST] DPDX (SPI) ==== ");
    SPIEnable(QPI);
    EnterDPD(SPI);
    #(10000); // Wait to DPD
    ExitDPD(SPI);
    #(10000); // Wait to Wake

    /* DPDX Test (DPI) */
    $display("==== [TEST] DPDX (DPI) ==== ");
    DPIEnable(SPI);
    EnterDPD(DPI);
    #(10000); // Wait to DPD
    ExitDPD(DPI);
    #(10000); // Wait to Wake

    /* DPDX Test (QPI) */
    $display("==== [TEST] DPDX (QPI) ==== ");
    QPIEnable(DPI);
    EnterDPD(QPI);
    #(10000); // Wait to DPD
    ExitDPD(QPI);
    #(10000); // Wait to Wake

    /* DPDX during normal mode (SPI) */
    $display("==== [TEST] DPDX during normal mode (SPI) ==== ");
    SPIEnable(QPI);
    ExitDPD(SPI);

    /* DPDX during normal mode (DPI) */
    $display("==== [TEST] DPDX during normal mode (DPI) ==== ");
    DPIEnable(SPI);
    ExitDPD(DPI);

    /* DPDX during normal mode (QPI) */
    $display("==== [TEST] DPDX during normal mode (QPI) ==== ");
    QPIEnable(DPI);
    ExitDPD(QPI);

    /* Deep power down invalid command Test (SPI) */
    $display("==== [TEST] Deep Power Down Invalid Command (SPI) ==== ");
    SPIEnable(QPI);
    EnterDPD(SPI);
    #(10000);
    WriteEnable(SPI); // will cause this to exit due to CSn width
    #(10000);
    ExitDPD(SPI);

    /* Deep power down invalid command Test (DPI) */
    $display("==== [TEST] Deep Power Down Invalid Command (DPI) ==== ");
    DPIEnable(SPI);
    EnterDPD(DPI);
    #(10000);
    WriteEnable(DPI);
    #(10000);
    ExitDPD(DPI);

    /* Deep power down invalid command Test (QPI) */
    $display("==== [TEST] Deep Power Down Invalid Command (QPI) ==== ");
    QPIEnable(DPI);
    EnterDPD(QPI);
    #(10000);
    WriteEnable(QPI);
    #(10000);
    ExitDPD(QPI);

    /* Software Reset (SPI) */
    $display("==== [TEST] Software Reset (SPI) ==== ");
    SPIEnable(QPI);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'hFF, 8'hFF, 8'hFF, 8'hFF);
    SoftwareReset(SPI);
    SoftwareResetEnable(SPI);
    SoftwareReset(SPI);
    #(10000);
   
    /* Software Reset (DPI) */
    $display("==== [TEST] Software Reset (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'hFF, 8'hFF, 8'hFF, 8'hFF);
    SoftwareReset(DPI);
    SoftwareResetEnable(DPI);
    SoftwareReset(DPI);
    #(10000);
   
    /* Software Reset (QPI) */
    $display("==== [TEST] Software Reset (QPI) ==== ");
    QPIEnable(SPI);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'hFF, 8'hFF, 8'hFF, 8'hFF);
    SoftwareReset(QPI);
    SoftwareResetEnable(QPI);
    SoftwareReset(QPI);
    #(10000);

    /* DPD CSn less than 50ns */
    $display("==== [TEST] DPD CSn toggle <50ns ==== ");
    EnterDPD(SPI);
    #(10000)
    CSn = 0;
    #(400) CSn = 1'b1;
    #(10000);
    CSn = 0;
    #(500) CSn = 1'b1;
    #(10000);

    /* RDC1 to RDC4 test (SPI) */
    $display("==== [TEST] RDC1 to RDC4 Test (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h1B, 8'h1B, 8'h1B, 8'h1B);
    ReadCR1(SPI, 8'h1B);
    ReadCR2(SPI, 8'h1B);
    ReadCR3(SPI, 8'h1B);
    ReadCR4(SPI, 8'h1B);

    /* RDC1 to RDC4 test (DPI) */
    $display("==== [TEST] RDC1 to RDC4 Test (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h2C, 8'h2C, 8'h2C, 8'h2C);
    ReadCR1(DPI, 8'h2C);
    ReadCR2(DPI, 8'h2C);
    ReadCR3(DPI, 8'h2C);
    ReadCR4(DPI, 8'h2C);

    /* RDC1 to RDC4 test (QPI) */
    $display("==== [TEST] RDC1 to RDC4 Test (QPI) ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h3D, 8'h3D, 8'h3D, 8'h3D);
    ReadCR1(QPI, 8'h3D);
    ReadCR2(QPI, 8'h3D);
    ReadCR3(QPI, 8'h3D);
    ReadCR4(QPI, 8'h3D);

    /* RDID (SPI) */
    $display("==== [TEST] RDID, RUID (SPI) ==== ");
    SPIEnable(QPI);
    ReadID(SPI); 
    ReadUID(SPI);

    /* RDID (DPI) */
    $display("==== [TEST] RDID, RUID (DPI) ==== ");
    DPIEnable(SPI);
    ReadID(DPI); 
    ReadUID(DPI);

    /* RDID (QPI) */
    $display("==== [TEST] RDID, RUID (QPI) ==== ");
    QPIEnable(DPI);
    ReadID(QPI); 
    ReadUID(QPI);

    $finish;
  end

endmodule
