`timescale 100ps / 10ps

/*
 * Options for modeling byte transmission (uncomment at most 1)
 * - SUHDCHK_ZHD
 *   - Line is released (data=HiZ) after hold time
 * - SUHDCK_CKFE
 *   - Data is set during falling edge of CLK, line released at falling edge
 *     of CLK
 * - default (no option uncommented)
 *   - Data is set to 0 after hold time, line released (tCL-tSU) after
 *     last falling edge of CLK
 *
 */
//`define SUHDCHK_ZHD
`define SUHDCHK_CKFE

module stimulus 
	(
		SO_IO1,
		CLK,
		IO3, 
		SI_IO0 ,
		WPn_IO2,
		CSn
	);

  inout SO_IO1;
	inout IO3;
	inout SI_IO0;
	inout WPn_IO2;
	
	output CSn;
	output CLK;
        
	reg CSn = 1;
	reg CLK = 0;
        
	reg SI_IO0_reg=0;
  reg SO_IO1_reg=0;
  reg WPn_IO2_reg=0;
  reg IO3_reg=0;
        
  reg [3:0] tx;

  parameter SPI = 3'h01;
  parameter DPI = 3'h02;
  parameter QPI = 3'h04;

  `include "../shr/ref/task.v"
	 
  assign SI_IO0 = tx[0] ? SI_IO0_reg : 1'bz;  
  assign SO_IO1 = tx[1] ? SO_IO1_reg : 1'bz;  
  assign WPn_IO2 = tx[2] ? WPn_IO2_reg : 1'bz;  
  assign IO3 = tx[3] ? IO3_reg : 1'bz;  

  /* Expected ERROR: 20 */
  initial begin
    tx = 4'd0;
    #10000;

    $display("==== [TEST] DDR Fast READ SPI ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteSeq(SPI, 24'ha1b2, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    DDRFastReadSeq(SPI, 24'ha1b2, 8'hc0, 8'h3, 8'h08, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    //FastReadSeq(SPI, 24'ha1b2, 8'hc0, 8'h3, 8'h08, 24'd10);

    $display("==== [TEST] DDR Fast READ SPI, Latency 0 ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h01);
    WriteSeq(SPI, 24'hb1b2, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    DDRFastReadSeq(SPI, 24'hb1b2, 8'hc0, 8'h3, 8'h00, 24'd10);

    $display("==== [TEST] DDR Fast READ 32-byte Wrap (SPI) ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h11, 8'h01); // 32-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'hcd00, 8'h10, 8'h1, 8'd40);
    DDRFastReadSeq(SPI, 24'hcd10, 8'h20, 8'h1, 8'h08, 8'd20);

    $display("==== [TEST] DRFR XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(SPI, 24'h100, 8'h70, 8'h1, 24'd10);
    DDRFastReadSeqXIP(SPI, 24'h100, 8'h70, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    DDRFastReadSeqXIP(SPI, 24'h105, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    DDRFastReadSeqXIP(SPI, 24'h105, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    DDRFastReadSeqXIP(SPI, 24'h100, 8'h70, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    DDRFastReadSeqXIP(SPI, 24'h100, 8'h70, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);
    
    $display("==== [TEST] DDR Fast READ DPI ====");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteStatus(DPI, 8'h00);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(DPI, 24'ha1b2, 8'hd0, 8'h3, 24'd10);
    ReadReg(DPI, 24'h0, 8'h00); //clear addr latch
    DDRFastReadSeq(DPI, 24'ha1b2, 8'hd0, 8'h3, 8'h08, 24'd10);
    ReadReg(DPI, 24'h0, 8'h00); //clear addr latch
    FastReadSeq(DPI, 24'ha1b2, 8'hd0, 8'h3, 8'h08, 24'd10);

    $display("==== [TEST] DDR Fast READ DPI, Latency 0 ====");
    WriteEnable(DPI);
    WriteStatus(DPI, 8'h00);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h00, 8'h00, 8'h01);
    FastWriteSeq(DPI, 24'hb1b2, 8'hd0, 8'h3, 24'd10);
    ReadReg(DPI, 24'h0, 8'h00); //clear addr latch
    DDRFastReadSeq(DPI, 24'hb1b2, 8'hd0, 8'h3, 8'h00, 24'd10);

    $display("==== [TEST] DDR Fast READ 32-byte Wrap (DPI) ==== ");
    WriteEnable(DPI);
    WriteStatus(DPI, 8'h00);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h08, 8'h11, 8'h01); // 32-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(DPI, 24'h6900, 8'h10, 8'h1, 8'd40);
    DDRFastReadSeq(DPI, 24'h6910, 8'h20, 8'h1, 8'h08, 8'd20);

    $display("==== [TEST] DRFR XIP test: WRENS=SRAM(0b01), MLATS=8 (DPI) ==== ");
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(DPI, 24'h100, 8'h90, 8'h1, 24'd10);
    DDRFastReadSeqXIP(DPI, 24'h100, 8'h90, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    DDRFastReadSeqXIP(DPI, 24'h105, 8'h95, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    DDRFastReadSeqXIP(DPI, 24'h105, 8'h95, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    DDRFastReadSeqXIP(DPI, 24'h100, 8'h90, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    DDRFastReadSeqXIP(DPI, 24'h100, 8'h90, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);
    
    $display("==== [TEST] DDR Fast READ QPI ====");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteStatus(QPI, 8'h00);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(QPI, 24'ha145, 8'hd0, 8'h3, 24'd10);
    ReadReg(QPI, 24'h0, 8'h00); //clear addr latch
    DDRFastReadSeq(QPI, 24'ha145, 8'hd0, 8'h3, 8'h08, 24'd10);
    ReadReg(QPI, 24'h0, 8'h00); //clear addr latch
    FastReadSeq(QPI, 24'ha145, 8'hd0, 8'h3, 8'h08, 24'd10);

    $display("==== [TEST] DDR Fast READ QPI, Latency 0 ====");
    WriteEnable(QPI);
    WriteStatus(QPI, 8'h00);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h00, 8'h00, 8'h01);
    FastWriteSeq(QPI, 24'hb145, 8'hd0, 8'h3, 24'd10);
    ReadReg(QPI, 24'h0, 8'h00); //clear addr latch
    DDRFastReadSeq(QPI, 24'hb145, 8'hd0, 8'h3, 8'h00, 24'd10);

    $display("==== [TEST] DDR Fast READ 32-byte Wrap (QPI) ==== ");
    WriteEnable(QPI);
    WriteStatus(QPI, 8'h00);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h08, 8'h11, 8'h01); // 32-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(QPI, 24'hd900, 8'h10, 8'h1, 8'd40);
    DDRFastReadSeq(QPI, 24'hd910, 8'h20, 8'h1, 8'h08, 8'd20);

    $display("==== [TEST] DRFR XIP test: WRENS=SRAM(0b01), MLATS=8 (QPI) ==== ");
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(QPI, 24'h200, 8'h90, 8'h1, 24'd10);
    DDRFastReadSeqXIP(QPI, 24'h200, 8'h90, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    DDRFastReadSeqXIP(QPI, 24'h205, 8'h95, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    DDRFastReadSeqXIP(QPI, 24'h205, 8'h95, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    DDRFastReadSeqXIP(QPI, 24'h200, 8'h90, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    DDRFastReadSeqXIP(QPI, 24'h200, 8'h90, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);

    $display("==== [TEST] DDR Fast Write SPI ====");
    SPIEnable(QPI);
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    DDRFastWriteSeq(SPI, 24'h6600, 8'h30, 8'h5, 24'd10);
    FastReadSeq(SPI, 24'h6600, 8'h30, 8'h5, 8'h08, 24'd10);
    
    $display("==== [TEST] DDR Fast Write SPI no WREN ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00);
    DDRFastWriteSeq(SPI, 24'h6600, 8'h40, 8'h5, 24'd10);
    //FastReadSeq(SPI, 24'h6600, 8'h40, 8'h5, 8'h08, 24'd10);
    
    $display("==== [TEST] DRFW XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    DDRFastWriteSeqXIP(SPI, 24'h100, 8'hb0, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    DDRFastWriteSeqXIP(SPI, 24'h105, 8'hb5, 8'h1, 24'd5, 1'b1, 8'hC9);
    DDRFastWriteSeqXIP(SPI, 24'h10a, 8'hba, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(SPI, 24'h100, 8'hb0, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);
  
    $display("==== [TEST] DDR Fast Write DPI ====");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteStatus(DPI, 8'h00);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h08, 8'h00, 8'h01);
    DDRFastWriteSeq(DPI, 24'h7700, 8'h30, 8'h5, 24'd10);
    FastReadSeq(DPI, 24'h7700, 8'h30, 8'h5, 8'h08, 24'd10);
    
    $display("==== [TEST] DRFW XIP test: WRENS=SRAM(0b01), MLATS=8 (DPI) ==== ");
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h08, 8'h00, 8'h01);
    DDRFastWriteSeqXIP(DPI, 24'h200, 8'hb0, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    DDRFastWriteSeqXIP(DPI, 24'h205, 8'hb5, 8'h1, 24'd5, 1'b1, 8'hC9);
    DDRFastWriteSeqXIP(DPI, 24'h20a, 8'hba, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(DPI, 24'h200, 8'hb0, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);
  
    $display("==== [TEST] DDR Fast Write QPI ====");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteStatus(QPI, 8'h00);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h08, 8'h00, 8'h01);
    DDRFastWriteSeq(QPI, 24'h8800, 8'h30, 8'h5, 24'd10);
    FastReadSeq(QPI, 24'h8800, 8'h30, 8'h5, 8'h08, 24'd10);
    
    $display("==== [TEST] DRFW XIP test: WRENS=SRAM(0b01), MLATS=8 (QPI) ==== ");
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h08, 8'h00, 8'h01);
    DDRFastWriteSeqXIP(QPI, 24'h300, 8'hb0, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    DDRFastWriteSeqXIP(QPI, 24'h305, 8'hb5, 8'h1, 24'd5, 1'b1, 8'hC9);
    DDRFastWriteSeqXIP(QPI, 24'h30a, 8'hba, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(QPI, 24'h300, 8'hb0, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);
  
    $display("==== [TEST] DRDI ====");
    SPIEnable(QPI);
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteSeq(SPI, 24'ha1b2, 8'hc5, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    DDRReadDualIOSeqXIP(SPI, 24'ha1b2, 8'hc5, 8'h3, 8'h08, 24'd10, 1'b0, 8'hff);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    FastReadSeq(SPI, 24'ha1b2, 8'hc5, 8'h3, 8'h08, 24'd10);
  
    $display("==== [TEST] DRDI, Latency 0 ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h01);
    WriteSeq(SPI, 24'hb1b2, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    DDRReadDualIOSeqXIP(SPI, 24'hb1b2, 8'hc0, 8'h3, 8'h00, 24'd10, 1'b0, 8'hff);

    $display("==== [TEST] DRDI 32-byte Wrap ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h11, 8'h01); // 32-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'h3f00, 8'h10, 8'h1, 8'd40);
    DDRReadDualIOSeqXIP(SPI, 24'h3f10, 8'h20, 8'h1, 8'h08, 8'd20, 1'b0, 8'hff);

    $display("==== [TEST] DRDI XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(SPI, 24'h100, 8'he0, 8'h1, 24'd10);
    DDRReadDualIOSeqXIP(SPI, 24'h100, 8'he0, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    DDRReadDualIOSeqXIP(SPI, 24'h105, 8'he5, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    DDRReadDualIOSeqXIP(SPI, 24'h105, 8'he5, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    DDRReadDualIOSeqXIP(SPI, 24'h100, 8'he0, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    DDRReadDualIOSeqXIP(SPI, 24'h100, 8'he0, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);
    
    $display("==== [TEST] DRQI ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteSeq(SPI, 24'hc8b2, 8'hc5, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    DDRReadQuadIOSeqXIP(SPI, 24'hc8b2, 8'hc5, 8'h3, 8'h08, 24'd10, 1'b0, 8'hff);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    FastReadSeq(SPI, 24'hc8b2, 8'hc5, 8'h3, 8'h08, 24'd10);
  
    $display("==== [TEST] DRQI, Latency 0 ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h01);
    WriteSeq(SPI, 24'h00b2, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    DDRReadQuadIOSeqXIP(SPI, 24'h00b2, 8'hc0, 8'h3, 8'h00, 24'd10, 1'b0, 8'hff);

    $display("==== [TEST] DRQI 32-byte Wrap ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h11, 8'h01); // 32-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'h4800, 8'h10, 8'h1, 8'd40);
    DDRReadQuadIOSeqXIP(SPI, 24'h4810, 8'h20, 8'h1, 8'h08, 8'd20, 1'b0, 8'hff);

    $display("==== [TEST] DRQI XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(SPI, 24'h1100, 8'he0, 8'h1, 24'd10);
    DDRReadQuadIOSeqXIP(SPI, 24'h1100, 8'he0, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    DDRReadQuadIOSeqXIP(SPI, 24'h1105, 8'he5, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    DDRReadQuadIOSeqXIP(SPI, 24'h1105, 8'he5, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    DDRReadQuadIOSeqXIP(SPI, 24'h1100, 8'he0, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    DDRReadQuadIOSeqXIP(SPI, 24'h1100, 8'he0, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);
    
    $display("==== [TEST] DWQI ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    DDRWriteQuadInputSeqXIP(SPI, 24'h7600, 8'h30, 8'h5, 24'd10, 1'b0, 8'hFF);
    FastReadSeq(SPI, 24'h7600, 8'h30, 8'h5, 8'h08, 24'd10);
    
    $display("==== [TEST] DWQI no WREN ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00);
    DDRWriteQuadInputSeqXIP(SPI, 24'h7600, 8'h50, 8'h5, 24'd10, 1'b0, 8'hFF);
    
    $display("==== [TEST] DWQI XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    DDRWriteQuadInputSeqXIP(SPI, 24'h100, 8'h40, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    DDRWriteQuadInputSeqXIP(SPI, 24'h105, 8'h45, 8'h1, 24'd5, 1'b1, 8'hC9);
    DDRWriteQuadInputSeqXIP(SPI, 24'h10a, 8'h4a, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(SPI, 24'h100, 8'h40, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);
  
    $display("==== [TEST] DWQO ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    DDRWriteQuadIOSeqXIP(SPI, 24'h9a00, 8'h30, 8'h7, 24'd10, 1'b0, 8'hFF);
    FastReadSeq(SPI, 24'h9a00, 8'h30, 8'h7, 8'h08, 24'd10);
    
    $display("==== [TEST] DWQO no WREN ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00);
    DDRWriteQuadIOSeqXIP(SPI, 24'ha900, 8'h50, 8'h5, 24'd10, 1'b0, 8'hFF);
    
    $display("==== [TEST] DWQO XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    DDRWriteQuadIOSeqXIP(SPI, 24'h700, 8'hc0, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    DDRWriteQuadIOSeqXIP(SPI, 24'h705, 8'hc5, 8'h1, 24'd5, 1'b1, 8'hC9);
    DDRWriteQuadIOSeqXIP(SPI, 24'h70a, 8'hca, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(SPI, 24'h700, 8'hc0, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);
  
    $finish;

  end
endmodule
