`timescale 100ps / 10ps

/*
 * Options for modeling byte transmission (uncomment at most 1)
 * - SUHDCHK_ZHD
 *   - Line is released (data=HiZ) after hold time
 * - SUHDCK_CKFE
 *   - Data is set during falling edge of CLK, line released at falling edge
 *     of CLK
 * - default (no option uncommented)
 *   - Data is set to 0 after hold time, line released (tCL-tSU) after
 *     last falling edge of CLK
 *
 */
//`define SUHDCHK_ZHD
`define SUHDCHK_CKFE

module stimulus 
	(
		SO_IO1,
		CLK,
		IO3, 
		SI_IO0 ,
		WPn_IO2,
		CSn
	);

  inout SO_IO1;
	inout IO3;
	inout SI_IO0;
	inout WPn_IO2;
	
	output CSn;
	output CLK;
        
	reg CSn = 1;
	reg CLK = 0;
        
	reg SI_IO0_reg=0;
  reg SO_IO1_reg=0;
  reg WPn_IO2_reg=0;
  reg IO3_reg=0;
        
  reg [3:0] tx;

  parameter SPI = 3'h01;
  parameter DPI = 3'h02;
  parameter QPI = 3'h04;

  `include "../shr/ref/task.v"
	 
  assign SI_IO0 = tx[0] ? SI_IO0_reg : 1'bz;  
  assign SO_IO1 = tx[1] ? SO_IO1_reg : 1'bz;  
  assign WPn_IO2 = tx[2] ? WPn_IO2_reg : 1'bz;  
  assign IO3 = tx[3] ? IO3_reg : 1'bz;  

  /* Expected ERROR: 16 */
  initial begin
    tx = 4'd0;
    #10000;

    $display("==== [TEST] RDDO ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteSeq(SPI, 24'ha1b2, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    ReadDualOutputSeqXIP(SPI, 24'ha1b2, 8'hc0, 8'h3, 8'h08, 24'd10, 1'b0, 8'hff);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    FastReadSeq(SPI, 24'ha1b2, 8'hc0, 8'h3, 8'h08, 24'd10);
  
    $display("==== [TEST] RDDO, Latency 0 ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h01);
    WriteSeq(SPI, 24'hb1b2, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    ReadDualOutputSeqXIP(SPI, 24'hb1b2, 8'hc0, 8'h3, 8'h00, 24'd10, 1'b0, 8'hff);

    $display("==== [TEST] RDDO 32-byte Wrap ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h11, 8'h01); // 32-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'hcd00, 8'h10, 8'h1, 8'd40);
    ReadDualOutputSeqXIP(SPI, 24'hcd10, 8'h20, 8'h1, 8'h08, 8'd20, 1'b0, 8'hff);

    $display("==== [TEST] RDDO XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(SPI, 24'h100, 8'h70, 8'h1, 24'd10);
    ReadDualOutputSeqXIP(SPI, 24'h100, 8'h70, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    ReadDualOutputSeqXIP(SPI, 24'h105, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    ReadDualOutputSeqXIP(SPI, 24'h105, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    ReadDualOutputSeqXIP(SPI, 24'h100, 8'h70, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    ReadDualOutputSeqXIP(SPI, 24'h100, 8'h70, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);
    
    $display("==== [TEST] RDQO ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteSeq(SPI, 24'hc3d4, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    ReadQuadOutputSeqXIP(SPI, 24'hc3d4, 8'hc0, 8'h3, 8'h08, 24'd10, 1'b0, 8'hff);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    FastReadSeq(SPI, 24'hc3d4, 8'hc0, 8'h3, 8'h08, 24'd10);

    $display("==== [TEST] RDQO, Latency 0 ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h01);
    WriteSeq(SPI, 24'hb001, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    ReadQuadOutputSeqXIP(SPI, 24'hb001, 8'hc0, 8'h3, 8'h00, 24'd10, 1'b0, 8'hff);

    $display("==== [TEST] RDQO 32-byte Wrap ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h11, 8'h01); // 32-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'h1300, 8'h10, 8'h1, 8'd40);
    ReadQuadOutputSeqXIP(SPI, 24'h1310, 8'h20, 8'h1, 8'h08, 8'd20, 1'b0, 8'hff);

    $display("==== [TEST] RDQO XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(SPI, 24'hc00, 8'h70, 8'h1, 24'd10);
    ReadQuadOutputSeqXIP(SPI, 24'hc00, 8'h70, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    ReadQuadOutputSeqXIP(SPI, 24'hc05, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    ReadQuadOutputSeqXIP(SPI, 24'hc05, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    ReadQuadOutputSeqXIP(SPI, 24'hc00, 8'h70, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    ReadQuadOutputSeqXIP(SPI, 24'hc00, 8'h70, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);
    
    $display("==== [TEST] RDDI ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteSeq(SPI, 24'h3456, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    ReadDualIOSeqXIP(SPI, 24'h3456, 8'hc0, 8'h3, 8'h08, 24'd10, 1'b0, 8'hff);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    FastReadSeq(SPI, 24'h3456, 8'hc0, 8'h3, 8'h08, 24'd10);
  
    $display("==== [TEST] RDDI, Latency 0 ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h01);
    WriteSeq(SPI, 24'h3829, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    ReadDualIOSeqXIP(SPI, 24'h3829, 8'hc0, 8'h3, 8'h00, 24'd10, 1'b0, 8'hff);

    $display("==== [TEST] RDDI 32-byte Wrap ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h11, 8'h01); // 32-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'hde00, 8'h10, 8'h1, 8'd40);
    ReadDualIOSeqXIP(SPI, 24'hde10, 8'h20, 8'h1, 8'h08, 8'd20, 1'b0, 8'hff);

    $display("==== [TEST] RDDI XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(SPI, 24'h600, 8'h70, 8'h1, 24'd10);
    ReadDualIOSeqXIP(SPI, 24'h600, 8'h70, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    ReadDualIOSeqXIP(SPI, 24'h605, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    ReadDualIOSeqXIP(SPI, 24'h605, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    ReadDualIOSeqXIP(SPI, 24'h600, 8'h70, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    ReadDualIOSeqXIP(SPI, 24'h600, 8'h70, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);
    
    $display("==== [TEST] RDQI ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteSeq(SPI, 24'h4567, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    ReadQuadIOSeqXIP(SPI, 24'h4567, 8'hc0, 8'h3, 8'h08, 24'd10, 1'b0, 8'hff);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    FastReadSeq(SPI, 24'h4567, 8'hc0, 8'h3, 8'h08, 24'd10);
  
    $display("==== [TEST] RDQI, Latency 0 ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h00, 8'h00, 8'h01);
    WriteSeq(SPI, 24'hab82, 8'hc0, 8'h3, 24'd10);
    ReadReg(SPI, 24'h0, 8'h00); //clear addr latch
    ReadQuadIOSeqXIP(SPI, 24'hab82, 8'hc0, 8'h3, 8'h00, 24'd10, 1'b0, 8'hff);

    $display("==== [TEST] RDQI 32-byte Wrap ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h11, 8'h01); // 32-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'ha100, 8'h10, 8'h1, 8'd40);
    ReadQuadIOSeqXIP(SPI, 24'ha110, 8'h20, 8'h1, 8'h08, 8'd20, 1'b0, 8'hff);

    $display("==== [TEST] RDQI XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    FastWriteSeq(SPI, 24'h1300, 8'h70, 8'h1, 24'd10);
    ReadQuadIOSeqXIP(SPI, 24'h1300, 8'h70, 8'h1, 8'h8, 24'd5, 1'b0, 8'hA5); // Enable XIP
    ReadQuadIOSeqXIP(SPI, 24'h1305, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'hA0);
    ReadQuadIOSeqXIP(SPI, 24'h1305, 8'h75, 8'h1, 8'h8, 24'd5, 1'b1, 8'h60);
    ReadQuadIOSeqXIP(SPI, 24'h1300, 8'h70, 8'h1, 8'h8, 24'd10, 1'b1, 8'hF3); // Disable XIP
    ReadQuadIOSeqXIP(SPI, 24'h1300, 8'h70, 8'h1, 8'h8, 24'd10, 1'b0, 8'hF3);

    $display("==== [TEST] WDUI ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteDualInputSeqXIP(SPI, 24'hdead, 8'h00, 8'h6, 24'd10, 1'b0, 8'hff);
    FastReadSeq(SPI, 24'hdead, 8'h00, 8'h6, 8'h08, 24'd10);
    
    $display("==== [TEST] WDUI no WREN ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00);
    WriteDualInputSeqXIP(SPI, 24'hdead, 8'h00, 8'h6, 24'd10, 1'b0, 8'hff);
    //FastReadSeq(SPI, 24'hdead, 8'h00, 8'h6, 8'h08, 24'd10);
    
    $display("==== [TEST] WDUI XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteDualInputSeqXIP(SPI, 24'h100, 8'h30, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    WriteDualInputSeqXIP(SPI, 24'h105, 8'h35, 8'h1, 24'd5, 1'b1, 8'hC9);
    WriteDualInputSeqXIP(SPI, 24'h10a, 8'h3a, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(SPI, 24'h100, 8'h30, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);
    
    $display("==== [TEST] WQDI ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteQuadInputSeqXIP(SPI, 24'hbeef, 8'h00, 8'h6, 24'd10, 1'b0, 8'hff);
    FastReadSeq(SPI, 24'hbeef, 8'h00, 8'h6, 8'h08, 24'd10);
    
    $display("==== [TEST] WQDI XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteQuadInputSeqXIP(SPI, 24'h200, 8'h30, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    WriteQuadInputSeqXIP(SPI, 24'h205, 8'h35, 8'h1, 24'd5, 1'b1, 8'hC9);
    WriteQuadInputSeqXIP(SPI, 24'h20a, 8'h3a, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(SPI, 24'h200, 8'h30, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);
    
    $display("==== [TEST] WDIO ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteDualIOSeqXIP(SPI, 24'hface, 8'h00, 8'h6, 24'd10, 1'b0, 8'hff);
    FastReadSeq(SPI, 24'hface, 8'h00, 8'h6, 8'h08, 24'd10);
    
    $display("==== [TEST] WDIO no WREN ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h00);
    WriteDualIOSeqXIP(SPI, 24'hb055, 8'h00, 8'h6, 24'd10, 1'b0, 8'hff);
    //FastReadSeq(SPI, 24'hdead, 8'h00, 8'h6, 8'h08, 24'd10);
    
    $display("==== [TEST] WDIO XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteDualIOSeqXIP(SPI, 24'h100, 8'hd0, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    WriteDualIOSeqXIP(SPI, 24'h105, 8'hd5, 8'h1, 24'd5, 1'b1, 8'hC9);
    WriteDualIOSeqXIP(SPI, 24'h10a, 8'hda, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(SPI, 24'h100, 8'hd0, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);

    $display("==== [TEST] WQIO ====");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteQuadIOSeqXIP(SPI, 24'h0ace, 8'h00, 8'h6, 24'd10, 1'b0, 8'hff);
    FastReadSeq(SPI, 24'h0ace, 8'h00, 8'h6, 8'h08, 24'd10);
    
    $display("==== [TEST] WQIO XIP test: WRENS=SRAM(0b01), MLATS=8 (SPI) ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h00, 8'h01);
    WriteQuadIOSeqXIP(SPI, 24'hb00, 8'h30, 8'h1, 24'd5, 1'b0, 8'hA1); // Enable XIP
    WriteQuadIOSeqXIP(SPI, 24'hb05, 8'h35, 8'h1, 24'd5, 1'b1, 8'hC9);
    WriteQuadIOSeqXIP(SPI, 24'hb0a, 8'h3a, 8'h1, 24'd5, 1'b1, 8'hFF); // Disable XIP
    FastReadSeqXIP(SPI, 24'hb00, 8'h30, 8'h1, 8'h8, 24'd15, 1'b0, 8'hFF);
    
    $finish;

  end
endmodule
