`timescale 100ps / 10ps

/*
 * Options for modeling byte transmission (uncomment at most 1)
 * - SUHDCHK_ZHD
 *   - Line is released (data=HiZ) after hold time
 * - SUHDCK_CKFE
 *   - Data is set during falling edge of CLK, line released at falling edge
 *     of CLK
 * - default (no option uncommented)
 *   - Data is set to 0 after hold time, line released (tCL-tSU) after
 *     last falling edge of CLK
 *
 */
//`define SUHDCHK_ZHD
`define SUHDCHK_CKFE

module stimulus 
	(
		SO_IO1,
		CLK,
		IO3, 
		SI_IO0 ,
		WPn_IO2,
		CSn
	);

  inout SO_IO1;
	inout IO3;
	inout SI_IO0;
	inout WPn_IO2;
	
	output CSn;
	output CLK;
        
	reg CSn = 1;
	reg CLK = 0;
        
	reg SI_IO0_reg=0;
  reg SO_IO1_reg=0;
  reg WPn_IO2_reg=0;
  reg IO3_reg=0;
        
  reg [3:0] tx;

  parameter SPI = 3'h01;
  parameter DPI = 3'h02;
  parameter QPI = 3'h04;

  `include "../shr/ref/task.v"
	 
  assign SI_IO0 = tx[0] ? SI_IO0_reg : 1'bz;  
  assign SO_IO1 = tx[1] ? SO_IO1_reg : 1'bz;  
  assign WPn_IO2 = WPn_IO2_reg;
  assign IO3 = tx[3] ? IO3_reg : 1'bz;  

  initial begin
    tx = 4'd0;
    #10000;

    WPn_IO2_reg = 1'b1;
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h80);
    $display("[TEST] WP#EN enabled");
    $display("[TEST] WP# pin is high"); 
    WriteEnable(SPI);
    WriteStatus(SPI, 8'hFF);
    ReadStatus(SPI, 8'hFF);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'hFF, 8'hFF, 8'hFF, 8'hFF);
    ReadConfig(SPI, 8'hFF, 8'hFF, 8'hFF, 8'hFF);

    WPn_IO2_reg = 1'b0;
    $display("[TEST] WP# pin is low");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'hA5);
    ReadStatus(SPI, 8'hA5);

    WPn_IO2_reg = 1'b1;
    $display("[TEST] WP# pin is high"); 
    WriteEnable(SPI);
    WPn_IO2_reg = 0;
    $display("[TEST] WP# pin is low");
    WriteConfig(SPI, 8'hC3, 8'hC3, 8'hC3, 8'hC3);
    ReadConfig(SPI, 8'hC3, 8'hC3, 8'hC3, 8'hC3);
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    ReadStatus(SPI, 8'h00);

    WPn_IO2_reg = 1'b1;
    $display("[TEST] WP# pin is high"); 
    DPIEnable(SPI);
    WriteEnable(DPI);
    WPn_IO2_reg = 0;
    $display("[TEST] WP# pin is low");
    WriteConfig(DPI, 8'h69, 8'h69, 8'h78, 8'h78);
    ReadConfig(DPI, 8'h69, 8'h69, 8'h78, 8'h78);

    $finish;
  end
endmodule
