`timescale 100ps / 10ps

/*
 * Options for modeling byte transmission (uncomment at most 1)
 * - SUHDCHK_ZHD
 *   - Line is released (data=HiZ) after hold time
 * - SUHDCK_CKFE
 *   - Data is set during falling edge of CLK, line released at falling edge
 *     of CLK
 * - default (no option uncommented)
 *   - Data is set to 0 after hold time, line released (tCL-tSU) after
 *     last falling edge of CLK
 *
 */
//`define SUHDCHK_ZHD
`define SUHDCHK_CKFE

module stimulus 
	(
		SO_IO1,
		CLK,
		IO3, 
		SI_IO0 ,
		WPn_IO2,
		CSn
	);

  inout SO_IO1;
	inout IO3;
	inout SI_IO0;
	inout WPn_IO2;
	
	output CSn;
	output CLK;
        
	reg CSn = 1;
	reg CLK = 0;
        
	reg SI_IO0_reg=0;
  reg SO_IO1_reg=0;
  reg WPn_IO2_reg=0;
  reg IO3_reg=0;
        
  reg [3:0] tx;

  parameter SPI = 3'h01;
  parameter DPI = 3'h02;
  parameter QPI = 3'h04;

  `include "../shr/ref/task.v"
	 
  assign SI_IO0 = tx[0] ? SI_IO0_reg : 1'bz;  
  assign SO_IO1 = tx[1] ? SO_IO1_reg : 1'bz;  
  assign WPn_IO2 = tx[2] ? WPn_IO2_reg : 1'bz;  
  assign IO3 = tx[3] ? IO3_reg : 1'bz;  

  initial begin
    tx = 4'd0;
    #10000;

    $display("==== [TEST] 16-byte Wrap (SPI) ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h10, 8'h01); // 16-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'hcd00, 8'h00, 8'h1, 8'd20);
    FastReadSeq(SPI, 24'hcd04, 8'h04, 8'h1, 8'h08, 8'd16);
    $display("==== ----> Turning off Wrap ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h04, 8'h01);
    FastReadSeq(SPI, 24'hcd04, 8'h04, 8'h1, 8'h08, 8'd16);

    $display("==== [TEST] 32-byte Wrap (SPI) ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h11, 8'h01); // 32-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'hcd00, 8'h10, 8'h1, 8'd40);
    FastReadSeq(SPI, 24'hcd10, 8'h20, 8'h1, 8'h08, 8'd40);

    $display("==== [TEST] 64-byte Wrap (SPI) ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h12, 8'h01); // 64-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'hcd00, 8'h20, 8'h1, 8'd70);
    FastReadSeq(SPI, 24'hcd20, 8'h40, 8'h1, 8'h08, 8'd54);

    $display("==== [TEST] 128-byte Wrap (SPI) ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h13, 8'h01); // 128-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'hab00, 8'h00, 8'h1, 8'd130);
    FastReadSeq(SPI, 24'hab50, 8'h50, 8'h1, 8'h08, 8'd60);

    $display("==== [TEST] 256-byte Wrap (SPI) ==== ");
    WriteEnable(SPI);
    WriteStatus(SPI, 8'h00);
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h14, 8'h01); // 256-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(SPI, 24'hc200, 8'h00, 8'h1, 8'd255);
    FastWriteSeq(SPI, 24'hc2ff, 8'hff, 8'h1, 8'd1);
    FastReadSeq(SPI, 24'hc2f0, 8'hf0, 8'h1, 8'h08, 8'd30);
    $display("==== ----> Turning off Wrap ==== ");
    WriteEnable(SPI);
    WriteConfig(SPI, 8'h00, 8'h08, 8'h04, 8'h01);
    FastReadSeq(SPI, 24'hc2f0, 8'hf0, 8'h1, 8'h08, 8'd30);

    $display("==== [TEST] 16-byte Wrap (DPI) ==== ");
    DPIEnable(SPI);
    WriteEnable(DPI);
    WriteStatus(DPI, 8'h00);
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h08, 8'h10, 8'h01); // 16-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(DPI, 24'hcd00, 8'h00, 8'h1, 8'd20);
    FastReadSeq(DPI, 24'hcd04, 8'h04, 8'h1, 8'h08, 8'd16);
    $display("==== ----> Turning off Wrap ==== ");
    WriteEnable(DPI);
    WriteConfig(DPI, 8'h00, 8'h08, 8'h04, 8'h01);
    FastReadSeq(DPI, 24'hcd04, 8'h04, 8'h1, 8'h08, 8'd16);

    $display("==== [TEST] 16-byte Wrap (QPI) ==== ");
    QPIEnable(DPI);
    WriteEnable(QPI);
    WriteStatus(QPI, 8'h00);
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h08, 8'h10, 8'h01); // 16-byte Wrap, Latency 8, SRAM WRENS
    FastWriteSeq(QPI, 24'hcd00, 8'h00, 8'h1, 8'd20);
    FastReadSeq(QPI, 24'hcd04, 8'h04, 8'h1, 8'h08, 8'd16);
    $display("==== ----> Turning off Wrap ==== ");
    WriteEnable(QPI);
    WriteConfig(QPI, 8'h00, 8'h08, 8'h04, 8'h01);
    FastReadSeq(QPI, 24'hcd04, 8'h04, 8'h1, 8'h08, 8'd16);

    $finish;
  end

endmodule
