///////////////////////////////////////////////////////////////////////////////////
//No part of this file can be copied or released without the consent of 
//Avalanche Technology
//                  
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
//                  
//Avalanche Technology Inc., Proprietary and Confidential     *
//                  
// Release:  2.1    Date 7/21/2021  
//                  
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
//  PART DESCRIPTION:
//
//  Technology: 40nm pMTJ STT-MRAM
//  Part:       AS3016204
//
//  Description: 16 Megabit HIGH PERFORMANCE SERIAL PERSISTANT SRAM MEMORY
//
////////////////////////////////////////////////////////////////////////////////////
//  FILE CONTENTS : SPI/DPI/QPI RELATED TASKS 
//
////////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////////
// MODULE DECLARATION                                                             //
////////////////////////////////////////////////////////////////////////////////////

`timescale 100ps / 10ps

/////////////////////////////////////////////////////////////////
//   Datasheet Timing
/////////////////////////////////////////////////////////////////
specify 
specparam tPOR_time  = 1; 
specparam tCSS       = 50; 
specparam tCSH       = 40; 
`ifdef DDR_TIMING
specparam tCH        = 92.5; 
specparam tCL        = 92.5; 
`else
specparam tCH        = 50; 
specparam tCL        = 50; 
`endif
specparam tCS        = 1300; 
specparam tSU        = 20;
specparam tHD        = 20;
specparam tCO        = 70;
endspecify

/////////////////////////////////////////////////////////////////
// Select chip
/////////////////////////////////////////////////////////////////
task SelectChip;
begin
  CSn = 0;
  #(tCSS);
end
endtask 
/////////////////////////////////////////////////////////////////
// Deselect chip
/////////////////////////////////////////////////////////////////
task DeSelChip;
begin
  #(tCSH);
  CSn = 1;
  #(tCS);
end
endtask
/////////////////////////////////////////////////////////////////
// Send byte
////////////////////////////////////////////////////////////////
task SendByte;
input [7:0] data;
integer i;
begin
  CLK = 0;

`ifdef SUHDCHK_ZHD
  for(i=0; i<8; i=i+1)
  begin
    tx[0] = 1;
    SI_IO0_reg = data[7-i];
    #(tSU);
    CLK = 1;
    #(tHD)
    tx[0] = 0;
    SI_IO0_reg = 0;
    #(tCH-tHD);
    CLK = 0;
    #(tCL-tSU);
  end

`elsif SUHDCHK_CKFE
  for(i=0; i<8; i=i+1)
  begin
    CLK = 0;
    tx[0] = 1;
    SI_IO0_reg = data[7-i];
    #(tCL);
    CLK = 1;
    #(tCH)
    tx[0] = 0;
    CLK = 0;
  end

`else
  tx[0] = 1;
  for(i=0; i<8; i=i+1)
  begin
    SI_IO0_reg = data[7-i];
    #(tSU);
    CLK = 1;
    #(tHD)
    SI_IO0_reg = 0;
    #(tCH-tHD);
    CLK = 0;
    #(tCL-tSU);
  end
  tx[0] = 0;
`endif

end
endtask 
/////////////////////////////////////////////////////////////////
// Send Dual byte
/////////////////////////////////////////////////////////////////
task SendByteDual;
input [7:0] data;
integer i;
begin
  CLK = 0;
`ifdef SUHDCHK_ZHD
  for(i=0; i<8; i=i+2)
  begin
    tx[1:0] = 2'b11;
    SO_IO1_reg = data[7-i];
    SI_IO0_reg = data[7-1-i];
    #(tSU);
    CLK = 1;
    #(tHD)
    tx[1:0] = 2'b00;
    SO_IO1_reg = 0;
    SI_IO0_reg = 0;
    #(tCH-tHD);
    CLK = 0;
    #(tCL-tSU);
  end

`elsif SUHDCHK_CKFE
  for(i=0; i<8; i=i+2)
  begin
    CLK = 0;
    tx[1:0] = 2'b11;
    SO_IO1_reg = data[7-i];
    SI_IO0_reg = data[7-1-i];
    #(tCL);
    CLK = 1;
    #(tCH)
    tx[1:0] = 2'b00;
    CLK = 0;
  end

`else
  tx[1:0] = 2'b11;
  for(i=0; i<8; i=i+2)
  begin
    SO_IO1_reg = data[7-i];
    SI_IO0_reg = data[7-1-i];
    #(tSU);
    CLK = 1;
    #(tHD)
    SO_IO1_reg = 0;
    SI_IO0_reg = 0;
    #(tCH-tHD);
    CLK = 0;
    #(tCL-tSU);
  end
  tx[1:0] = 2'b0;
`endif

end
endtask 
/////////////////////////////////////////////////////////////////
// Send Quad byte
/////////////////////////////////////////////////////////////////
task SendByteQuad;
input [7:0] data;
integer i;
begin
  CLK = 0;
  
`ifdef SUHDCHK_ZHD
  for(i=0; i<8; i=i+4)
  begin
    tx = 4'b1111;
    IO3_reg = data[7-i];
    WPn_IO2_reg = data[7-1-i];
    SO_IO1_reg = data[7-2-i];
    SI_IO0_reg = data[7-3-i];
    #(tSU);
    CLK = 1;
    #(tHD)
    tx = 4'b0000;
    IO3_reg = 0; 
    WPn_IO2_reg = 0;
    SO_IO1_reg = 0;
    SI_IO0_reg = 0;
    #(tCH-tHD);
    CLK = 0;
    #(tCL-tSU);
  end

`elsif SUHDCHK_CKFE
  for(i=0; i<8; i=i+4)
  begin
    CLK = 0;
    tx = 4'b1111;
    IO3_reg = data[7-i];
    WPn_IO2_reg = data[7-1-i];
    SO_IO1_reg = data[7-2-i];
    SI_IO0_reg = data[7-3-i];
    #(tCL);
    CLK = 1;
    #(tCH)
    tx = 4'b0000;
    CLK = 0;
  end

`else
  tx = 4'b1111;
  for(i=0; i<8; i=i+4)
  begin
    IO3_reg = data[7-i];
    WPn_IO2_reg = data[7-1-i];
    SO_IO1_reg = data[7-2-i];
    SI_IO0_reg = data[7-3-i];
    #(tSU);
    CLK = 1;
    #(tHD)
    IO3_reg = 0; 
    WPn_IO2_reg = 0;
    SO_IO1_reg = 0;
    SI_IO0_reg = 0;
    #(tCH-tHD);
    CLK = 0;
    #(tCL-tSU);
  end
  tx = 4'b0000;
`endif
end
endtask 
/////////////////////////////////////////////////////////////////
// Send byte DDR
////////////////////////////////////////////////////////////////
task SendByte_DDR;
input [7:0] data;
integer i;
begin
  CLK = 0;

`ifdef SUHDCHK_ZHD
  for(i=0; i<8; i=i+2)
  begin
    tx[0] = 1;
    SI_IO0_reg = data[7-i];
    #(tSU);
    CLK = 1;
    #(tHD)
    tx[0] = 0;
    SI_IO0_reg = 0;
    #(tCH-tHD-tSU);
    tx[0] = 1;
    SI_IO0_reg = data[6-i];
    #(tSU);
    CLK = 0;
    #(tHD)
    tx[0] = 0;
    SI_IO0_reg = 0;
    #(tCL-tSU-tHD);
  end

`elsif SUHDCHK_CKFE
  tx[0] = 1;
  for(i=0; i<8; i=i+2)
  begin
    CLK = 0;
    #(tCL/2.0)
    SI_IO0_reg = data[7-i];
    #(tCL/2.0);
    CLK = 1;
    #(tCH/2.0)
    SI_IO0_reg = data[6-i];
    #(tCH/2.0)
    CLK = 0;
  end
  tx[0] = 0;

`else
  tx[0] = 1;
  for(i=0; i<8; i=i+2)
  begin
    SI_IO0_reg = data[7-i];
    #(tSU);
    CLK = 1;
    #(tHD)
    SI_IO0_reg = data[6-i];
    #(tCH-tHD);
    CLK = 0;
    #(tCL-tSU);
  end
  tx[0] = 0;
`endif

end
endtask 
/////////////////////////////////////////////////////////////////
// Send Dual byte DDR
/////////////////////////////////////////////////////////////////
task SendByteDual_DDR;
input [7:0] data;
integer i;
begin
  CLK = 0;
`ifdef SUHDCHK_ZHD
  for(i=0; i<8; i=i+4)
  begin
    tx[1:0] = 2'b11;
    SO_IO1_reg = data[7-i];
    SI_IO0_reg = data[6-i];
    #(tSU);
    CLK = 1;
    #(tHD)
    tx[1:0] = 2'b00;
    SO_IO1_reg = 0;
    SI_IO0_reg = 0;
    #(tCH-tHD-tSU);
    tx[1:0] = 2'b11;
    SO_IO1_reg = data[5-i];
    SI_IO0_reg = data[4-i];
    #(tSU);
    CLK = 0;
    #(tHD)
    tx[1:0] = 2'b00;
    SO_IO1_reg = 0;
    SI_IO0_reg = 0;
    #(tCL-tSU-tHD);
  end

`elsif SUHDCHK_CKFE
  tx[1:0] = 2'b11;
  for(i=0; i<8; i=i+4)
  begin
    CLK = 0;
    #(tCL/2.0)
    SO_IO1_reg = data[7-i];
    SI_IO0_reg = data[6-i];
    #(tCL/2.0);
    CLK = 1;
    #(tCH/2.0)
    SO_IO1_reg = data[5-i];
    SI_IO0_reg = data[4-i];
    #(tCH/2.0)
    CLK = 0;
  end
  tx[1:0] = 2'b00;

`else
  tx[1:0] = 2'b11;
  for(i=0; i<8; i=i+4)
  begin
    SO_IO1_reg = data[7-i];
    SI_IO0_reg = data[6-i];
    #(tSU);
    CLK = 1;
    #(tHD)
    SO_IO1_reg = data[5-i];
    SI_IO0_reg = data[4-i];
    #(tCH-tHD);
    CLK = 0;
    #(tCL-tSU);
  end
  tx[1:0] = 2'b0;
`endif

end
endtask 
/////////////////////////////////////////////////////////////////
// Send Quad byte DDR
/////////////////////////////////////////////////////////////////
task SendByteQuad_DDR;
input [7:0] data;
integer i;
begin
  CLK = 0;
  
`ifdef SUHDCHK_ZHD
    tx = 4'b1111;
    IO3_reg = data[7];
    WPn_IO2_reg = data[6];
    SO_IO1_reg = data[5];
    SI_IO0_reg = data[4];
    #(tSU);
    CLK = 1;
    #(tHD)
    tx = 4'b0000;
    IO3_reg = 0; 
    WPn_IO2_reg = 0;
    SO_IO1_reg = 0;
    SI_IO0_reg = 0;
    #(tCH-tHD-tSU);
    tx = 4'b1111;
    IO3_reg = data[3];
    WPn_IO2_reg = data[2];
    SO_IO1_reg = data[1];
    SI_IO0_reg = data[0];
    #(tSU);
    CLK = 0;
    #(tHD)
    tx = 4'b0000;
    IO3_reg = 0; 
    WPn_IO2_reg = 0;
    SO_IO1_reg = 0;
    SI_IO0_reg = 0;
    #(tCL-tSU-tHD);

`elsif SUHDCHK_CKFE
    tx = 4'b1111;
    CLK = 0;
    #(tCL/2.0)
    IO3_reg = data[7];
    WPn_IO2_reg = data[6];
    SO_IO1_reg = data[5];
    SI_IO0_reg = data[4];
    #(tCL/2.0);
    CLK = 1;
    #(tCH/2.0)
    IO3_reg = data[3];
    WPn_IO2_reg = data[2];
    SO_IO1_reg = data[1];
    SI_IO0_reg = data[0];
    #(tCH/2.0);
    CLK = 0;
    tx = 4'b0000;

`else
  tx = 4'b1111;
    IO3_reg = data[7];
    WPn_IO2_reg = data[6];
    SO_IO1_reg = data[5];
    SI_IO0_reg = data[4];
    #(tSU);
    CLK = 1;
    #(tHD)
    IO3_reg = data[3];
    WPn_IO2_reg = data[2];
    SO_IO1_reg = data[1];
    SI_IO0_reg = data[0];
    #(tCH-tHD);
    CLK = 0;
    #(tCL-tSU);
  tx = 4'b0000;
`endif
end
endtask 
/////////////////////////////////////////////////////////////////
// Task to receive byte
/////////////////////////////////////////////////////////////////
task RcvByte;
output [7:0] data;
integer i;
begin
  CLK = 0;
  tx=0;
`ifdef SUHDCHK_CKFE
  for(i=0; i<8; i=i+1)
  begin
    CLK = 0;
    #(tCL)
    CLK = 1;
    #(tCH)
    CLK = 0;
    data[7-i] = SO_IO1;
  end
`else
  for(i=0; i<8; i=i+1)
  begin
    SI_IO0_reg = 0;
    #(tSU);
    CLK = 1;
    #(tHD)
    SI_IO0_reg = 0;
    #(tCH-tHD);
    SO_IO1_reg = 1'bz;
    SI_IO0_reg = 0;
    data[7-i] = SO_IO1;
    CLK = 0;
    #(tCL-tSU);
  end
`endif
end
endtask 
/////////////////////////////////////////////////////////////////
// Task to receive Dualbyte
/////////////////////////////////////////////////////////////////
task RcvByteDual;
output [7:0] data;
integer i;
begin
  CLK = 0;
  tx=0;
`ifdef SUHDCHK_CKFE
  for(i=0; i<8; i=i+2)
  begin
    CLK = 0;
    #(tCL)
    CLK = 1;
    #(tCH)
    CLK = 0;
    data[7-i] = SO_IO1;
    data[7-1-i] = SI_IO0;
  end
`else
  for(i=0; i<8; i=i+2)
  begin
    SO_IO1_reg = 0; 
    SI_IO0_reg = 0;    
    #(tSU);
    CLK = 1;
    #(tHD)
    SO_IO1_reg = 0;       
    SI_IO0_reg = 0;   
    #(tCH-tHD);
    SO_IO1_reg=1'bz;
    SI_IO0_reg=1'bz;  
    data[7-i] = SO_IO1;
    data[7-1-i] = SI_IO0;
    CLK = 0;
    #(tCL-tSU);
  end
`endif
end
endtask 
/////////////////////////////////////////////////////////////////
// Task to receive Quadbyte
/////////////////////////////////////////////////////////////////
task RcvByteQuad;
output [7:0] data;
integer i;
begin
  CLK = 0;
  tx=0;
`ifdef SUHDCHK_CKFE
  for(i=0; i<8; i=i+4)
  begin
    CLK = 0;
    #(tCL)
    CLK = 1;
    #(tCH)
    CLK = 0;
    data[7-i]= IO3;
    data[7-1-i]= WPn_IO2;
    data[7-2-i]= SO_IO1;
    data[7-3-i]= SI_IO0;
  end
`else
  for(i=0; i<8; i=i+4)
  begin
    IO3_reg = 0; 
    WPn_IO2_reg = 0;
    SO_IO1_reg = 0; 
    SI_IO0_reg = 0; 
    #(tSU);
    CLK = 1;
    #(tHD)
    IO3_reg = 0; 
    WPn_IO2_reg = 0;
    SO_IO1_reg = 0;
    SI_IO0_reg = 0;
    #(tCH-tHD);
    IO3_reg=1'bz;
    WPn_IO2_reg=1'bz;
    SO_IO1_reg=1'bz;
    SI_IO0_reg=1'bz;
    data[7-i]= IO3;
    data[7-1-i]= WPn_IO2;
    data[7-2-i]= SO_IO1;
    data[7-3-i]= SI_IO0;
    CLK = 0;
    #(tCL-tSU);
  end
`endif
end
endtask 
/////////////////////////////////////////////////////////////////
// Task to receive byte DDR
/////////////////////////////////////////////////////////////////
task RcvByte_DDR;
output [7:0] data;
integer i;
begin
  CLK = 0;
  tx=0;
`ifdef SUHDCHK_CKFE
  for(i=0; i<8; i=i+2)
  begin
    CLK = 0;
    #(tCL)
    CLK = 1;
    data[7-i] = SO_IO1;
    #(tCH)
    CLK = 0;
    data[6-i] = SO_IO1;
  end
`else
  for(i=0; i<8; i=i+2)
  begin
    SI_IO0_reg = 0;
    #(tSU);
    CLK = 1;
    data[7-i] = SO_IO1;
    #(tHD)
    SI_IO0_reg = 0;
    #(tCH-tHD);
    SO_IO1_reg = 1'bz;
    SI_IO0_reg = 0;
    data[6-i] = SO_IO1;
    CLK = 0;
    #(tCL-tSU);
  end
`endif
end
endtask 
/////////////////////////////////////////////////////////////////
// Task to receive Dualbyte DDR
/////////////////////////////////////////////////////////////////
task RcvByteDual_DDR;
output [7:0] data;
integer i;
begin
  CLK = 0;
  tx=0;
`ifdef SUHDCHK_CKFE
  for(i=0; i<8; i=i+4)
  begin
    CLK = 0;
    #(tCL)
    CLK = 1;
    data[7-i] = SO_IO1;
    data[6-i] = SI_IO0;
    #(tCH)
    CLK = 0;
    data[5-i] = SO_IO1;
    data[4-i] = SI_IO0;
  end
`else
  for(i=0; i<8; i=i+4)
  begin
    SO_IO1_reg = 0; 
    SI_IO0_reg = 0;    
    #(tSU);
    CLK = 1;
    data[7-i] = SO_IO1;
    data[6-i] = SI_IO0;
    #(tHD)
    SO_IO1_reg = 0;       
    SI_IO0_reg = 0;   
    #(tCH-tHD);
    SO_IO1_reg=1'bz;
    SI_IO0_reg=1'bz;  
    data[5-i] = SO_IO1;
    data[4-i] = SI_IO0;
    CLK = 0;
    #(tCL-tSU);
  end
`endif
end
endtask 
/////////////////////////////////////////////////////////////////
// Task to receive Quadbyte DDR
/////////////////////////////////////////////////////////////////
task RcvByteQuad_DDR;
output [7:0] data;
integer i;
begin
  CLK = 0;
  tx=0;
`ifdef SUHDCHK_CKFE
    CLK = 0;
    #(tCL)
    CLK = 1;
    data[7]= IO3;
    data[6]= WPn_IO2;
    data[5]= SO_IO1;
    data[4]= SI_IO0;
    #(tCH)
    CLK = 0;
    data[3]= IO3;
    data[2]= WPn_IO2;
    data[1]= SO_IO1;
    data[0]= SI_IO0;
`else
    IO3_reg = 0; 
    WPn_IO2_reg = 0;
    SO_IO1_reg = 0; 
    SI_IO0_reg = 0; 
    #(tSU);
    CLK = 1;
    data[7]= IO3;
    data[6]= WPn_IO2;
    data[5]= SO_IO1;
    data[4]= SI_IO0;
    #(tHD)
    IO3_reg = 0; 
    WPn_IO2_reg = 0;
    SO_IO1_reg = 0;
    SI_IO0_reg = 0;
    #(tCH-tHD);
    IO3_reg=1'bz;
    WPn_IO2_reg=1'bz;
    SO_IO1_reg=1'bz;
    SI_IO0_reg=1'bz;
    data[3]= IO3;
    data[2]= WPn_IO2;
    data[1]= SO_IO1;
    data[0]= SI_IO0;
    CLK = 0;
  end
`endif
end
endtask 
/////////////////////////////////////////////////////////////////
// Send command
////////////////////////////////////////////////////////////////
task Cmd;
input [2:0] Mode;
input [7:0] data;
begin
  if (Mode == 3'h1)
  begin
    SendByte(data);
  end
  else if (Mode == 3'h2)
  begin
    SendByteDual(data);
  end
  else if (Mode == 3'h4)
  begin
    SendByteQuad(data);
  end
end
endtask
/////////////////////////////////////////////////////////////////
// Send 24-bit address
/////////////////////////////////////////////////////////////////
task Addr_24bit;
input [2:0] Mode;
input [23:0] addr;
begin
  if (Mode == 3'h1)
  begin
    SendByte(addr[23:16]);
    SendByte(addr[15:8]);
    SendByte(addr[7:0]);
  end

  else if (Mode == 3'h2)
  begin
    SendByteDual(addr[23:16]);
    SendByteDual(addr[15:8]);
    SendByteDual(addr[7:0]);
  end

  else if (Mode == 3'h4)
  begin
    SendByteQuad(addr[23:16]);
    SendByteQuad(addr[15:8]);
    SendByteQuad(addr[7:0]);
  end
end
endtask
/////////////////////////////////////////////////////////////////
// Send 24-bit address DDR
/////////////////////////////////////////////////////////////////
task Addr_24bit_DDR;
input [2:0] Mode;
input [23:0] addr;
begin
  if (Mode == 3'h1)
  begin
    SendByte_DDR(addr[23:16]);
    SendByte_DDR(addr[15:8]);
    SendByte_DDR(addr[7:0]);
  end

  else if (Mode == 3'h2)
  begin
    SendByteDual_DDR(addr[23:16]);
    SendByteDual_DDR(addr[15:8]);
    SendByteDual_DDR(addr[7:0]);
  end

  else if (Mode == 3'h4)
  begin
    SendByteQuad_DDR(addr[23:16]);
    SendByteQuad_DDR(addr[15:8]);
    SendByteQuad_DDR(addr[7:0]);
  end
end
endtask
/////////////////////////////////////////////////////////////////
// Task to send mode byte
/////////////////////////////////////////////////////////////////
task ModeByte;
input [2:0] Mode;
input [7:0] data;
begin
  if (Mode == 3'h1)
  begin
    SendByte(data);
  end

  else if (Mode == 3'h2)
  begin
    SendByteDual(data);
  end

  else if (Mode == 3'h4)
  begin
    SendByteQuad(data);
  end
end
endtask
/////////////////////////////////////////////////////////////////
// Task to send mode byte DDR
/////////////////////////////////////////////////////////////////
task ModeByte_DDR;
input [2:0] Mode;
input [7:0] data;
begin
  if (Mode == 3'h1)
  begin
    SendByte_DDR(data);
  end

  else if (Mode == 3'h2)
  begin
    SendByteDual_DDR(data);
  end

  else if (Mode == 3'h4)
  begin
    SendByteQuad_DDR(data);
  end
end
endtask
/////////////////////////////////////////////////////////////////
// Task to send latency
/////////////////////////////////////////////////////////////////
task Latency;
input [7:0] num;
integer i;
begin
  CLK = 0;
  for(i=0; i<num; i=i+1)
  begin
`ifdef SUHDCHK_CKFE
    CLK = 0;
    #(tCL);
    CLK = 1;
    #(tCH);
    CLK = 0;
`else
    IO3_reg =1'bz;
    WPn_IO2_reg =1'bz;
    SO_IO1_reg = 1'bz;
    SI_IO0_reg = 1'bz;
    #(tSU);
    CLK = 1;
    #(tHD)
    IO3_reg =1'bz;
    WPn_IO2_reg =1'bz;
    SO_IO1_reg = 1'bz;
    SI_IO0_reg = 1'bz;
    #(tCH-tHD);
    CLK = 0;
    #(tCL-tSU);
`endif
  end
end
endtask
/////////////////////////////////////////////////////////////////
// SPI enable for Modes
/////////////////////////////////////////////////////////////////
task SPIEnable;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'hff);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// DPI enable for Modes
/////////////////////////////////////////////////////////////////
task DPIEnable;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'h37);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// QPI enable for Modes
/////////////////////////////////////////////////////////////////
task QPIEnable;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'h38);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Write enable
/////////////////////////////////////////////////////////////////
task WriteEnable;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'h06);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// SPI enable
/////////////////////////////////////////////////////////////////
task SPI_Enable;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'hff);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// DPI enable
/////////////////////////////////////////////////////////////////
task DPI_Enable;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'h37);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// QPI enable
/////////////////////////////////////////////////////////////////
task QPI_Enable;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'h38);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Write disable
/////////////////////////////////////////////////////////////////
task WriteDisable;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'h04);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Enter Hibernate 
/////////////////////////////////////////////////////////////////
task EnterHibernate;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'hBA);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Exit Hibernate 
/////////////////////////////////////////////////////////////////
task ExitHibernate;
begin
  SelectChip();
  #(4500000);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Enter DPD 
/////////////////////////////////////////////////////////////////
task EnterDPD;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'hB9);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Exit DPD 
/////////////////////////////////////////////////////////////////
task ExitDPD;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'hAB);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Software Reset Enable 
/////////////////////////////////////////////////////////////////
task SoftwareResetEnable;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'h66);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Software Reset 
/////////////////////////////////////////////////////////////////
task SoftwareReset;
input [2:0] Mode;
begin
  SelectChip();
  Cmd(Mode,8'h99);
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
//  Write status
/////////////////////////////////////////////////////////////////
task WriteStatus();
input [2:0] Mode;
input [7:0]data;
begin
  SelectChip();
  Cmd(Mode,8'h01);

  if (Mode == 3'h1)
  begin
    SendByte(data);
  end

  else if (Mode == 3'h2)
  begin
    SendByteDual(data);
  end

  else if (Mode == 3'h4)
  begin
    SendByteQuad(data);
  end
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
//  Write ASP 
/////////////////////////////////////////////////////////////////
task WriteASP();
input [2:0] Mode;
input [7:0]data;
begin
  SelectChip();
  Cmd(Mode,8'h1A);

  if (Mode == 3'h1)
  begin
    SendByte(data);
  end

  else if (Mode == 3'h2)
  begin
    SendByteDual(data);
  end

  else if (Mode == 3'h4)
  begin
    SendByteQuad(data);
  end
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
//  Write SN 
/////////////////////////////////////////////////////////////////
task WriteSN();
input [2:0] Mode;
input [63:0]data;
begin
  SelectChip();
  Cmd(Mode,8'hC2);

  if (Mode == 3'h1)
  begin
    SendByte(data[63:56]);
    SendByte(data[55:48]);
    SendByte(data[47:40]);
    SendByte(data[39:32]);
    SendByte(data[31:24]);
    SendByte(data[23:16]);
    SendByte(data[15:8]);
    SendByte(data[7:0]);
  end

  else if (Mode == 3'h2)
  begin
    SendByteDual(data[63:56]);
    SendByteDual(data[55:48]);
    SendByteDual(data[47:40]);
    SendByteDual(data[39:32]);
    SendByteDual(data[31:24]);
    SendByteDual(data[23:16]);
    SendByteDual(data[15:8]);
    SendByteDual(data[7:0]);
  end

  else if (Mode == 3'h4)
  begin
    SendByteQuad(data[63:56]);
    SendByteQuad(data[55:48]);
    SendByteQuad(data[47:40]);
    SendByteQuad(data[39:32]);
    SendByteQuad(data[31:24]);
    SendByteQuad(data[23:16]);
    SendByteQuad(data[15:8]);
    SendByteQuad(data[7:0]);
  end
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Write Config
/////////////////////////////////////////////////////////////////
task WriteConfig();
input [2:0] Mode;
input [7:0] data1;
input [7:0] data2;
input [7:0] data3;
input [7:0] data4;
begin
  SelectChip();
  Cmd(Mode,8'h87);

  if (Mode == 3'h1)
  begin
    SendByte(data1);
    SendByte(data2);
    SendByte(data3);
    SendByte(data4);
  end

  else if (Mode == 3'h2)
  begin
    SendByteDual(data1);
    SendByteDual(data2);
    SendByteDual(data3);
    SendByteDual(data4);
  end

  else if (Mode == 3'h4)
  begin
    SendByteQuad(data1);
    SendByteQuad(data2);
    SendByteQuad(data3);
    SendByteQuad(data4);
  end
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Write Any Register
/////////////////////////////////////////////////////////////////
task WriteReg();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data;

begin
  $display("[INFO] Write Any Register : Address %x data %x", addr, data);
  case(addr)
    24'h0: $display("       --> Status Register");
    24'h2: $display("       --> Config Register 1");
    24'h3: $display("       --> Config Register 2");
    24'h4: $display("       --> Config Register 3");
    24'h5: $display("       --> Config Register 4");
    default: $display("       --> Invalid Register address");
  endcase

  SelectChip();
  Cmd(Mode, 8'h71);
  Addr_24bit(Mode, addr);

  if (Mode == 3'h1)
  begin
    SendByte(data);
  end
  else if (Mode == 3'h2)
  begin
    SendByteDual(data);
  end
  else if (Mode == 3'h4)
  begin
    SendByteQuad(data);
  end
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Read status
/////////////////////////////////////////////////////////////////
task ReadStatus();
input [2:0] Mode;
input [7:0] data;
reg [7:0] data_rcv;
begin
  SelectChip();
  Cmd(Mode,8'h05);

  if (Mode == 3'h1)
  begin
    RcvByte(data_rcv);
  end

  else if (Mode == 3'h2)
  begin
    RcvByteDual(data_rcv);
  end

  else if (Mode == 3'h4)
  begin
    RcvByteQuad(data_rcv);
  end

  DeSelChip();

  if((data_rcv & 8'hfc) !== (data & 8'hfc))
    $display("[ERROR] Status Compare Error : Expected %x Received %x",data,data_rcv);
  else
    $display("[INFO] Status Compare Pass : Expected %x Received %x",data,data_rcv);
end
endtask
/////////////////////////////////////////////////////////////////
// Read ASP
/////////////////////////////////////////////////////////////////
task ReadASP();
input [2:0] Mode;
input [7:0] data;
reg [7:0] data_rcv;
begin
  SelectChip();
  Cmd(Mode,8'h14);

  if (Mode == 3'h1)
  begin
    RcvByte(data_rcv);
  end

  else if (Mode == 3'h2)
  begin
    RcvByteDual(data_rcv);
  end

  else if (Mode == 3'h4)
  begin
    RcvByteQuad(data_rcv);
  end

  DeSelChip();

  if(data_rcv !== data)
    $display("[ERROR] Status Compare Error : Expected %x Received %x",data,data_rcv);
  else
    $display("[INFO] Status Compare Pass : Expected %x Received %x",data,data_rcv);
end
endtask
/////////////////////////////////////////////////////////////////
// Read SN
/////////////////////////////////////////////////////////////////
task ReadSN();
input [2:0] Mode;
input [63:0] data;
reg [63:0] data_rcv;
begin
  SelectChip();
  Cmd(Mode,8'hC3);

  if (Mode == 3'h1)
  begin
    RcvByte(data_rcv[63:56]);
    RcvByte(data_rcv[55:48]);
    RcvByte(data_rcv[47:40]);
    RcvByte(data_rcv[39:32]);
    RcvByte(data_rcv[31:24]);
    RcvByte(data_rcv[23:16]);
    RcvByte(data_rcv[15:8]);
    RcvByte(data_rcv[7:0]);
  end

  else if (Mode == 3'h2)
  begin
    RcvByteDual(data_rcv[63:56]);
    RcvByteDual(data_rcv[55:48]);
    RcvByteDual(data_rcv[47:40]);
    RcvByteDual(data_rcv[39:32]);
    RcvByteDual(data_rcv[31:24]);
    RcvByteDual(data_rcv[23:16]);
    RcvByteDual(data_rcv[15:8]);
    RcvByteDual(data_rcv[7:0]);
  end

  else if (Mode == 3'h4)
  begin
    RcvByteQuad(data_rcv[63:56]);
    RcvByteQuad(data_rcv[55:48]);
    RcvByteQuad(data_rcv[47:40]);
    RcvByteQuad(data_rcv[39:32]);
    RcvByteQuad(data_rcv[31:24]);
    RcvByteQuad(data_rcv[23:16]);
    RcvByteQuad(data_rcv[15:8]);
    RcvByteQuad(data_rcv[7:0]);
  end

  DeSelChip();

  if(data_rcv !== data)
    $display("[ERROR] Status Compare Error : Expected %x Received %x",data,data_rcv);
  else
    $display("[INFO] Status Compare Pass : Expected %x Received %x",data,data_rcv);
end
endtask
/////////////////////////////////////////////////////////////////
// Read Config
/////////////////////////////////////////////////////////////////
task ReadConfig ();
input [2:0] Mode;
input [7:0] data1;
input [7:0] data2;
input [7:0] data3;
input [7:0] data4;
reg [7:0] data_rcv1;
reg [7:0] data_rcv2;
reg [7:0] data_rcv3;
reg [7:0] data_rcv4;
begin
  SelectChip();
  Cmd(Mode,8'h46);

  if (Mode == 3'h1)
  begin
    RcvByte(data_rcv1);
    RcvByte(data_rcv2);
    RcvByte(data_rcv3);
    RcvByte(data_rcv4);
  end

  else if (Mode == 3'h2)
  begin
    RcvByteDual(data_rcv1);
    RcvByteDual(data_rcv2);
    RcvByteDual(data_rcv3);
    RcvByteDual(data_rcv4);
  end
 
  else if (Mode == 3'h4)
  begin
    RcvByteQuad(data_rcv1);
    RcvByteQuad(data_rcv2);
    RcvByteQuad(data_rcv3);
    RcvByteQuad(data_rcv4);
  end
  DeSelChip();


  if(((data_rcv1 & 8'h05) !== (data1 & 8'h05)) || 
     ((data_rcv2 & 8'h0f) !== (data2 & 8'h0f)) || 
     ((data_rcv3 & 8'hf7) !== (data3 & 8'hf7)) || 
     (data_rcv4 !== data4)
    )
    $display("[ERROR] CFG Reg Compare Error : Expected %x %x %x %x Received %x %x %x %x",data1, data2, data3, data4, data_rcv1, data_rcv2, data_rcv3, data_rcv4);
  else
    $display("[INFO] CFG Reg Compare Pass : Expected %x %x %x %x Received %x %x %x %x",data1, data2, data3, data4, data_rcv1, data_rcv2, data_rcv3, data_rcv4);

end
endtask
/////////////////////////////////////////////////////////////////
// Read Config 1 
/////////////////////////////////////////////////////////////////
task ReadCR1();
input [2:0] Mode;
input [7:0] data;
reg [7:0] data_rcv;
begin
  SelectChip();
  Cmd(Mode,8'h35);

  if (Mode == 3'h1)
  begin
    RcvByte(data_rcv);
  end

  else if (Mode == 3'h2)
  begin
    RcvByteDual(data_rcv);
  end

  else if (Mode == 3'h4)
  begin
    RcvByteQuad(data_rcv);
  end

  DeSelChip();

  if((data_rcv & 8'h05) !== (data & 8'h05))
    $display("[ERROR] Status Compare Error : Expected %x Received %x",data,data_rcv);
  else
    $display("[INFO] Status Compare Pass : Expected %x Received %x",data,data_rcv);
end
endtask
/////////////////////////////////////////////////////////////////
// Read Config 2 
/////////////////////////////////////////////////////////////////
task ReadCR2();
input [2:0] Mode;
input [7:0] data;
reg [7:0] data_rcv;
begin
  SelectChip();
  Cmd(Mode,8'h3f);

  if (Mode == 3'h1)
  begin
    RcvByte(data_rcv);
  end

  else if (Mode == 3'h2)
  begin
    RcvByteDual(data_rcv);
  end

  else if (Mode == 3'h4)
  begin
    RcvByteQuad(data_rcv);
  end

  DeSelChip();

  if((data_rcv & 8'h0F) !== (data & 8'h0F))
    $display("[ERROR] Status Compare Error : Expected %x Received %x",data,data_rcv);
  else
    $display("[INFO] Status Compare Pass : Expected %x Received %x",data,data_rcv);
end
endtask
/////////////////////////////////////////////////////////////////
// Read Config 3 
/////////////////////////////////////////////////////////////////
task ReadCR3();
input [2:0] Mode;
input [7:0] data;
reg [7:0] data_rcv;
begin
  SelectChip();
  Cmd(Mode,8'h44);

  if (Mode == 3'h1)
  begin
    RcvByte(data_rcv);
  end

  else if (Mode == 3'h2)
  begin
    RcvByteDual(data_rcv);
  end

  else if (Mode == 3'h4)
  begin
    RcvByteQuad(data_rcv);
  end

  DeSelChip();

  if((data_rcv & 8'hF7) !== (data & 8'hF7))
    $display("[ERROR] Status Compare Error : Expected %x Received %x",data,data_rcv);
  else
    $display("[INFO] Status Compare Pass : Expected %x Received %x",data,data_rcv);
end
endtask
/////////////////////////////////////////////////////////////////
// Read Config 4 
/////////////////////////////////////////////////////////////////
task ReadCR4();
input [2:0] Mode;
input [7:0] data;
reg [7:0] data_rcv;
begin
  SelectChip();
  Cmd(Mode,8'h45);

  if (Mode == 3'h1)
  begin
    RcvByte(data_rcv);
  end

  else if (Mode == 3'h2)
  begin
    RcvByteDual(data_rcv);
  end

  else if (Mode == 3'h4)
  begin
    RcvByteQuad(data_rcv);
  end

  DeSelChip();

  if((data_rcv & 8'hFF) !== (data & 8'hFF))
    $display("[ERROR] Status Compare Error : Expected %x Received %x",data,data_rcv);
  else
    $display("[INFO] Status Compare Pass : Expected %x Received %x",data,data_rcv);
end
endtask
/////////////////////////////////////////////////////////////////
// Read ID
/////////////////////////////////////////////////////////////////
task ReadID();
input [2:0] Mode;
reg [31:0] data_rcv;
begin
  SelectChip();
  Cmd(Mode,8'h9F);

  if (Mode == 3'h1)
  begin
    RcvByte(data_rcv[31:24]);
    RcvByte(data_rcv[23:16]);
    RcvByte(data_rcv[15:8]);
    RcvByte(data_rcv[7:0]);
  end

  else if (Mode == 3'h2)
  begin
    RcvByteDual(data_rcv[31:24]);
    RcvByteDual(data_rcv[23:16]);
    RcvByteDual(data_rcv[15:8]);
    RcvByteDual(data_rcv[7:0]);
  end

  else if (Mode == 3'h4)
  begin
    RcvByteQuad(data_rcv[31:24]);
    RcvByteQuad(data_rcv[23:16]);
    RcvByteQuad(data_rcv[15:8]);
    RcvByteQuad(data_rcv[7:0]);
  end

  DeSelChip();

  $display("[INFO] Read ID: Received %x", data_rcv);
end
endtask
/////////////////////////////////////////////////////////////////
// Read UID
/////////////////////////////////////////////////////////////////
task ReadUID();
input [2:0] Mode;
reg [63:0] data_rcv;
begin
  SelectChip();
  Cmd(Mode,8'h4C);

  if (Mode == 3'h1)
  begin
    RcvByte(data_rcv[63:56]);
    RcvByte(data_rcv[55:48]);
    RcvByte(data_rcv[47:40]);
    RcvByte(data_rcv[39:32]);
    RcvByte(data_rcv[31:24]);
    RcvByte(data_rcv[23:16]);
    RcvByte(data_rcv[15:8]);
    RcvByte(data_rcv[7:0]);
  end

  else if (Mode == 3'h2)
  begin
    RcvByteDual(data_rcv[63:56]);
    RcvByteDual(data_rcv[55:48]);
    RcvByteDual(data_rcv[47:40]);
    RcvByteDual(data_rcv[39:32]);
    RcvByteDual(data_rcv[31:24]);
    RcvByteDual(data_rcv[23:16]);
    RcvByteDual(data_rcv[15:8]);
    RcvByteDual(data_rcv[7:0]);
  end

  else if (Mode == 3'h4)
  begin
    RcvByteQuad(data_rcv[63:56]);
    RcvByteQuad(data_rcv[55:48]);
    RcvByteQuad(data_rcv[47:40]);
    RcvByteQuad(data_rcv[39:32]);
    RcvByteQuad(data_rcv[31:24]);
    RcvByteQuad(data_rcv[23:16]);
    RcvByteQuad(data_rcv[15:8]);
    RcvByteQuad(data_rcv[7:0]);
  end

  DeSelChip();

  $display("[INFO] Read UID: Received %x", data_rcv);
end
endtask
/////////////////////////////////////////////////////////////////
// Read Any Register
/////////////////////////////////////////////////////////////////
task ReadReg();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data;
reg [7:0] data_rcv;
reg error_flag;
integer i;
reg [31:0] id;
reg [63:0] uid;
begin

  $display("[INFO] Read Any Register : Address %x data %x", addr, data);
  case(addr)
    24'h0: $display("       --> Status Register");
    24'h2: $display("       --> Config Register 1");
    24'h3: $display("       --> Config Register 2");
    24'h4: $display("       --> Config Register 3");
    24'h5: $display("       --> Config Register 4");
    24'h30: $display("       --> Device Identification Register");
    24'h40: $display("       --> Device Unique Identification Register");
    default: $display("       --> Invalid Register address");
  endcase

  SelectChip();
  Cmd(Mode, 8'h65);
  Addr_24bit(Mode, addr);

  // 32-bit output
  if (addr == 24'h30)
  begin
  if (Mode == 3'h1)
  begin
    Latency(8);
    for (i=0; i<4; i=i+1) begin
      RcvByte(data_rcv);
      id = {id[23:0], data_rcv};
      //id[31-(i*8) : 24-(i*8)] = data_rcv;
    end
  end
  else if (Mode == 3'h2)
  begin
    Latency(4);
    for (i=0; i<4; i=i+1) begin
      RcvByteDual(data_rcv);
      id = {id[23:0], data_rcv};
      //id[31-(i*8) : 24-(i*8)] = data_rcv;
    end
  end
  else if (Mode == 3'h4)
  begin
    Latency(2);
    for (i=0; i<4; i=i+1) begin
      RcvByteQuad(data_rcv);
      id = {id[23:0], data_rcv};
      //id[31-(i*8) : 24-(i*8)] = data_rcv;
    end
  end
  end

  // 64-bit output 
  else if (addr == 24'h40)
  begin
  if (Mode == 3'h1)
  begin
    Latency(8);
    for (i=0; i<8; i=i+1) begin
      RcvByte(data_rcv);
      uid = {uid[55:0], data_rcv};
      //uid[63-(i*8) : 63-(i*8)] = data_rcv;
    end
  end
  else if (Mode == 3'h2)
  begin
    Latency(4);
    for (i=0; i<8; i=i+1) begin
      RcvByteDual(data_rcv);
      uid = {uid[55:0], data_rcv};
      //uid[63-(i*8) : 63-(i*8)] = data_rcv;
    end
  end
  else if (Mode == 3'h4)
  begin
    Latency(2);
    for (i=0; i<8; i=i+1) begin
      RcvByteQuad(data_rcv);
      uid = {uid[55:0], data_rcv};
      //uid[63-(i*8) : 63-(i*8)] = data_rcv;
    end
  end
  end

  // 8-bit output 
  else begin
  if (Mode == 3'h1)
  begin
    Latency(8);
    RcvByte(data_rcv);
  end
  else if (Mode == 3'h2)
  begin
    Latency(4);
    RcvByteDual(data_rcv);
  end
  else if (Mode == 3'h4)
  begin
    Latency(2);
    RcvByteQuad(data_rcv);
  end
  end

  DeSelChip();

  error_flag = 0;
  case(addr)
    24'h0:
      if (data_rcv != (data & 8'hfc))
        error_flag = 1'b1;
    24'h2: //CR1
      if ((data_rcv & 8'h05) != (data & 8'h05))
        error_flag = 1'b1;
    24'h3: //CR2
      if ((data_rcv & 8'h0f) != (data & 8'h0f))
        error_flag = 1'b1;
    24'h4: //CR3
      if ((data_rcv & 8'hf7) != (data & 8'hf7))
        error_flag = 1'b1;
    24'h5: //CR4
      if (data_rcv != data)
        error_flag = 1'b1;
  endcase

  if (error_flag)
    $display("[ERROR] Register Compare Error (Reg 0x%x) : Expected 0x%x Received 0x%x (time:%t)", addr, data, data_rcv, $time);
  else
    case(addr)
      24'h0, 24'h2, 24'h3, 24'h4, 24'h5:
        $display("[INFO] Register Compare Pass (Reg 0x%x) : Expected 0x%x Received 0x%x", addr, data, data_rcv);
      24'h30:
        $display("[INFO] Register Read ID : 0x%x", id);
      24'h40:
        $display("[INFO] Register Read UID : 0x%x", uid);
      default:
        $display("[ERROR] Register Compare not performed due to invalid/unsupported address (Reg 0x%x)", addr);
    endcase
end
endtask
/////////////////////////////////////////////////////////////////
// Memory Write
/////////////////////////////////////////////////////////////////
task Write();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data;
input [23:0] num_data_i;
integer i; 
begin
  $display("[INFO] Mem write  : Address %x data %x num_data %x",addr, data, num_data_i);
  SelectChip();
  Cmd(Mode,8'h02);
  Addr_24bit(Mode,addr);
  for (i=0;i<num_data_i;i=i+1)
  begin

    if (Mode == 3'h1)
    begin
      SendByte(data);
    end

    else if (Mode == 3'h2)
    begin
      SendByteDual(data);
    end

    else if (Mode == 3'h4)
    begin
      SendByteQuad(data);
    end

  end
  DeSelChip();
end
endtask

task WriteSeq();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] Mem write (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  Cmd(Mode,8'h02);
  Addr_24bit(Mode,addr);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByte(data);
    end
    else if (Mode == 3'h2)
    begin
      SendByteDual(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad(data);
    end

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task WriteAugSeq();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] ASA write (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  Cmd(Mode,8'h42);
  Addr_24bit(Mode,addr);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByte(data);
    end
    else if (Mode == 3'h2)
    begin
      SendByteDual(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad(data);
    end

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task FastWrite();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data;
input [23:0] num_data_i;
integer i; 
begin
  $display("[INFO] Fast Mem write  : Address %x data %x num_data %x",addr, data, num_data_i);
  SelectChip();
  Cmd(Mode,8'hDA);
  Addr_24bit(Mode,addr);
  ModeByte(Mode,8'hFF);
  for (i=0;i<num_data_i;i=i+1)
  begin

    if (Mode == 3'h1)
    begin
      SendByte(data);
    end

    else if (Mode == 3'h2)
    begin
      SendByteDual(data);
    end

    else if (Mode == 3'h4)
    begin
      SendByteQuad(data);
    end

  end
  DeSelChip();
end
endtask

task FastWriteSeq();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] Fast Mem write (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  Cmd(Mode,8'hDA);
  Addr_24bit(Mode,addr);
  ModeByte(Mode,8'hFF);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByte(data);
    end
    else if (Mode == 3'h2)
    begin
      SendByteDual(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad(data);
    end

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task DDRFastWriteSeq();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] Fast Mem write DDR (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  Cmd(Mode,8'hDE);
  Addr_24bit_DDR(Mode,addr);
  ModeByte_DDR(Mode,8'hFF);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByte_DDR(data);
    end
    else if (Mode == 3'h2)
    begin
      SendByteDual_DDR(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad_DDR(data);
    end

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task FastWriteSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
input xip_on;
input [7:0] mode_byte;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] Fast Mem write (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'hDA);
  end
  Addr_24bit(Mode,addr);
  ModeByte(Mode,mode_byte);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByte(data);
    end
    else if (Mode == 3'h2)
    begin
      SendByteDual(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad(data);
    end

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task WriteDualInputSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
input xip_on;
input [7:0] mode_byte;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] Write Dual Input (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'hA2);
  end
  Addr_24bit(Mode,addr);
  ModeByte(Mode,mode_byte);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByteDual(data);
    end
    /*else if (Mode == 3'h2)
    begin
      SendByteDual(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad(data);
    end*/

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task WriteDualIOSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
input xip_on;
input [7:0] mode_byte;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] Write Dual I/O (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'hA1);
  end
  Addr_24bit(3'h2,addr);
  ModeByte(3'h2,mode_byte);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByteDual(data);
    end
    /*else if (Mode == 3'h2)
    begin
      SendByteDual(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad(data);
    end*/

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task WriteQuadInputSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
input xip_on;
input [7:0] mode_byte;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] Write Quad Input (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'h32);
  end
  Addr_24bit(Mode,addr);
  ModeByte(Mode,mode_byte);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByteQuad(data);
    end
    /*else if (Mode == 3'h2)
    begin
      SendByteDual(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad(data);
    end*/

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task WriteQuadIOSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
input xip_on;
input [7:0] mode_byte;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] Write Quad I/O (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'hd2);
  end
  Addr_24bit(3'h4,addr);
  ModeByte(3'h4,mode_byte);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByteQuad(data);
    end
    /*else if (Mode == 3'h2)
    begin
      SendByteDual(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad(data);
    end*/

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task DDRFastWriteSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
input xip_on;
input [7:0] mode_byte;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] Fast Mem write (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'hDE);
  end
  Addr_24bit_DDR(Mode,addr);
  ModeByte_DDR(Mode,mode_byte);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByte_DDR(data);
    end
    else if (Mode == 3'h2)
    begin
      SendByteDual_DDR(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad_DDR(data);
    end

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task DDRWriteQuadInputSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
input xip_on;
input [7:0] mode_byte;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] DDR Write Quad Input (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'h31);
  end
  Addr_24bit_DDR(Mode,addr);
  ModeByte_DDR(Mode,mode_byte);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByteQuad_DDR(data);
    end
    /*else if (Mode == 3'h2)
    begin
      SendByteDual_DDR(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad_DDR(data);
    end*/

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task DDRWriteQuadIOSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [23:0] num_data_i;
input xip_on;
input [7:0] mode_byte;
reg [7:0] data;
integer i; 
begin
  $display("[INFO] DDR Write Quad I/O (sequential)  : Address %x num_data %x data_base %x data_inc %x",addr, num_data_i, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'hD1);
  end
  Addr_24bit_DDR(3'h4,addr);
  ModeByte_DDR(3'h4,mode_byte);
  for (i=0; i<num_data_i; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      SendByteQuad_DDR(data);
    end
    /*else if (Mode == 3'h2)
    begin
      SendByteDual_DDR(data);
    end
    else if (Mode == 3'h4)
    begin
      SendByteQuad_DDR(data);
    end*/

    data = data + data_inc;
  end
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Memory Read
/////////////////////////////////////////////////////////////////
task Read();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data;
input [7:0] latency_val;
input [23:0] num_data_o;
integer i;
reg [7:0] data_rcv;
begin
  $display("[INFO] Mem read  : Address %x data %x latency %x num_data %x",addr, data, latency_val, num_data_o);
  SelectChip();
  Cmd(Mode,8'h03);
  Addr_24bit(Mode,addr);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      RcvByte(data_rcv);
    end
    else if (Mode == 3'h2)
    begin
      RcvByteDual(data_rcv);
    end
    else if (Mode == 3'h4)
    begin
    RcvByteQuad(data_rcv);
    end

  if(data_rcv !== data)
    $display("[ERROR] Mem Read Compare Error : Expected %x Received %x",data,data_rcv);
  else
    $display("[INFO] Mem Read Compare Pass : Expected %x Received %x",data,data_rcv);
end
    DeSelChip();
end
endtask

task ReadSeq();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem read (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  Cmd(Mode, 8'h03);
  Addr_24bit(Mode, addr);
  Latency(latency_val);

  for (i=0; i<num_data_o; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      RcvByte(data_rcv);
    end
    else if (Mode == 3'h2)
    begin
      RcvByteDual(data_rcv);
    end
    else if (Mode == 3'h4)
    begin
      RcvByteQuad(data_rcv);
    end

    if (data_rcv !== data)
    $display("[ERROR] Mem Read Compare Error : Expected %x Received %x",data,data_rcv);
    else
    $display("[INFO] Mem Read Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();

end
endtask

task ReadAugSeq();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] ASA read (sequential) : Address %x num_data %x data_base %x data_inc %x",addr, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  Cmd(Mode, 8'h4B);
  Addr_24bit(Mode, addr);
  Latency(latency_val);

  for (i=0; i<num_data_o; i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      RcvByte(data_rcv);
    end
    else if (Mode == 3'h2)
    begin
      RcvByteDual(data_rcv);
    end
    else if (Mode == 3'h4)
    begin
      RcvByteQuad(data_rcv);
    end

    if (data_rcv !== data)
    $display("[ERROR] ASA Read Compare Error : Expected %x Received %x",data,data_rcv);
    else
    $display("[INFO] ASA Read Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();

end
endtask

/////////////////////////////////////////////////////////////////
// Memory Read using Fast Read
/////////////////////////////////////////////////////////////////
task FastRead();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data;
input [7:0] latency_val;
input [23:0] num_data_o;
integer i;
reg [7:0] data_rcv;
begin
  $display("[INFO] Mem Fast read  : Address %x data %x latency %x num_data %x",addr, data, latency_val, num_data_o);
  SelectChip();
  Cmd(Mode,8'h0B);
  Addr_24bit(Mode,addr);
  ModeByte(Mode,8'hFF);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
    RcvByte(data_rcv);
    end
    else if (Mode == 3'h2)
    begin
      RcvByteDual(data_rcv);
    end
    else if (Mode == 3'h4)
    begin
      RcvByteQuad(data_rcv);
    end

    if(data_rcv !== data)
      $display("[ERROR] Mem Fast Read Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Fast Read Compare Pass : Expected %x Received %x",data,data_rcv);
  end
  DeSelChip();
end
endtask

task FastReadSeq();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem Fast read (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  Cmd(Mode,8'h0B);
  Addr_24bit(Mode,addr);
  ModeByte(Mode,8'hFF);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
    RcvByte(data_rcv);
    end
    else if (Mode == 3'h2)
    begin
      RcvByteDual(data_rcv);
    end
    else if (Mode == 3'h4)
    begin
      RcvByteQuad(data_rcv);
    end

    if(data_rcv !== data)
      $display("[ERROR] Mem Fast Read Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Fast Read Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task DDRFastReadSeq();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem Fast read DDR (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  Cmd(Mode,8'h0D);
  Addr_24bit_DDR(Mode,addr);
  ModeByte_DDR(Mode,8'hFF);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      RcvByte_DDR(data_rcv);
    end
    else if (Mode == 3'h2)
    begin
      RcvByteDual_DDR(data_rcv);
    end
    else if (Mode == 3'h4)
    begin
      RcvByteQuad_DDR(data_rcv);
    end

    if(data_rcv !== data)
      $display("[ERROR] Mem Fast Read Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Fast Read Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task FastReadSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
input xip_on;
input [7:0] mode_byte;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem Fast read (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'h0B);
  end
  Addr_24bit(Mode,addr);
  ModeByte(Mode,mode_byte);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
    RcvByte(data_rcv);
    end
    else if (Mode == 3'h2)
    begin
      RcvByteDual(data_rcv);
    end
    else if (Mode == 3'h4)
    begin
      RcvByteQuad(data_rcv);
    end

    if(data_rcv !== data)
      $display("[ERROR] Mem Fast Read Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Fast Read Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task ReadDualOutputSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
input xip_on;
input [7:0] mode_byte;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem Read Dual Output (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'h3B);
  end
  Addr_24bit(Mode,addr);
  ModeByte(Mode,mode_byte);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      RcvByteDual(data_rcv);
    end
    //else if (Mode == 3'h2)
    //begin
    //  RcvByteDual(data_rcv);
    //end
    //else if (Mode == 3'h4)
    //begin
    //  RcvByteQuad(data_rcv);
    //end

    if(data_rcv !== data)
      $display("[ERROR] Mem Read Dual Output Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Read Dual Output Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task ReadQuadOutputSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
input xip_on;
input [7:0] mode_byte;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem Read Quad Output (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'h6B);
  end
  Addr_24bit(Mode,addr);
  ModeByte(Mode,mode_byte);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      RcvByteQuad(data_rcv);
    end
    //else if (Mode == 3'h2)
    //begin
    //  RcvByteDual(data_rcv);
    //end
    //else if (Mode == 3'h4)
    //begin
    //  RcvByteQuad(data_rcv);
    //end

    if(data_rcv !== data)
      $display("[ERROR] Mem Read Quad Output Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Read Quad Output Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task ReadDualIOSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
input xip_on;
input [7:0] mode_byte;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem Read Dual I/O (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'hBB);
  end
  Addr_24bit(3'h2,addr);
  ModeByte(3'h2,mode_byte);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      RcvByteDual(data_rcv);
    end
    //else if (Mode == 3'h2)
    //begin
    //  RcvByteDual(data_rcv);
    //end
    //else if (Mode == 3'h4)
    //begin
    //  RcvByteQuad(data_rcv);
    //end

    if(data_rcv !== data)
      $display("[ERROR] Mem Read Dual I/O Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Read Dual I/O Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task ReadQuadIOSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
input xip_on;
input [7:0] mode_byte;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem Read Quad I/O (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'hEB);
  end
  Addr_24bit(3'h4,addr);
  ModeByte(3'h4,mode_byte);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
      RcvByteQuad(data_rcv);
    end
    //else if (Mode == 3'h2)
    //begin
    //  RcvByteDual(data_rcv);
    //end
    //else if (Mode == 3'h4)
    //begin
    //  RcvByteQuad(data_rcv);
    //end

    if(data_rcv !== data)
      $display("[ERROR] Mem Read Quad I/O Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Read Quad I/O Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task DDRFastReadSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
input xip_on;
input [7:0] mode_byte;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem Fast read DDR (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'h0D);
  end
  Addr_24bit_DDR(Mode,addr);
  ModeByte_DDR(Mode,mode_byte);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
    RcvByte_DDR(data_rcv);
    end
    else if (Mode == 3'h2)
    begin
      RcvByteDual_DDR(data_rcv);
    end
    else if (Mode == 3'h4)
    begin
      RcvByteQuad_DDR(data_rcv);
    end

    if(data_rcv !== data)
      $display("[ERROR] Mem Fast Read DDR Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Fast Read DDR Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();
end
endtask

task DDRReadDualIOSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
input xip_on;
input [7:0] mode_byte;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem Read Dual I/O DDR (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'hBD);
  end
  Addr_24bit_DDR(3'h2,addr);
  ModeByte_DDR(3'h2,mode_byte);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
    RcvByteDual_DDR(data_rcv);
    end
    /*else if (Mode == 3'h2)
    begin
      RcvByteDual_DDR(data_rcv);
    end
    else if (Mode == 3'h4)
    begin
      RcvByteQuad_DDR(data_rcv);
    end*/

    if(data_rcv !== data)
      $display("[ERROR] Mem Read Dual I/O DDR Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Read Dual I/O DDR Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();
end
endtask
task DDRReadQuadIOSeqXIP();
input [2:0] Mode;
input [23:0] addr;
input [7:0] data_base;
input [7:0] data_inc;
input [7:0] latency_val;
input [23:0] num_data_o;
input xip_on;
input [7:0] mode_byte;
integer i;
reg [7:0] data_rcv;
reg [7:0] data;
begin
  $display("[INFO] Mem Read Quad I/O DDR (sequential) : Address %x latency %x num_data %x data_base %x data_inc %x",addr, latency_val, num_data_o, data_base, data_inc);
  data = data_base;
  SelectChip();
  if (!xip_on) begin
    Cmd(Mode,8'hED);
  end
  Addr_24bit_DDR(3'h4,addr);
  ModeByte_DDR(3'h4,mode_byte);
  Latency(latency_val);
  for (i=0;i<num_data_o;i=i+1)
  begin
    if (Mode == 3'h1)
    begin
    RcvByteQuad_DDR(data_rcv);
    end
    /*else if (Mode == 3'h2)
    begin
      RcvByteQuad_DDR(data_rcv);
    end
    else if (Mode == 3'h4)
    begin
      RcvByteQuad_DDR(data_rcv);
    end*/

    if(data_rcv !== data)
      $display("[ERROR] Mem Read Quad I/O DDR Compare Error : Expected %x Received %x",data,data_rcv);
    else
      $display("[INFO] Mem Read Quad I/O DDR Compare Pass : Expected %x Received %x",data,data_rcv);

    data = data + data_inc;
  end
  DeSelChip();
end
endtask
/////////////////////////////////////////////////////////////////
// Status register write read check
/////////////////////////////////////////////////////////////////
task Status_Register_Check();
input [2:0] Mode;
begin 
   parameter SPI = 3'h1;
        parameter DPI = 3'h2;
        parameter QPI = 3'h4;
       
        if(Mode == 3'h1)
begin
WriteEnable(SPI);
   WriteStatus(SPI,8'hFF);
   ReadStatus(SPI,8'hFF);
   
WriteStatus(SPI,8'haa);
   ReadStatus(SPI,8'hFF);
end

        else if(Mode  == 3'h2)
begin
WriteEnable(DPI);
   WriteStatus(DPI,8'hFF);
   ReadStatus(DPI,8'hFF);
   
WriteStatus(DPI,8'haa);
   ReadStatus(DPI,8'hFF);
        end
 
        else if(Mode == 3'h4)
begin
WriteEnable(QPI);
   WriteStatus(QPI,8'hFF);
   ReadStatus(QPI,8'hFF);
   
WriteStatus(QPI,8'haa);
   ReadStatus(QPI,8'hFF);
        end
end
endtask
/////////////////////////////////////////////////////////////////
// Config register write read check
/////////////////////////////////////////////////////////////////
task Config_Register_Check();
input [2:0] Mode;
begin 
parameter SPI = 3'h1;
        parameter DPI = 3'h2;
        parameter QPI = 3'h4;

$display("*****************************************************************");
$display("** Config register write read check");
$display("*****************************************************************");
        if(Mode == 3'h1)
begin
   WriteEnable(SPI);
   WriteConfig(SPI,8'hFF,8'hFF,8'hFF,8'hFF);
   ReadConfig(SPI,8'hFF,8'hFF,8'hFF,8'hFF);

   WriteConfig(SPI,8'haa,8'haa,8'haa,8'haa);
   ReadConfig(SPI,8'hFF,8'hFF,8'hFF,8'hFF);

  WriteEnable(SPI);
  WriteDisable(SPI);
   WriteConfig(SPI,8'haa,8'haa,8'haa,8'haa);
   ReadConfig(SPI,8'hFF,8'hFF,8'hFF,8'hFF);

  WriteEnable(SPI);
   WriteConfig(SPI,8'h00,8'h08,8'h00,8'h00);
   ReadConfig(SPI,8'h00,8'h08,8'h00,8'h00);

        end 

        else if(Mode == 3'h2)
begin
WriteEnable(DPI);
   WriteConfig(DPI,8'hFF,8'hFF,8'hFF,8'hFF);
   ReadConfig(DPI,8'hFF,8'hFF,8'hFF,8'hFF);

   WriteConfig(DPI,8'haa,8'haa,8'haa,8'haa);
   ReadConfig(DPI,8'hFF,8'hFF,8'hFF,8'hFF);

  WriteEnable(DPI);
  WriteDisable(DPI);
   WriteConfig(DPI,8'haa,8'haa,8'haa,8'haa);
   ReadConfig(DPI,8'hFF,8'hFF,8'hFF,8'hFF);

  WriteEnable(DPI);
   WriteConfig(DPI,8'h00,8'h08,8'h00,8'h00);
   ReadConfig(DPI,8'h00,8'h08,8'h00,8'h00);
        end

        else if(Mode == 3'h4)
begin     
  WriteEnable(QPI);
  WriteConfig(QPI,8'hFF,8'hFF,8'hFF,8'hFF);
  ReadConfig(QPI,8'hFF,8'hFF,8'hFF,8'hFF);

  WriteConfig(QPI,8'haa,8'haa,8'haa,8'haa);
   ReadConfig(QPI,8'hFF,8'hFF,8'hFF,8'hFF);

  WriteEnable(QPI);
  WriteDisable(QPI);
   WriteConfig(QPI,8'haa,8'haa,8'haa,8'haa);
   ReadConfig(QPI,8'hFF,8'hFF,8'hFF,8'hFF);

  WriteEnable(QPI);
   //WriteConfig(QPI,8'h00,8'h08,8'h00,8'h00);
   WriteConfig(QPI,8'h00,8'h08,8'haa,8'h55);
   //ReadConfig(QPI,8'h00,8'h08,8'h00,8'h00);
   ReadConfig(QPI,8'h00,8'h08,8'haa,8'h55);
        end
end
endtask
/////////////////////////////////////////////////////////////////
// Write Read Memory check
/////////////////////////////////////////////////////////////////
task Memory_Write_Read_Check();
input [2:0] Mode;
begin 

  parameter SPI = 3'h1;
  parameter DPI = 3'h2;
  parameter QPI = 3'h4;
        
  $display("*****************************************************************");
  $display("** Write Read Memory check ");
  $display("*****************************************************************");
  if(Mode == SPI)
  begin
    WriteEnable(SPI);
    WriteStatus(SPI,8'h00);
    ReadStatus(SPI,8'h00);

    WriteEnable(SPI);
    WriteConfig(SPI,8'h00,8'h08,8'h00,8'h00);
    ReadConfig(SPI,8'h00,8'h08,8'h00,8'h00);

    WriteEnable(SPI);
    Write(SPI,24'h000000,8'ha3,8'h0f);
    //Read(SPI,24'h000000,8'ha3,8'h08,8'h0f);
    Read(SPI,24'h000000,8'ha3,8'h00,8'h0f); // Latency 0 since SPI Read
    FastRead(SPI,24'h000000,8'ha3,8'h08,8'h0f);
  end
   
  else if(Mode == DPI)
  begin
    WriteEnable(DPI);
    WriteStatus(DPI,8'h00);
    ReadStatus(DPI,8'h00);

    WriteEnable(DPI);
    WriteConfig(DPI,8'h00,8'h08,8'h00,8'h00);
    ReadConfig(DPI,8'h00,8'h08,8'h00,8'h00);

    WriteEnable(DPI);
    //Write(DPI,24'h000100,8'hb4,8'h0f);
    FastWrite(DPI,24'h000100,8'hb4,8'h0f);
    //Read(DPI,24'h000100,8'hb4,8'h08,8'h0f);
    FastRead(DPI,24'h000100,8'hb4,8'h08,8'h0f);

    //Read(DPI,24'h000000,8'ha3,8'h08,8'h0f);
    FastRead(DPI,24'h000000,8'ha3,8'h08,8'h0f);

  end
   
  else if(Mode == QPI)
  begin
    WriteEnable(QPI);
    WriteStatus(QPI,8'h00);
    ReadStatus(QPI,8'h00);

    WriteEnable(QPI);
    WriteConfig(QPI,8'h00,8'h08,8'h00,8'h00);
    ReadConfig(QPI,8'h00,8'h08,8'h00,8'h00);

    WriteEnable(QPI);
    //Write(QPI,24'h000200,8'hc5,8'h0f);
    FastWrite(QPI,24'h000200,8'hc5,8'h0f);

    //Read(QPI,24'h000200,8'hc5,8'h08,8'h0f);
    FastRead(QPI,24'h000200,8'hc5,8'h08,8'h0f);

    //Read(QPI,24'h000100,8'hb4,8'h08,8'h0f);
    FastRead(QPI,24'h000100,8'hb4,8'h08,8'h0f);

    //Read(QPI,24'h000000,8'ha3,8'h08,8'h0f);
    FastRead(QPI,24'h000000,8'ha3,8'h08,8'h0f);

    WriteEnable(QPI);
    WriteConfig(QPI,8'h00,8'h0c,8'h00,8'h00);
    ReadConfig(QPI,8'h00,8'h0c,8'h00,8'h00);


    //Read(QPI,24'h000200,8'hc5,8'h0c,8'h0f);
    FastRead(QPI,24'h000200,8'hc5,8'h0c,8'h0f);

    //Read(QPI,24'h000100,8'hb4,8'h0c,8'h0f);
    FastRead(QPI,24'h000100,8'hb4,8'h0c,8'h0f);

    //Read(QPI,24'h000000,8'ha3,8'h0c,8'h0f);
    FastRead(QPI,24'h000000,8'ha3,8'h0c,8'h0f);

    WriteEnable(QPI);
    WriteConfig(QPI,8'h00,8'h08,8'h00,8'h00);
    ReadConfig(QPI,8'h00,8'h08,8'h00,8'h00);

    WriteEnable(QPI);
    //Write(QPI,24'h100002,8'hc5,8'h0f);
    FastWrite(QPI,24'h100002,8'hc5,8'h0f);
    //Read(QPI,24'h100002,8'hc5,8'h08,8'h0f);

    WriteEnable(QPI);
    WriteConfig(QPI,8'h00,8'h00,8'h00,8'h00);
    ReadConfig(QPI,8'h00,8'h00,8'h00,8'h00);

    SPI_Enable(QPI);
    Read(SPI,24'h100002,8'hc5,8'h00,8'h0f);

  end
end
endtask


/////////////////////////////////////////////////////////////////
// Write Read Registers Check
/////////////////////////////////////////////////////////////////
task Register_Write_Read_Check();
input [2:0] Mode;
begin
  parameter SPI = 3'h1;
  parameter DPI = 3'h2;
  parameter QPI = 3'h4;
  parameter REG_SR = 24'h0;
  parameter REG_CR1 = 24'h2;
  parameter REG_CR2 = 24'h3;
  parameter REG_CR3 = 24'h4;
  parameter REG_CR4 = 24'h5;
  $display("");
  $display("*****************************************************************");
  $display("** Write Read Registers check ");
  $display("*****************************************************************");

  $display("");
  $display("*** Status register check ***");
  /* SR: All 1's */
  $display("");
  $display("**** Write 0xff ****");
  WriteEnable(Mode);
  WriteReg(Mode, REG_SR, 8'hff); 
  ReadReg(Mode, REG_SR, 8'hff);
  ReadStatus(Mode, 8'hff);
  /* SR: All 0's */
  $display("");
  $display("**** Write 0x00 ****");
  WriteEnable(Mode);
  WriteReg(Mode, REG_SR, 8'h00); 
  ReadReg(Mode, REG_SR, 8'h00);
  ReadStatus(Mode, 8'h00);

`ifndef SIM_CORRECTONLY
  /* SR: All 1's, no write enable */
  $display("");
  $display("**** Write 0xff, no write enable ****");
  WriteReg(Mode, REG_SR, 8'hff); 
  ReadReg(Mode, REG_SR, 8'hff);
  ReadStatus(Mode, 8'hff);
`endif

  $display("");
  $display("*** Config register check ***");
  /* CR1 to CR4: All 1's */
  $display("");
  $display("**** Write all 0xff ****");
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR1, 8'hff);
  ReadReg(Mode, REG_CR1, 8'hff);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR2, 8'hff);
  ReadReg(Mode, REG_CR2, 8'hff);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR3, 8'hff);
  ReadReg(Mode, REG_CR3, 8'hff);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR4, 8'hff);
  ReadReg(Mode, REG_CR4, 8'hff);
  ReadConfig(Mode, 8'hff, 8'hff, 8'hff, 8'hff);

  /* CR1 to CR4: All 0's */
  $display("");
  $display("**** Write all 0x00 ****");
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR1, 8'h00);
  ReadReg(Mode, REG_CR1, 8'h00);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR2, 8'h00);
  ReadReg(Mode, REG_CR2, 8'h00);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR3, 8'h00);
  ReadReg(Mode, REG_CR3, 8'h00);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR4, 8'h00);
  ReadReg(Mode, REG_CR4, 8'h00);
  ReadConfig(Mode, 8'h00, 8'h00, 8'h00, 8'h00);

`ifndef SIM_CORRECTONLY
  /* CR1 to CR4: All 1's, no write enables */
  $display("");
  $display("**** Write 0xff, no write enables ****");
  WriteReg(Mode, REG_CR1, 8'hff);
  ReadReg(Mode, REG_CR1, 8'hff);
  WriteReg(Mode, REG_CR2, 8'hff);
  ReadReg(Mode, REG_CR2, 8'hff);
  WriteReg(Mode, REG_CR3, 8'hff);
  ReadReg(Mode, REG_CR3, 8'hff);
  WriteReg(Mode, REG_CR4, 8'hff);
  ReadReg(Mode, REG_CR4, 8'hff);
  ReadConfig(Mode, 8'hff, 8'hff, 8'hff, 8'hff);
`endif

  /* CR1 to CR4: 0x00, 0x08, 0xaa, 0x55*/
  $display("");
  //$display("**** Write 0x00, 0x08, 0xaa, 0x55 ****");
  $display("**** Write 0xaf, 0x08, 0xaa, 0x55 ****");
  WriteEnable(Mode);
  //WriteReg(Mode, REG_CR1, 8'h00);
  //ReadReg(Mode, REG_CR1, 8'h00);
  WriteReg(Mode, REG_CR1, 8'hAF);
  ReadReg(Mode, REG_CR1, 8'hAF);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR2, 8'h08);
  ReadReg(Mode, REG_CR2, 8'h08);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR3, 8'haa);
  ReadReg(Mode, REG_CR3, 8'haa);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR4, 8'h55);
  ReadReg(Mode, REG_CR4, 8'h55);
  //ReadConfig(Mode, 8'h00, 8'h08, 8'haa, 8'h55);
  ReadConfig(Mode, 8'haf, 8'h08, 8'haa, 8'h55);

  /* CR1 to CR4: All 0's */
  $display("");
  $display("**** Write all 0x00 ****");
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR1, 8'h00);
  ReadReg(Mode, REG_CR1, 8'h00);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR2, 8'h00);
  ReadReg(Mode, REG_CR2, 8'h00);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR3, 8'h00);
  ReadReg(Mode, REG_CR3, 8'h00);
  WriteEnable(Mode);
  WriteReg(Mode, REG_CR4, 8'h00);
  ReadReg(Mode, REG_CR4, 8'h00);
  ReadConfig(Mode, 8'h00, 8'h00, 8'h00, 8'h00);
end
endtask

